open BinNums
open Coqlib
open Errors
open Floats
open Integers
open List0

type ident = positive

let ident_eq :  positive -> positive -> bool =
fun x0 x1  -> peq x0 x1 

type typ =
| Tint
| Tfloat

let typ_eq :  typ -> typ -> bool =
  fun t1 t2  ->
  match t1 with
  | Tint ->
    (match t2 with
     | Tint -> true
     | Tfloat -> false)
  | Tfloat ->
    (match t2 with
     | Tint -> false
     | Tfloat -> true)

let opt_typ_eq :  typ option -> typ option -> bool =
  fun t1 t2  ->
  match t1 with
  | Some x ->
    (match t2 with
     | Some t -> typ_eq x t
     | None -> false)
  | None ->
    (match t2 with
     | Some t -> false
     | None -> true)

type signature = { sig_args : typ list; sig_res : typ option }

let sig_args :  signature -> typ list =
  fun x  ->
    x.sig_args

let sig_res :  signature -> typ option =
  fun x  ->
    x.sig_res

let proj_sig_res :  signature -> typ =
  fun s  ->
  match s.sig_res with
  | Some t -> t
  | None -> Tint

type memory_chunk =
| Mint8signed
| Mint8unsigned
| Mint16signed
| Mint16unsigned
| Mint32
| Mfloat32
| Mfloat64
| Mfloat64al32

let type_of_chunk :  memory_chunk -> typ =
    function
| Mfloat32 -> Tfloat
| Mfloat64 -> Tfloat
| Mfloat64al32 -> Tfloat
| _ -> Tint

type init_data =
| Init_int8 of Int.int
| Init_int16 of Int.int
| Init_int32 of Int.int
| Init_float32 of float
| Init_float64 of float
| Init_space of coq_Z
| Init_addrof of ident * Int.int

type 'v globvar = { gvar_info : 'v; gvar_init : init_data list;
                    gvar_readonly : bool; gvar_volatile : bool }

let gvar_info :  'a1 globvar -> 'a1 =
  fun x  ->
    x.gvar_info

let gvar_init :  'a1 globvar -> init_data list =
  fun x  ->
    x.gvar_init

let gvar_readonly :  'a1 globvar -> bool =
  fun x  ->
    x.gvar_readonly

let gvar_volatile :  'a1 globvar -> bool =
  fun x  ->
    x.gvar_volatile

type ('f, 'v) globdef =
| Gfun of 'f
| Gvar of 'v globvar

type ('f, 'v) program = { prog_defs : (ident * ('f, 'v) globdef) list;
                          prog_main : ident }

let prog_defs :     ('a1, 'a2) program -> (ident * ('a1, 'a2) globdef) list =
  fun x  ->
    x.prog_defs

let prog_main :  ('a1, 'a2) program -> ident =
  fun x  ->
    x.prog_main

let transform_program_globdef :     ('a1 -> 'a2) -> (ident * ('a1, 'a3) globdef) -> ident * ('a2, 'a3)
    globdef =
  fun transf  ->
    function
| (id, g) ->
  (match g with
   | Gfun f -> (id, (Gfun (transf f)))
   | Gvar v -> (id, (Gvar v)))

let transform_program :     ('a1 -> 'a2) -> ('a1, 'a3) program -> ('a2, 'a3) program =
  fun transf p  ->
  { prog_defs = (map (transform_program_globdef transf) p.prog_defs);
    prog_main = p.prog_main }

let transf_globvar :     ('a1 -> 'a2 res) -> 'a1 globvar -> 'a2 globvar res =
  fun transf_var g  ->
  match transf_var g.gvar_info with
  | OK x ->
    OK { gvar_info = x; gvar_init = g.gvar_init; gvar_readonly =
      g.gvar_readonly; gvar_volatile = g.gvar_volatile }
  | Error msg -> Error msg

let rec transf_globdefs :     ('a1 -> 'a2 res) -> ('a3 -> 'a4 res) -> (ident * ('a1, 'a3) globdef) list
    -> (ident * ('a2, 'a4) globdef) list res =
  fun transf_fun transf_var  ->
    function
| [] -> OK []
| p :: l' ->
  let (id, g) = p in
  (match g with
   | Gfun f ->
     (match transf_fun f with
      | OK tf ->
        (match transf_globdefs transf_fun transf_var l' with
         | OK x -> OK ((id, (Gfun tf)) :: x)
         | Error msg -> Error msg)
      | Error msg ->
        Error ((MSG
          ('I'::('n'::(' '::('f'::('u'::('n'::('c'::('t'::('i'::('o'::('n'::(' '::[]))))))))))))) :: ((CTX
          id) :: ((MSG (':'::(' '::[]))) :: msg))))
   | Gvar v ->
     (match transf_globvar transf_var v with
      | OK tv ->
        (match transf_globdefs transf_fun transf_var l' with
         | OK x -> OK ((id, (Gvar tv)) :: x)
         | Error msg -> Error msg)
      | Error msg ->
        Error ((MSG
          ('I'::('n'::(' '::('v'::('a'::('r'::('i'::('a'::('b'::('l'::('e'::(' '::[]))))))))))))) :: ((CTX
          id) :: ((MSG (':'::(' '::[]))) :: msg)))))

let transform_partial_program2 :     ('a1 -> 'a2 res) -> ('a3 -> 'a4 res) -> ('a1, 'a3) program -> ('a2, 'a4)
    program res =
  fun transf_fun transf_var p  ->
  match transf_globdefs transf_fun transf_var p.prog_defs with
  | OK x -> OK { prog_defs = x; prog_main = p.prog_main }
  | Error msg -> Error msg

let transform_partial_program :     ('a1 -> 'a2 res) -> ('a1, 'a3) program -> ('a2, 'a3) program res =
  fun transf_partial p  ->
  transform_partial_program2 transf_partial (fun v -> OK v) p

type external_function =
| EF_external of ident * signature
| EF_builtin of ident * signature
| EF_vload of memory_chunk
| EF_vstore of memory_chunk
| EF_vload_global of memory_chunk * ident * Int.int
| EF_vstore_global of memory_chunk * ident * Int.int
| EF_malloc
| EF_free
| EF_memcpy of coq_Z * coq_Z
| EF_annot of ident * annot_arg list
| EF_annot_val of ident * typ
| EF_inline_asm of ident
and annot_arg =
| AA_arg of typ
| AA_int of Int.int
| AA_float of float

let rec annot_args_typ :  annot_arg list -> typ list =
    function
| [] -> []
| a :: targs' ->
  (match a with
   | AA_arg ty -> ty :: (annot_args_typ targs')
   | _ -> annot_args_typ targs')

let ef_sig :  external_function -> signature =
    function
| EF_external (name, sg) -> sg
| EF_builtin (name, sg) -> sg
| EF_vload chunk ->
  { sig_args = (Tint :: []); sig_res = (Some (type_of_chunk chunk)) }
| EF_vstore chunk ->
  { sig_args = (Tint :: ((type_of_chunk chunk) :: [])); sig_res = None }
| EF_vload_global (chunk, id, ofs) ->
  { sig_args = []; sig_res = (Some (type_of_chunk chunk)) }
| EF_vstore_global (chunk, id, ofs) ->
  { sig_args = ((type_of_chunk chunk) :: []); sig_res = None }
| EF_malloc -> { sig_args = (Tint :: []); sig_res = (Some Tint) }
| EF_free -> { sig_args = (Tint :: []); sig_res = None }
| EF_memcpy (sz, al) -> { sig_args = (Tint :: (Tint :: [])); sig_res = None }
| EF_annot (text, targs) ->
  { sig_args = (annot_args_typ targs); sig_res = None }
| EF_annot_val (text, targ) ->
  { sig_args = (targ :: []); sig_res = (Some targ) }
| EF_inline_asm text -> { sig_args = []; sig_res = None }

let ef_inline :  external_function -> bool =
    function
| EF_external (name, sg) -> false
| EF_malloc -> false
| EF_free -> false
| _ -> true

let ef_reloads :  external_function -> bool =
    function
| EF_annot (text, targs) -> false
| _ -> true

type 'f fundef =
| Internal of 'f
| External of external_function

let transf_fundef :  ('a1 -> 'a2) -> 'a1 fundef -> 'a2 fundef =
  fun transf  ->
    function
| Internal f -> Internal (transf f)
| External ef -> External ef

let transf_partial_fundef :     ('a1 -> 'a2 res) -> 'a1 fundef -> 'a2 fundef res =
  fun transf_partial  ->
    function
| Internal f ->
  (match transf_partial f with
   | OK x -> OK (Internal x)
   | Error msg -> Error msg)
| External ef -> OK (External ef)

