open AST
open BinNums
open Coloring
open Coqlib
open Datatypes
open Errors
open LTL
open Lattice
open List0
open Locations
open Maps
open Op
open RTL
open RTLtyping
open Registers

let reg_option_live :  reg option -> Regset.t -> Regset.t =
  fun or0 lv  ->
  match or0 with
  | Some r -> Regset.add r lv
  | None -> lv

let reg_sum_live :  (reg, ident) sum -> Regset.t -> Regset.t =
  fun ros lv  ->
  match ros with
  | Coq_inl r -> Regset.add r lv
  | Coq_inr s -> lv

let rec reg_list_live :  reg list -> Regset.t -> Regset.t =
  fun rl lv  ->
  match rl with
  | [] -> lv
  | r1 :: rs -> reg_list_live rs (Regset.add r1 lv)

let transfer :  coq_function -> node -> Regset.t -> Regset.t =
  fun f pc after  ->
  match PTree.get pc f.fn_code with
  | Some i ->
    (match i with
     | Inop s -> after
     | Iop (op, args, res0, s) ->
       if Regset.mem res0 after
       then reg_list_live args (Regset.remove res0 after)
       else after
     | Iload (chunk, addr, args, dst, s) ->
       if Regset.mem dst after
       then reg_list_live args (Regset.remove dst after)
       else after
     | Istore (chunk, addr, args, src, s) ->
       reg_list_live args (Regset.add src after)
     | Icall (sig0, ros, args, res0, s) ->
       reg_list_live args (reg_sum_live ros (Regset.remove res0 after))
     | Itailcall (sig0, ros, args) ->
       reg_list_live args (reg_sum_live ros Regset.empty)
     | Ibuiltin (ef, args, res0, s) ->
       reg_list_live args (Regset.remove res0 after)
     | Icond (cond, args, ifso, ifnot) -> reg_list_live args after
     | Ijumptable (arg, tbl) -> Regset.add arg after
     | Ireturn optarg -> reg_option_live optarg Regset.empty)
  | None -> Regset.empty

module DS = 
 struct 
  module L = 
   struct 
    type t = Regset.t
    
    (** val beq : t -> t -> bool **)
    
    let beq x y =
      Regset.equal x y
    
    (** val bot : t **)
    
    let bot =
      Regset.empty
    
    (** val lub : t -> t -> t **)
    
    let lub x y =
      Regset.union x y
   end
  
  (** val fixpoint :
      positive list PTree.t -> (positive -> L.t -> L.t) -> (positive * L.t)
      list -> L.t PMap.t option **)
  
  let fixpoint :
      positive list PTree.t -> (positive -> L.t -> L.t) -> (positive * L.t)
      list -> L.t PMap.t option =
    fun x y z ->
    failwith "AXIOM TO BE REALIZED"
 end

let analyze :  coq_function -> Regset.t PMap.t option =
  fun f  ->
  DS.fixpoint (successors f) (transfer f) []

let is_redundant_move :     operation -> reg list -> reg -> (reg -> loc) -> bool =
  fun op args res0 assign  ->
  match is_move_operation op args with
  | Some src -> if Loc.eq (assign src) (assign res0) then true else false
  | None -> false

let transf_instr :     coq_function -> Regset.t PMap.t -> (reg -> loc) -> LTL.node ->
    instruction -> LTL.instruction =
  fun f live assign pc  ->
    function
| Inop s -> Lnop s
| Iop (op, args, res0, s) ->
  if Regset.mem res0 (PMap.get pc live)
  then if is_redundant_move op args res0 assign
       then Lnop s
       else Lop (op, (map assign args), (assign res0), s)
  else Lnop s
| Iload (chunk, addr, args, dst, s) ->
  if Regset.mem dst (PMap.get pc live)
  then Lload (chunk, addr, (map assign args), (assign dst), s)
  else Lnop s
| Istore (chunk, addr, args, src, s) ->
  Lstore (chunk, addr, (map assign args), (assign src), s)
| Icall (sig0, ros, args, res0, s) ->
  Lcall (sig0, (sum_left_map assign ros), (map assign args), (assign res0),
    s)
| Itailcall (sig0, ros, args) ->
  Ltailcall (sig0, (sum_left_map assign ros), (map assign args))
| Ibuiltin (ef, args, res0, s) ->
  Lbuiltin (ef, (map assign args), (assign res0), s)
| Icond (cond, args, ifso, ifnot) ->
  Lcond (cond, (map assign args), ifso, ifnot)
| Ijumptable (arg, tbl) -> Ljumptable ((assign arg), tbl)
| Ireturn optarg -> Lreturn (option_map assign optarg)

let transf_fun :     coq_function -> Regset.t PMap.t -> (reg -> loc) -> LTL.coq_function =
  fun f live assign  ->
  { LTL.fn_sig = f.fn_sig; LTL.fn_params = (map assign f.fn_params);
    LTL.fn_stacksize = f.fn_stacksize; LTL.fn_code =
    (PTree.map (transf_instr f live assign) f.fn_code); LTL.fn_entrypoint =
    f.fn_entrypoint }

let live0 :  coq_function -> Regset.t PMap.t -> Regset.t =
  fun f live  ->
  transfer f f.fn_entrypoint (PMap.get f.fn_entrypoint live)

let transf_function :  coq_function -> LTL.coq_function res =
  fun f  ->
  match type_function f with
  | OK env ->
    (match analyze f with
     | Some live ->
       (match regalloc f live (live0 f live) env with
        | Some assign -> OK (transf_fun f live assign)
        | None ->
          Error
            (msg
              ('I'::('n'::('c'::('o'::('r'::('r'::('e'::('c'::('t'::(' '::('g'::('r'::('a'::('p'::('h'::(' '::('c'::('o'::('l'::('o'::('r'::('i'::('n'::('g'::[]))))))))))))))))))))))))))
     | None ->
       Error
         (msg
           ('L'::('i'::('v'::('e'::('n'::('e'::('s'::('s'::(' '::('a'::('n'::('a'::('l'::('y'::('s'::('i'::('s'::(' '::('f'::('a'::('i'::('l'::('u'::('r'::('e'::[])))))))))))))))))))))))))))
  | Error msg0 -> Error msg0

let transf_fundef :  fundef -> LTL.fundef res =
  fun fd  ->
  transf_partial_fundef transf_function fd

let transf_program :  program -> LTL.program res =
  fun p  ->
  transform_partial_program transf_fundef p

