open AST
open BinInt
open BinNums
open Conventions1
open Linear
open List0
open Locations
open Machregs

type bounds = { bound_int_local : coq_Z; bound_float_local : coq_Z;
                bound_int_callee_save : coq_Z;
                bound_float_callee_save : coq_Z; bound_outgoing : coq_Z;
                bound_stack_data : coq_Z }

let bound_int_local :  bounds -> coq_Z =
  fun x  ->
    x.bound_int_local

let bound_float_local :  bounds -> coq_Z =
  fun x  ->
    x.bound_float_local

let bound_int_callee_save :  bounds -> coq_Z =
  fun x  ->
    x.bound_int_callee_save

let bound_float_callee_save :  bounds -> coq_Z =
  fun x  ->
    x.bound_float_callee_save

let bound_outgoing :  bounds -> coq_Z =
  fun x  ->
    x.bound_outgoing

let bound_stack_data :  bounds -> coq_Z =
  fun x  ->
    x.bound_stack_data

let regs_of_instr :  instruction -> mreg list =
    function
| Lgetstack (s, r) -> r :: []
| Lsetstack (r, s) -> r :: []
| Lop (op, args, res) -> res :: []
| Lload (chunk, addr, args, dst) -> dst :: []
| Lbuiltin (ef, args, res) -> res :: []
| _ -> []

let rec slots_of_locs :  loc list -> slot list =
    function
| [] -> []
| l0 :: l' ->
  (match l0 with
   | R r -> slots_of_locs l'
   | S s -> s :: (slots_of_locs l'))

let slots_of_instr :  instruction -> slot list =
    function
| Lgetstack (s, r) -> s :: []
| Lsetstack (r, s) -> s :: []
| Lannot (ef, args) -> slots_of_locs args
| _ -> []

let max_over_list :  ('a1 -> coq_Z) -> 'a1 list -> coq_Z =
  fun valu l  ->
  fold_left (fun m l0 -> Z.max m (valu l0)) l Z0

let max_over_instrs :  coq_function -> (instruction -> coq_Z) -> coq_Z =
  fun f valu  ->
  max_over_list valu f.fn_code

let max_over_regs_of_instr :  (mreg -> coq_Z) -> instruction -> coq_Z =
  fun valu i  ->
  max_over_list valu (regs_of_instr i)

let max_over_slots_of_instr :  (slot -> coq_Z) -> instruction -> coq_Z =
  fun valu i  ->
  max_over_list valu (slots_of_instr i)

let max_over_regs_of_funct :  coq_function -> (mreg -> coq_Z) -> coq_Z =
  fun f valu  ->
  max_over_instrs f (max_over_regs_of_instr valu)

let max_over_slots_of_funct :     coq_function -> (slot -> coq_Z) -> coq_Z =
  fun f valu  ->
  max_over_instrs f (max_over_slots_of_instr valu)

let int_callee_save :  mreg -> coq_Z =
  fun r  ->
  Z.add (Zpos Coq_xH) (index_int_callee_save r)

let float_callee_save :  mreg -> coq_Z =
  fun r  ->
  Z.add (Zpos Coq_xH) (index_float_callee_save r)

let int_local :  slot -> coq_Z =
    function
| Local (ofs, t) ->
  (match t with
   | Tint -> Z.add (Zpos Coq_xH) ofs
   | Tfloat -> Z0)
| _ -> Z0

let float_local :  slot -> coq_Z =
    function
| Local (ofs, t) ->
  (match t with
   | Tint -> Z0
   | Tfloat -> Z.add (Zpos Coq_xH) ofs)
| _ -> Z0

let outgoing_slot :  slot -> coq_Z =
    function
| Outgoing (ofs, ty) -> Z.add ofs (typesize ty)
| _ -> Z0

let outgoing_space :  instruction -> coq_Z =
    function
| Lcall (sig0, s) -> size_arguments sig0
| _ -> Z0

let function_bounds :  coq_function -> bounds =
  fun f  ->
  { bound_int_local = (max_over_slots_of_funct f int_local);
    bound_float_local = (max_over_slots_of_funct f float_local);
    bound_int_callee_save = (max_over_regs_of_funct f int_callee_save);
    bound_float_callee_save = (max_over_regs_of_funct f float_callee_save);
    bound_outgoing =
    (Z.max (max_over_instrs f outgoing_space)
      (max_over_slots_of_funct f outgoing_slot)); bound_stack_data =
    (Z.max f.fn_stacksize Z0) }

