open AST
open BinNums
open BinPos
open CombineOp
open Coqlib
open Datatypes
open Errors
open Kildall
open List0
open Maps
open Op
open RTL
open RTLtyping
open Registers

let eq_valnum :  valnum -> valnum -> bool =
fun x0 x1  -> peq x0 x1 

let rec eq_list_valnum :  valnum list -> valnum list -> bool =
  fun l y0  ->
  match l with
  | [] ->
    (match y0 with
     | [] -> true
     | v :: l0 -> false)
  | y :: l0 ->
    (match y0 with
     | [] -> false
     | v :: l1 -> if eq_valnum y v then eq_list_valnum l0 l1 else false)

let eq_rhs :  rhs -> rhs -> bool =
  fun x y  ->
  let h3 = fun x0 y0 ->
    match x0 with
    | Mint8signed ->
      (match y0 with
       | Mint8signed -> true
       | _ -> false)
    | Mint8unsigned ->
      (match y0 with
       | Mint8unsigned -> true
       | _ -> false)
    | Mint16signed ->
      (match y0 with
       | Mint16signed -> true
       | _ -> false)
    | Mint16unsigned ->
      (match y0 with
       | Mint16unsigned -> true
       | _ -> false)
    | Mint32 ->
      (match y0 with
       | Mint32 -> true
       | _ -> false)
    | Mfloat32 ->
      (match y0 with
       | Mfloat32 -> true
       | _ -> false)
    | Mfloat64 ->
      (match y0 with
       | Mfloat64 -> true
       | _ -> false)
    | Mfloat64al32 ->
      (match y0 with
       | Mfloat64al32 -> true
       | _ -> false)
  in
  (match x with
   | Op (x0, x1) ->
     (match y with
      | Op (o0, l0) ->
        if eq_operation x0 o0 then eq_list_valnum x1 l0 else false
      | Load (m, a, l0) -> false)
   | Load (x0, x1, x2) ->
     (match y with
      | Op (o, l0) -> false
      | Load (m0, a0, l0) ->
        if h3 x0 m0
        then if eq_addressing x1 a0 then eq_list_valnum x2 l0 else false
        else false))

type numbering = { num_next : valnum; num_eqs : (valnum * rhs) list;
                   num_reg : valnum PTree.t; num_val : reg list PMap.t }

let num_next :  numbering -> valnum =
  fun x  ->
    x.num_next

let num_eqs :  numbering -> (valnum * rhs) list =
  fun x  ->
    x.num_eqs

let num_reg :  numbering -> valnum PTree.t =
  fun x  ->
    x.num_reg

let num_val :  numbering -> reg list PMap.t =
  fun x  ->
    x.num_val

let empty_numbering :  numbering =
  { num_next = Coq_xH; num_eqs = []; num_reg = PTree.empty; num_val =
    (PMap.init []) }

let valnum_reg :  numbering -> reg -> numbering * valnum =
  fun n r  ->
  match PTree.get r n.num_reg with
  | Some v -> (n, v)
  | None ->
    let v = n.num_next in
    ({ num_next = (Pos.succ v); num_eqs = n.num_eqs; num_reg =
    (PTree.set r v n.num_reg); num_val = (PMap.set v (r :: []) n.num_val) },
    v)

let rec valnum_regs :  numbering -> reg list -> numbering * valnum list =
  fun n  ->
    function
| [] -> (n, [])
| r1 :: rs ->
  let (n1, v1) = valnum_reg n r1 in
  let (ns, vs) = valnum_regs n1 rs in (ns, (v1 :: vs))

let rec find_valnum_rhs :  rhs -> (valnum * rhs) list -> valnum option =
  fun r  ->
    function
| [] -> None
| p :: eqs1 ->
  let (v, r') = p in if eq_rhs r r' then Some v else find_valnum_rhs r eqs1

let rec find_valnum_num :  valnum -> (valnum * rhs) list -> rhs option =
  fun v  ->
    function
| [] -> None
| p :: eqs1 ->
  let (v', r') = p in if peq v v' then Some r' else find_valnum_num v eqs1

let forget_reg :  numbering -> reg -> reg list PMap.t =
  fun n rd  ->
  match PTree.get rd n.num_reg with
  | Some v -> PMap.set v (remove peq rd (PMap.get v n.num_val)) n.num_val
  | None -> n.num_val

let update_reg :  numbering -> reg -> valnum -> reg list PMap.t =
  fun n rd vn  ->
  let nv = forget_reg n rd in PMap.set vn (rd :: (PMap.get vn nv)) nv

let add_rhs :  numbering -> reg -> rhs -> numbering =
  fun n rd rh  ->
  match find_valnum_rhs rh n.num_eqs with
  | Some vres ->
    { num_next = n.num_next; num_eqs = n.num_eqs; num_reg =
      (PTree.set rd vres n.num_reg); num_val = (update_reg n rd vres) }
  | None ->
    { num_next = (Pos.succ n.num_next); num_eqs = ((n.num_next,
      rh) :: n.num_eqs); num_reg = (PTree.set rd n.num_next n.num_reg);
      num_val = (update_reg n rd n.num_next) }

let add_op :  numbering -> reg -> operation -> reg list -> numbering =
  fun n rd op rs  ->
  match is_move_operation op rs with
  | Some r ->
    let (n1, v) = valnum_reg n r in
    { num_next = n1.num_next; num_eqs = n1.num_eqs; num_reg =
    (PTree.set rd v n1.num_reg); num_val = (update_reg n1 rd v) }
  | None -> let (n1, vs) = valnum_regs n rs in add_rhs n1 rd (Op (op, vs))

let add_load :     numbering -> reg -> memory_chunk -> addressing -> reg list -> numbering =
  fun n rd chunk addr rs  ->
  let (n1, vs) = valnum_regs n rs in add_rhs n1 rd (Load (chunk, addr, vs))

let add_unknown :  numbering -> reg -> numbering =
  fun n rd  ->
  { num_next = (Pos.succ n.num_next); num_eqs = n.num_eqs; num_reg =
    (PTree.set rd n.num_next n.num_reg); num_val = (forget_reg n rd) }

let rec kill_eqs :     (rhs -> bool) -> (valnum * rhs) list -> (valnum * rhs) list =
  fun pred  ->
    function
| [] -> []
| eq :: rem ->
  if pred (snd eq) then kill_eqs pred rem else eq :: (kill_eqs pred rem)

let kill_equations :  (rhs -> bool) -> numbering -> numbering =
  fun pred n  ->
  { num_next = n.num_next; num_eqs = (kill_eqs pred n.num_eqs); num_reg =
    n.num_reg; num_val = n.num_val }

let filter_loads :  rhs -> bool =
    function
| Op (op, l) -> op_depends_on_memory op
| Load (m, a, l) -> true

let kill_loads :  numbering -> numbering =
  fun n  ->
  kill_equations filter_loads n

let filter_after_store :     memory_chunk -> addressing -> valnum list -> rhs -> bool =
  fun chunk addr vl  ->
    function
| Op (op, vl') -> op_depends_on_memory op
| Load (chunk', addr', vl') ->
  negb
    ((&&) (proj_sumbool (eq_list_valnum vl vl'))
      (addressing_separated chunk addr chunk' addr'))

let add_store :     numbering -> memory_chunk -> addressing -> reg list -> reg -> numbering =
  fun n chunk addr rargs rsrc  ->
  let (n1, vargs) = valnum_regs n rargs in
  let n2 = kill_equations (filter_after_store chunk addr vargs) n1 in
  (match chunk with
   | Mint32 -> add_rhs n2 rsrc (Load (chunk, addr, vargs))
   | Mfloat64 -> add_rhs n2 rsrc (Load (chunk, addr, vargs))
   | _ -> n2)

let reg_valnum :  numbering -> valnum -> reg option =
  fun n vn  ->
  match PMap.get vn n.num_val with
  | [] -> None
  | r :: rs -> Some r

let rec regs_valnums :  numbering -> valnum list -> reg list option =
  fun n  ->
    function
| [] -> Some []
| v1 :: vs ->
  (match reg_valnum n v1 with
   | Some r1 ->
     (match regs_valnums n vs with
      | Some rs -> Some (r1 :: rs)
      | None -> None)
   | None -> None)

let find_rhs :  numbering -> rhs -> reg option =
  fun n rh  ->
  match find_valnum_rhs rh n.num_eqs with
  | Some vres -> reg_valnum n vres
  | None -> None

let rec reduce_rec :     ((valnum -> rhs option) -> 'a1 -> valnum list -> ('a1 * valnum list)
    option) -> numbering -> nat -> 'a1 -> valnum list -> ('a1 * reg list)
    option =
  fun f n niter op args  ->
  match niter with
  | O -> None
  | S niter' ->
    (match f (fun v -> find_valnum_num v n.num_eqs) op args with
     | Some p ->
       let (op', args') = p in
       (match reduce_rec f n niter' op' args' with
        | Some p0 -> Some p0
        | None ->
          (match regs_valnums n args' with
           | Some rl -> Some (op', rl)
           | None -> None))
     | None -> None)

let reduce :     ((valnum -> rhs option) -> 'a1 -> valnum list -> ('a1 * valnum list)
    option) -> numbering -> 'a1 -> reg list -> valnum list -> 'a1 * reg list =
  fun f n op rl vl  ->
  match reduce_rec f n (S (S (S (S O)))) op vl with
  | Some res0 -> res0
  | None -> (op, rl)

module Solver = 
 struct 
  module L = 
   struct 
    type t = numbering
    
    (** val top : numbering **)
    
    let top =
      empty_numbering
   end
  
  (** val fixpoint :
      positive list PTree.t -> (positive -> L.t -> L.t) -> positive -> L.t
      PMap.t option **)
  
  let fixpoint :
      positive list PTree.t -> (positive -> L.t -> L.t) -> positive -> L.t
      PMap.t option =
    fun x y z ->
    failwith "AXIOM TO BE REALIZED"
 end

let transfer :  coq_function -> node -> numbering -> numbering =
  fun f pc before  ->
  match PTree.get pc f.fn_code with
  | Some i ->
    (match i with
     | Iop (op, args, res0, s) -> add_op before res0 op args
     | Iload (chunk, addr, args, dst, s) ->
       add_load before dst chunk addr args
     | Istore (chunk, addr, args, src, s) ->
       add_store before chunk addr args src
     | Icall (sig0, ros, args, res0, s) -> empty_numbering
     | Itailcall (sig0, ros, args) -> empty_numbering
     | Ibuiltin (ef, args, res0, s) -> add_unknown (kill_loads before) res0
     | _ -> before)
  | None -> before

let analyze :  coq_function -> numbering PMap.t option =
  fun f  ->
  Solver.fixpoint (successors f) (transfer f) f.fn_entrypoint

let transf_instr :  numbering -> instruction -> instruction =
  fun n instr  ->
    match instr with
| Iop (op, args, res0, s) ->
  if is_trivial_op op
  then instr
  else let (n1, vl) = valnum_regs n args in
       (match find_rhs n1 (Op (op, vl)) with
        | Some r -> Iop (Omove, (r :: []), res0, s)
        | None ->
          let (op', args') = reduce combine_op n1 op args vl in
          Iop (op', args', res0, s))
| Iload (chunk, addr, args, dst, s) ->
  let (n1, vl) = valnum_regs n args in
  (match find_rhs n1 (Load (chunk, addr, vl)) with
   | Some r -> Iop (Omove, (r :: []), dst, s)
   | None ->
     let (addr', args') = reduce combine_addr n1 addr args vl in
     Iload (chunk, addr', args', dst, s))
| Istore (chunk, addr, args, src, s) ->
  let (n1, vl) = valnum_regs n args in
  let (addr', args') = reduce combine_addr n1 addr args vl in
  Istore (chunk, addr', args', src, s)
| Icond (cond, args, s1, s2) ->
  let (n1, vl) = valnum_regs n args in
  let (cond', args') = reduce combine_cond n1 cond args vl in
  Icond (cond', args', s1, s2)
| _ -> instr

let transf_code :  numbering PMap.t -> code -> code =
  fun approxs instrs  ->
  PTree.map (fun pc instr -> transf_instr (PMap.get pc approxs) instr) instrs

let transf_function :  coq_function -> coq_function res =
  fun f  ->
  match type_function f with
  | OK tyenv ->
    (match analyze f with
     | Some approxs ->
       OK { fn_sig = f.fn_sig; fn_params = f.fn_params; fn_stacksize =
         f.fn_stacksize; fn_code = (transf_code approxs f.fn_code);
         fn_entrypoint = f.fn_entrypoint }
     | None ->
       Error
         (msg
           ('C'::('S'::('E'::(' '::('f'::('a'::('i'::('l'::('u'::('r'::('e'::[])))))))))))))
  | Error msg0 -> Error msg0

let transf_fundef :  fundef -> fundef res =
  fun f  ->
  transf_partial_fundef transf_function f

let transf_program :  program -> program res =
  fun p  ->
  transform_partial_program transf_fundef p

