open AST
open Cop
open Ctypes
open Datatypes
open Floats
open Integers
open List0

type expr =
| Econst_int of Int.int * coq_type
| Econst_float of float * coq_type
| Evar of ident * coq_type
| Etempvar of ident * coq_type
| Ederef of expr * coq_type
| Eaddrof of expr * coq_type
| Eunop of unary_operation * expr * coq_type
| Ebinop of binary_operation * expr * expr * coq_type
| Ecast of expr * coq_type
| Efield of expr * ident * coq_type

let coq_Esizeof :  coq_type -> coq_type -> expr =
  fun ty' ty  ->
  Econst_int ((Int.repr (sizeof ty')), ty)

let coq_Ealignof :  coq_type -> coq_type -> expr =
  fun ty' ty  ->
  Econst_int ((Int.repr (alignof ty')), ty)

let typeof :  expr -> coq_type =
    function
| Econst_int (i, ty) -> ty
| Econst_float (f, ty) -> ty
| Evar (i, ty) -> ty
| Etempvar (i, ty) -> ty
| Ederef (e0, ty) -> ty
| Eaddrof (e0, ty) -> ty
| Eunop (u, e0, ty) -> ty
| Ebinop (b, e0, e1, ty) -> ty
| Ecast (e0, ty) -> ty
| Efield (e0, i, ty) -> ty

type label = ident

type statement =
| Sskip
| Sassign of expr * expr
| Sset of ident * expr
| Scall of ident option * expr * expr list
| Sbuiltin of ident option * external_function * typelist * expr list
| Ssequence of statement * statement
| Sifthenelse of expr * statement * statement
| Sloop of statement * statement
| Sbreak
| Scontinue
| Sreturn of expr option
| Sswitch of expr * labeled_statements
| Slabel of label * statement
| Sgoto of label
and labeled_statements =
| LSdefault of statement
| LScase of Int.int * statement * labeled_statements

type coq_function = { fn_return : coq_type;
                      fn_params : (ident * coq_type) list;
                      fn_vars : (ident * coq_type) list;
                      fn_temps : (ident * coq_type) list; fn_body : statement }

let fn_return :  coq_function -> coq_type =
  fun x  ->
    x.fn_return

let fn_params :  coq_function -> (ident * coq_type) list =
  fun x  ->
    x.fn_params

let fn_vars :  coq_function -> (ident * coq_type) list =
  fun x  ->
    x.fn_vars

let fn_temps :  coq_function -> (ident * coq_type) list =
  fun x  ->
    x.fn_temps

let fn_body :  coq_function -> statement =
  fun x  ->
    x.fn_body

let var_names :  (ident * coq_type) list -> ident list =
  fun vars  ->
  map fst vars

type fundef =
| Internal of coq_function
| External of external_function * typelist * coq_type

type program = (fundef, coq_type) AST.program

