open AST
open BinNums
open Conventions
open Conventions1
open Coqlib
open Datatypes
open InterfGraph
open List0
open Locations
open Machregs
open Maps
open Op
open RTL
open RTLtyping
open Registers

let add_interf_live :     (reg -> bool) -> reg -> Regset.t -> graph -> graph =
  fun filter res live g  ->
  Regset.fold (fun r g0 -> if filter r then add_interf r res g0 else g0) live
    g

let add_interf_op :  reg -> Regset.t -> graph -> graph =
  fun res live g  ->
  add_interf_live (fun r -> if Reg.eq r res then false else true) res live g

let add_interf_move :  reg -> reg -> Regset.t -> graph -> graph =
  fun arg res live g  ->
  add_interf_live (fun r ->
    if Reg.eq r res then false else if Reg.eq r arg then false else true) res
    live g

let add_interf_destroyed :  Regset.t -> mreg list -> graph -> graph =
  fun live destroyed g  ->
  fold_left (fun g0 mr ->
    Regset.fold (fun r g1 -> add_interf_mreg r mr g1) live g0) destroyed g

let add_interfs_indirect_call :  reg -> loc list -> graph -> graph =
  fun rfun locs g  ->
  fold_left (fun g0 loc0 ->
    match loc0 with
    | R mr -> add_interf_mreg rfun mr g0
    | S s -> g0) locs g

let add_interf_call :  (reg, ident) sum -> loc list -> graph -> graph =
  fun ros locs g  ->
  match ros with
  | Coq_inl rfun -> add_interfs_indirect_call rfun locs g
  | Coq_inr idfun -> g

let rec add_prefs_call :  reg list -> loc list -> graph -> graph =
  fun args locs g  ->
  match args with
  | [] -> g
  | a1 :: al ->
    (match locs with
     | [] -> g
     | l1 :: ll ->
       add_prefs_call al ll
         (match l1 with
          | R mr -> add_pref_mreg a1 mr g
          | S s -> g))

let add_prefs_builtin :     external_function -> reg list -> reg -> graph -> graph =
  fun ef args res g  ->
  match ef with
  | EF_annot_val (txt, targ) ->
    (match args with
     | [] -> g
     | arg1 :: l -> add_pref arg1 res g)
  | _ -> g

let add_interf_entry :  reg list -> Regset.t -> graph -> graph =
  fun coqzz_params live g  ->
  fold_left (fun g0 r -> add_interf_op r live g0) coqzz_params g

let rec add_interf_params :  reg list -> graph -> graph =
  fun coqzz_params g  ->
  match coqzz_params with
  | [] -> g
  | p1 :: pl ->
    add_interf_params pl
      (fold_left (fun g0 r -> if Reg.eq r p1 then g0 else add_interf r p1 g0)
        pl g)

let add_edges_instr :     signature -> instruction -> Regset.t -> graph -> graph =
  fun sig0 i live g  ->
  match i with
  | Iop (op, args, res, s) ->
    if Regset.mem res live
    then (match is_move_operation op args with
          | Some arg -> add_pref arg res (add_interf_move arg res live g)
          | None -> add_interf_op res live g)
    else g
  | Iload (chunk, addr, args, dst, s) ->
    if Regset.mem dst live then add_interf_op dst live g else g
  | Icall (sig1, ros, args, res, s) ->
    let largs = loc_arguments sig1 in
    let lres = loc_result sig1 in
    add_prefs_call args largs
      (add_pref_mreg res lres
        (add_interf_op res live
          (add_interf_call ros largs
            (add_interf_destroyed (Regset.remove res live)
              destroyed_at_call_regs g))))
  | Itailcall (sig1, ros, args) ->
    let largs = loc_arguments sig1 in
    add_prefs_call args largs (add_interf_call ros largs g)
  | Ibuiltin (ef, args, res, s) ->
    add_prefs_builtin ef args res (add_interf_op res live g)
  | Ireturn o ->
    (match o with
     | Some r -> add_pref_mreg r (loc_result sig0) g
     | None -> g)
  | _ -> g

let add_edges_instrs :  coq_function -> Regset.t PMap.t -> graph =
  fun f live  ->
  PTree.fold (fun g pc i -> add_edges_instr f.fn_sig i (PMap.get pc live) g)
    f.fn_code empty_graph

let interf_graph :     coq_function -> Regset.t PMap.t -> Regset.t -> graph =
  fun f live live0  ->
  add_prefs_call f.fn_params (loc_parameters f.fn_sig)
    (add_interf_params f.fn_params
      (add_interf_entry f.fn_params live0 (add_edges_instrs f live)))

let graph_coloring :     coq_function -> graph -> regenv -> Regset.t -> reg -> loc =
  fun x y z t u ->
    Coloringaux.graph_coloring x y z t u

let check_coloring_1 :  graph -> (reg -> loc) -> bool =
  fun g coloring  ->
  SetRegReg.for_all (fun r1r2 ->
    if Loc.eq (coloring (fst r1r2)) (coloring (snd r1r2))
    then false
    else true) g.interf_reg_reg

let check_coloring_2 :  graph -> (reg -> loc) -> bool =
  fun g coloring  ->
  SetRegMreg.for_all (fun r1mr2 ->
    if Loc.eq (coloring (fst r1mr2)) (R (snd r1mr2)) then false else true)
    g.interf_reg_mreg

let same_typ :  typ -> typ -> bool =
  fun t1 t2  ->
  match t1 with
  | Tint ->
    (match t2 with
     | Tint -> true
     | Tfloat -> false)
  | Tfloat ->
    (match t2 with
     | Tint -> false
     | Tfloat -> true)

let loc_is_acceptable :  loc -> bool =
  fun l  ->
    match l with
| R r -> if in_dec Loc.eq l temporaries then false else true
| S s ->
  (match s with
   | Local (ofs, ty) -> if zlt ofs Z0 then false else true
   | _ -> false)

let check_coloring_3 :  Regset.t -> regenv -> (reg -> loc) -> bool =
  fun rs env coloring  ->
  Regset.for_all (fun r ->
    let l = coloring r in
    (&&) (loc_is_acceptable l) (same_typ (env r) (Loc.coq_type l))) rs

let check_coloring :     graph -> regenv -> Regset.t -> (reg -> loc) -> bool =
  fun g env rs coloring  ->
  (&&) (check_coloring_1 g coloring)
    ((&&) (check_coloring_2 g coloring) (check_coloring_3 rs env coloring))

let alloc_of_coloring :     (reg -> loc) -> regenv -> Regset.t -> Regset.elt -> loc =
  fun coloring env rs r  ->
  if Regset.mem r rs
  then coloring r
  else (match env r with
        | Tint -> R dummy_int_reg
        | Tfloat -> R dummy_float_reg)

let regalloc :     coq_function -> Regset.t PMap.t -> Regset.t -> regenv -> (Regset.elt ->
    loc) option =
  fun f live live0 env  ->
  let g = interf_graph f live live0 in
  let rs = all_interf_regs g in
  let coloring = graph_coloring f g env rs in
  if check_coloring g env rs coloring
  then Some (alloc_of_coloring coloring env rs)
  else None

