open AST
open BinNums
open Coqlib
open Ctypes
open Datatypes
open Floats
open Integers
open Memory
open Values

type unary_operation =
| Onotbool
| Onotint
| Oneg

type binary_operation =
| Oadd
| Osub
| Omul
| Odiv
| Omod
| Oand
| Oor
| Oxor
| Oshl
| Oshr
| Oeq
| One
| Olt
| Ogt
| Ole
| Oge

type incr_or_decr =
| Incr
| Decr

type classify_cast_cases =
| Coq_cast_case_neutral
| Coq_cast_case_i2i of intsize * signedness
| Coq_cast_case_f2f of floatsize
| Coq_cast_case_i2f of signedness * floatsize
| Coq_cast_case_f2i of intsize * signedness
| Coq_cast_case_f2bool
| Coq_cast_case_p2bool
| Coq_cast_case_struct of ident * fieldlist * ident * fieldlist
| Coq_cast_case_union of ident * fieldlist * ident * fieldlist
| Coq_cast_case_void
| Coq_cast_case_default

let classify_cast :  coq_type -> coq_type -> classify_cast_cases =
  fun tfrom  ->
    function
| Tvoid -> Coq_cast_case_void
| Tint (sz2, si2, a) ->
  (match sz2 with
   | I32 ->
     (match tfrom with
      | Tint (sz1, si1, a0) -> Coq_cast_case_neutral
      | Tfloat (sz1, a0) -> Coq_cast_case_f2i (sz2, si2)
      | Tpointer (t, a0) -> Coq_cast_case_neutral
      | Tarray (t, z, a0) -> Coq_cast_case_neutral
      | Tfunction (t, t0) -> Coq_cast_case_neutral
      | _ -> Coq_cast_case_default)
   | IBool ->
     (match tfrom with
      | Tint (sz1, si1, a0) -> Coq_cast_case_i2i (sz2, si2)
      | Tfloat (sz1, a0) -> Coq_cast_case_f2bool
      | Tpointer (t, a0) -> Coq_cast_case_p2bool
      | Tarray (t, z, a0) -> Coq_cast_case_p2bool
      | Tfunction (t, t0) -> Coq_cast_case_p2bool
      | _ -> Coq_cast_case_default)
   | _ ->
     (match tfrom with
      | Tint (sz1, si1, a0) -> Coq_cast_case_i2i (sz2, si2)
      | Tfloat (sz1, a0) -> Coq_cast_case_f2i (sz2, si2)
      | _ -> Coq_cast_case_default))
| Tfloat (sz2, a) ->
  (match tfrom with
   | Tint (sz1, si1, a0) -> Coq_cast_case_i2f (si1, sz2)
   | Tfloat (sz1, a0) -> Coq_cast_case_f2f sz2
   | _ -> Coq_cast_case_default)
| Tpointer (t, a) ->
  (match tfrom with
   | Tint (i, s, a0) -> Coq_cast_case_neutral
   | Tpointer (t0, a0) -> Coq_cast_case_neutral
   | Tarray (t0, z, a0) -> Coq_cast_case_neutral
   | Tfunction (t0, t1) -> Coq_cast_case_neutral
   | _ -> Coq_cast_case_default)
| Tstruct (id2, fld2, a) ->
  (match tfrom with
   | Tstruct (id1, fld1, a0) -> Coq_cast_case_struct (id1, fld1, id2, fld2)
   | _ -> Coq_cast_case_default)
| Tunion (id2, fld2, a) ->
  (match tfrom with
   | Tunion (id1, fld1, a0) -> Coq_cast_case_union (id1, fld1, id2, fld2)
   | _ -> Coq_cast_case_default)
| _ -> Coq_cast_case_default

let cast_int_int :  intsize -> signedness -> Int.int -> Int.int =
  fun sz sg i  ->
  match sz with
  | I8 ->
    (match sg with
     | Signed -> Int.sign_ext (Zpos (Coq_xO (Coq_xO (Coq_xO Coq_xH)))) i
     | Unsigned -> Int.zero_ext (Zpos (Coq_xO (Coq_xO (Coq_xO Coq_xH)))) i)
  | I16 ->
    (match sg with
     | Signed ->
       Int.sign_ext (Zpos (Coq_xO (Coq_xO (Coq_xO (Coq_xO Coq_xH))))) i
     | Unsigned ->
       Int.zero_ext (Zpos (Coq_xO (Coq_xO (Coq_xO (Coq_xO Coq_xH))))) i)
  | I32 -> i
  | IBool -> if Int.eq i Int.zero then Int.zero else Int.one

let cast_int_float :  signedness -> Int.int -> float =
  fun si i  ->
  match si with
  | Signed -> Float.floatofint i
  | Unsigned -> Float.floatofintu i

let cast_float_int :  signedness -> float -> Int.int option =
  fun si f  ->
  match si with
  | Signed -> Float.intoffloat f
  | Unsigned -> Float.intuoffloat f

let cast_float_float :  floatsize -> float -> float =
  fun sz f  ->
  match sz with
  | F32 -> Float.singleoffloat f
  | F64 -> f

let sem_cast :  coq_val -> coq_type -> coq_type -> coq_val option =
  fun v t1 t2  ->
  match classify_cast t1 t2 with
  | Coq_cast_case_neutral ->
    (match v with
     | Vundef -> None
     | Vfloat f -> None
     | _ -> Some v)
  | Coq_cast_case_i2i (sz2, si2) ->
    (match v with
     | Vint i -> Some (Vint (cast_int_int sz2 si2 i))
     | _ -> None)
  | Coq_cast_case_f2f sz2 ->
    (match v with
     | Vfloat f -> Some (Vfloat (cast_float_float sz2 f))
     | _ -> None)
  | Coq_cast_case_i2f (si1, sz2) ->
    (match v with
     | Vint i -> Some (Vfloat (cast_float_float sz2 (cast_int_float si1 i)))
     | _ -> None)
  | Coq_cast_case_f2i (sz2, si2) ->
    (match v with
     | Vfloat f ->
       (match cast_float_int si2 f with
        | Some i -> Some (Vint (cast_int_int sz2 si2 i))
        | None -> None)
     | _ -> None)
  | Coq_cast_case_f2bool ->
    (match v with
     | Vfloat f ->
       Some (Vint (if Float.cmp Ceq f Float.zero then Int.zero else Int.one))
     | _ -> None)
  | Coq_cast_case_p2bool ->
    (match v with
     | Vint i -> Some (Vint (cast_int_int IBool Signed i))
     | Vptr (b, i) -> Some (Vint Int.one)
     | _ -> None)
  | Coq_cast_case_struct (id1, fld1, id2, fld2) ->
    if (&&) (proj_sumbool (ident_eq id1 id2))
         (proj_sumbool (fieldlist_eq fld1 fld2))
    then Some v
    else None
  | Coq_cast_case_union (id1, fld1, id2, fld2) ->
    if (&&) (proj_sumbool (ident_eq id1 id2))
         (proj_sumbool (fieldlist_eq fld1 fld2))
    then Some v
    else None
  | Coq_cast_case_void -> Some v
  | Coq_cast_case_default -> None

type classify_bool_cases =
| Coq_bool_case_i
| Coq_bool_case_f
| Coq_bool_case_p
| Coq_bool_default

let classify_bool :  coq_type -> classify_bool_cases =
  fun ty  ->
  match typeconv ty with
  | Tint (i, s, a) -> Coq_bool_case_i
  | Tfloat (f, a) -> Coq_bool_case_f
  | Tpointer (t, a) -> Coq_bool_case_p
  | _ -> Coq_bool_default

let bool_val :  coq_val -> coq_type -> bool option =
  fun v t  ->
  match classify_bool t with
  | Coq_bool_case_i ->
    (match v with
     | Vint n -> Some (negb (Int.eq n Int.zero))
     | _ -> None)
  | Coq_bool_case_f ->
    (match v with
     | Vfloat f -> Some (negb (Float.cmp Ceq f Float.zero))
     | _ -> None)
  | Coq_bool_case_p ->
    (match v with
     | Vint n -> Some (negb (Int.eq n Int.zero))
     | Vptr (b, ofs) -> Some true
     | _ -> None)
  | Coq_bool_default -> None

let sem_notbool :  coq_val -> coq_type -> coq_val option =
  fun v ty  ->
  match classify_bool ty with
  | Coq_bool_case_i ->
    (match v with
     | Vint n -> Some (Val.of_bool (Int.eq n Int.zero))
     | _ -> None)
  | Coq_bool_case_f ->
    (match v with
     | Vfloat f -> Some (Val.of_bool (Float.cmp Ceq f Float.zero))
     | _ -> None)
  | Coq_bool_case_p ->
    (match v with
     | Vint n -> Some (Val.of_bool (Int.eq n Int.zero))
     | Vptr (b, i) -> Some coq_Vfalse
     | _ -> None)
  | Coq_bool_default -> None

type classify_neg_cases =
| Coq_neg_case_i of signedness
| Coq_neg_case_f
| Coq_neg_default

let classify_neg :  coq_type -> classify_neg_cases =
    function
| Tint (i, s, a) ->
  (match i with
   | I32 -> Coq_neg_case_i s
   | _ -> Coq_neg_case_i Signed)
| Tfloat (f, a) -> Coq_neg_case_f
| _ -> Coq_neg_default

let sem_neg :  coq_val -> coq_type -> coq_val option =
  fun v ty  ->
  match classify_neg ty with
  | Coq_neg_case_i sg ->
    (match v with
     | Vint n -> Some (Vint (Int.neg n))
     | _ -> None)
  | Coq_neg_case_f ->
    (match v with
     | Vfloat f -> Some (Vfloat (Float.neg f))
     | _ -> None)
  | Coq_neg_default -> None

type classify_notint_cases =
| Coq_notint_case_i of signedness
| Coq_notint_default

let classify_notint :  coq_type -> classify_notint_cases =
    function
| Tint (i, s, a) ->
  (match i with
   | I32 -> Coq_notint_case_i s
   | _ -> Coq_notint_case_i Signed)
| _ -> Coq_notint_default

let sem_notint :  coq_val -> coq_type -> coq_val option =
  fun v ty  ->
  match classify_notint ty with
  | Coq_notint_case_i sg ->
    (match v with
     | Vint n -> Some (Vint (Int.xor n Int.mone))
     | _ -> None)
  | Coq_notint_default -> None

type classify_add_cases =
| Coq_add_case_ii of signedness
| Coq_add_case_ff
| Coq_add_case_if of signedness
| Coq_add_case_fi of signedness
| Coq_add_case_pi of coq_type * attr
| Coq_add_case_ip of coq_type * attr
| Coq_add_default

let classify_add :  coq_type -> coq_type -> classify_add_cases =
  fun ty1 ty2  ->
  match typeconv ty1 with
  | Tint (i, sg, a0) ->
    (match i with
     | I32 ->
       (match sg with
        | Signed ->
          (match typeconv ty2 with
           | Tint (i0, s, a) ->
             (match i0 with
              | I32 -> Coq_add_case_ii s
              | _ -> Coq_add_case_ii Signed)
           | Tfloat (f, a) -> Coq_add_case_if sg
           | Tpointer (ty, a) -> Coq_add_case_ip (ty, a)
           | _ -> Coq_add_default)
        | Unsigned ->
          (match typeconv ty2 with
           | Tint (i0, s, a) -> Coq_add_case_ii Unsigned
           | Tfloat (f, a) -> Coq_add_case_if sg
           | Tpointer (ty, a) -> Coq_add_case_ip (ty, a)
           | _ -> Coq_add_default))
     | _ ->
       (match typeconv ty2 with
        | Tint (i0, s, a) ->
          (match i0 with
           | I32 -> Coq_add_case_ii s
           | _ -> Coq_add_case_ii Signed)
        | Tfloat (f, a) -> Coq_add_case_if sg
        | Tpointer (ty, a) -> Coq_add_case_ip (ty, a)
        | _ -> Coq_add_default))
  | Tfloat (f, a) ->
    (match typeconv ty2 with
     | Tint (i, sg, a0) -> Coq_add_case_fi sg
     | Tfloat (f0, a0) -> Coq_add_case_ff
     | _ -> Coq_add_default)
  | Tpointer (ty, a) ->
    (match typeconv ty2 with
     | Tint (i, s, a0) -> Coq_add_case_pi (ty, a)
     | _ -> Coq_add_default)
  | _ -> Coq_add_default

let sem_add :     coq_val -> coq_type -> coq_val -> coq_type -> coq_val option =
  fun v1 t1 v2 t2  ->
  match classify_add t1 t2 with
  | Coq_add_case_ii sg ->
    (match v1 with
     | Vint n1 ->
       (match v2 with
        | Vint n2 -> Some (Vint (Int.add n1 n2))
        | _ -> None)
     | _ -> None)
  | Coq_add_case_ff ->
    (match v1 with
     | Vfloat n1 ->
       (match v2 with
        | Vfloat n2 -> Some (Vfloat (Float.add n1 n2))
        | _ -> None)
     | _ -> None)
  | Coq_add_case_if sg ->
    (match v1 with
     | Vint n1 ->
       (match v2 with
        | Vfloat n2 -> Some (Vfloat (Float.add (cast_int_float sg n1) n2))
        | _ -> None)
     | _ -> None)
  | Coq_add_case_fi sg ->
    (match v1 with
     | Vfloat n1 ->
       (match v2 with
        | Vint n2 -> Some (Vfloat (Float.add n1 (cast_int_float sg n2)))
        | _ -> None)
     | _ -> None)
  | Coq_add_case_pi (ty, a) ->
    (match v1 with
     | Vptr (b1, ofs1) ->
       (match v2 with
        | Vint n2 ->
          Some (Vptr (b1,
            (Int.add ofs1 (Int.mul (Int.repr (sizeof ty)) n2))))
        | _ -> None)
     | _ -> None)
  | Coq_add_case_ip (ty, a) ->
    (match v1 with
     | Vint n1 ->
       (match v2 with
        | Vptr (b2, ofs2) ->
          Some (Vptr (b2,
            (Int.add ofs2 (Int.mul (Int.repr (sizeof ty)) n1))))
        | _ -> None)
     | _ -> None)
  | Coq_add_default -> None

type classify_sub_cases =
| Coq_sub_case_ii of signedness
| Coq_sub_case_ff
| Coq_sub_case_if of signedness
| Coq_sub_case_fi of signedness
| Coq_sub_case_pi of coq_type
| Coq_sub_case_pp of coq_type
| Coq_sub_default

let classify_sub :  coq_type -> coq_type -> classify_sub_cases =
  fun ty1 ty2  ->
  match typeconv ty1 with
  | Tint (i, sg, a) ->
    (match i with
     | I32 ->
       (match sg with
        | Signed ->
          (match typeconv ty2 with
           | Tint (i0, s, a0) ->
             (match i0 with
              | I32 -> Coq_sub_case_ii s
              | _ -> Coq_sub_case_ii Signed)
           | Tfloat (f, a0) -> Coq_sub_case_if sg
           | _ -> Coq_sub_default)
        | Unsigned ->
          (match typeconv ty2 with
           | Tint (i0, s, a0) -> Coq_sub_case_ii Unsigned
           | Tfloat (f, a0) -> Coq_sub_case_if sg
           | _ -> Coq_sub_default))
     | _ ->
       (match typeconv ty2 with
        | Tint (i0, s, a0) ->
          (match i0 with
           | I32 -> Coq_sub_case_ii s
           | _ -> Coq_sub_case_ii Signed)
        | Tfloat (f, a0) -> Coq_sub_case_if sg
        | _ -> Coq_sub_default))
  | Tfloat (f, a) ->
    (match typeconv ty2 with
     | Tint (i, sg, a0) -> Coq_sub_case_fi sg
     | Tfloat (f0, a0) -> Coq_sub_case_ff
     | _ -> Coq_sub_default)
  | Tpointer (ty, a) ->
    (match typeconv ty2 with
     | Tint (i, s, a0) -> Coq_sub_case_pi ty
     | Tpointer (t, a0) -> Coq_sub_case_pp ty
     | _ -> Coq_sub_default)
  | _ -> Coq_sub_default

let sem_sub :     coq_val -> coq_type -> coq_val -> coq_type -> coq_val option =
  fun v1 t1 v2 t2  ->
  match classify_sub t1 t2 with
  | Coq_sub_case_ii sg ->
    (match v1 with
     | Vint n1 ->
       (match v2 with
        | Vint n2 -> Some (Vint (Int.sub n1 n2))
        | _ -> None)
     | _ -> None)
  | Coq_sub_case_ff ->
    (match v1 with
     | Vfloat f1 ->
       (match v2 with
        | Vfloat f2 -> Some (Vfloat (Float.sub f1 f2))
        | _ -> None)
     | _ -> None)
  | Coq_sub_case_if sg ->
    (match v1 with
     | Vint n1 ->
       (match v2 with
        | Vfloat n2 -> Some (Vfloat (Float.sub (cast_int_float sg n1) n2))
        | _ -> None)
     | _ -> None)
  | Coq_sub_case_fi sg ->
    (match v1 with
     | Vfloat n1 ->
       (match v2 with
        | Vint n2 -> Some (Vfloat (Float.sub n1 (cast_int_float sg n2)))
        | _ -> None)
     | _ -> None)
  | Coq_sub_case_pi ty ->
    (match v1 with
     | Vptr (b1, ofs1) ->
       (match v2 with
        | Vint n2 ->
          Some (Vptr (b1,
            (Int.sub ofs1 (Int.mul (Int.repr (sizeof ty)) n2))))
        | _ -> None)
     | _ -> None)
  | Coq_sub_case_pp ty ->
    (match v1 with
     | Vptr (b1, ofs1) ->
       (match v2 with
        | Vptr (b2, ofs2) ->
          if zeq b1 b2
          then if Int.eq (Int.repr (sizeof ty)) Int.zero
               then None
               else Some (Vint
                      (Int.divu (Int.sub ofs1 ofs2) (Int.repr (sizeof ty))))
          else None
        | _ -> None)
     | _ -> None)
  | Coq_sub_default -> None

type classify_mul_cases =
| Coq_mul_case_ii of signedness
| Coq_mul_case_ff
| Coq_mul_case_if of signedness
| Coq_mul_case_fi of signedness
| Coq_mul_default

let classify_mul :  coq_type -> coq_type -> classify_mul_cases =
  fun ty1 ty2  ->
  match typeconv ty1 with
  | Tint (i, sg, a) ->
    (match i with
     | I32 ->
       (match sg with
        | Signed ->
          (match typeconv ty2 with
           | Tint (i0, s, a0) ->
             (match i0 with
              | I32 -> Coq_mul_case_ii s
              | _ -> Coq_mul_case_ii Signed)
           | Tfloat (f, a0) -> Coq_mul_case_if sg
           | _ -> Coq_mul_default)
        | Unsigned ->
          (match typeconv ty2 with
           | Tint (i0, s, a0) -> Coq_mul_case_ii Unsigned
           | Tfloat (f, a0) -> Coq_mul_case_if sg
           | _ -> Coq_mul_default))
     | _ ->
       (match typeconv ty2 with
        | Tint (i0, s, a0) ->
          (match i0 with
           | I32 -> Coq_mul_case_ii s
           | _ -> Coq_mul_case_ii Signed)
        | Tfloat (f, a0) -> Coq_mul_case_if sg
        | _ -> Coq_mul_default))
  | Tfloat (f, a) ->
    (match typeconv ty2 with
     | Tint (i, sg, a0) -> Coq_mul_case_fi sg
     | Tfloat (f0, a0) -> Coq_mul_case_ff
     | _ -> Coq_mul_default)
  | _ -> Coq_mul_default

let sem_mul :     coq_val -> coq_type -> coq_val -> coq_type -> coq_val option =
  fun v1 t1 v2 t2  ->
  match classify_mul t1 t2 with
  | Coq_mul_case_ii sg ->
    (match v1 with
     | Vint n1 ->
       (match v2 with
        | Vint n2 -> Some (Vint (Int.mul n1 n2))
        | _ -> None)
     | _ -> None)
  | Coq_mul_case_ff ->
    (match v1 with
     | Vfloat f1 ->
       (match v2 with
        | Vfloat f2 -> Some (Vfloat (Float.mul f1 f2))
        | _ -> None)
     | _ -> None)
  | Coq_mul_case_if sg ->
    (match v1 with
     | Vint n1 ->
       (match v2 with
        | Vfloat n2 -> Some (Vfloat (Float.mul (cast_int_float sg n1) n2))
        | _ -> None)
     | _ -> None)
  | Coq_mul_case_fi sg ->
    (match v1 with
     | Vfloat n1 ->
       (match v2 with
        | Vint n2 -> Some (Vfloat (Float.mul n1 (cast_int_float sg n2)))
        | _ -> None)
     | _ -> None)
  | Coq_mul_default -> None

type classify_div_cases =
| Coq_div_case_ii of signedness
| Coq_div_case_ff
| Coq_div_case_if of signedness
| Coq_div_case_fi of signedness
| Coq_div_default

let classify_div :  coq_type -> coq_type -> classify_div_cases =
  fun ty1 ty2  ->
  match typeconv ty1 with
  | Tint (i, sg, a) ->
    (match i with
     | I32 ->
       (match sg with
        | Signed ->
          (match typeconv ty2 with
           | Tint (i0, s, a0) ->
             (match i0 with
              | I32 -> Coq_div_case_ii s
              | _ -> Coq_div_case_ii Signed)
           | Tfloat (f, a0) -> Coq_div_case_if sg
           | _ -> Coq_div_default)
        | Unsigned ->
          (match typeconv ty2 with
           | Tint (i0, s, a0) -> Coq_div_case_ii Unsigned
           | Tfloat (f, a0) -> Coq_div_case_if sg
           | _ -> Coq_div_default))
     | _ ->
       (match typeconv ty2 with
        | Tint (i0, s, a0) ->
          (match i0 with
           | I32 -> Coq_div_case_ii s
           | _ -> Coq_div_case_ii Signed)
        | Tfloat (f, a0) -> Coq_div_case_if sg
        | _ -> Coq_div_default))
  | Tfloat (f, a) ->
    (match typeconv ty2 with
     | Tint (i, sg, a0) -> Coq_div_case_fi sg
     | Tfloat (f0, a0) -> Coq_div_case_ff
     | _ -> Coq_div_default)
  | _ -> Coq_div_default

let sem_div :     coq_val -> coq_type -> coq_val -> coq_type -> coq_val option =
  fun v1 t1 v2 t2  ->
  match classify_div t1 t2 with
  | Coq_div_case_ii s ->
    (match s with
     | Signed ->
       (match v1 with
        | Vint n1 ->
          (match v2 with
           | Vint n2 ->
             if (||) (Int.eq n2 Int.zero)
                  ((&&) (Int.eq n1 (Int.repr Int.min_signed))
                    (Int.eq n2 Int.mone))
             then None
             else Some (Vint (Int.divs n1 n2))
           | _ -> None)
        | _ -> None)
     | Unsigned ->
       (match v1 with
        | Vint n1 ->
          (match v2 with
           | Vint n2 ->
             if Int.eq n2 Int.zero
             then None
             else Some (Vint (Int.divu n1 n2))
           | _ -> None)
        | _ -> None))
  | Coq_div_case_ff ->
    (match v1 with
     | Vfloat f1 ->
       (match v2 with
        | Vfloat f2 -> Some (Vfloat (Float.div f1 f2))
        | _ -> None)
     | _ -> None)
  | Coq_div_case_if sg ->
    (match v1 with
     | Vint n1 ->
       (match v2 with
        | Vfloat n2 -> Some (Vfloat (Float.div (cast_int_float sg n1) n2))
        | _ -> None)
     | _ -> None)
  | Coq_div_case_fi sg ->
    (match v1 with
     | Vfloat n1 ->
       (match v2 with
        | Vint n2 -> Some (Vfloat (Float.div n1 (cast_int_float sg n2)))
        | _ -> None)
     | _ -> None)
  | Coq_div_default -> None

type classify_binint_cases =
| Coq_binint_case_ii of signedness
| Coq_binint_default

let classify_binint :  coq_type -> coq_type -> classify_binint_cases =
  fun ty1 ty2  ->
  match typeconv ty1 with
  | Tint (i, s, a) ->
    (match i with
     | I32 ->
       (match s with
        | Signed ->
          (match typeconv ty2 with
           | Tint (i0, s0, a0) ->
             (match i0 with
              | I32 -> Coq_binint_case_ii s0
              | _ -> Coq_binint_case_ii Signed)
           | _ -> Coq_binint_default)
        | Unsigned ->
          (match typeconv ty2 with
           | Tint (i0, s0, a0) -> Coq_binint_case_ii Unsigned
           | _ -> Coq_binint_default))
     | _ ->
       (match typeconv ty2 with
        | Tint (i0, s0, a0) ->
          (match i0 with
           | I32 -> Coq_binint_case_ii s0
           | _ -> Coq_binint_case_ii Signed)
        | _ -> Coq_binint_default))
  | _ -> Coq_binint_default

let sem_mod :     coq_val -> coq_type -> coq_val -> coq_type -> coq_val option =
  fun v1 t1 v2 t2  ->
  match classify_binint t1 t2 with
  | Coq_binint_case_ii s ->
    (match s with
     | Signed ->
       (match v1 with
        | Vint n1 ->
          (match v2 with
           | Vint n2 ->
             if (||) (Int.eq n2 Int.zero)
                  ((&&) (Int.eq n1 (Int.repr Int.min_signed))
                    (Int.eq n2 Int.mone))
             then None
             else Some (Vint (Int.mods n1 n2))
           | _ -> None)
        | _ -> None)
     | Unsigned ->
       (match v1 with
        | Vint n1 ->
          (match v2 with
           | Vint n2 ->
             if Int.eq n2 Int.zero
             then None
             else Some (Vint (Int.modu n1 n2))
           | _ -> None)
        | _ -> None))
  | Coq_binint_default -> None

let sem_and :     coq_val -> coq_type -> coq_val -> coq_type -> coq_val option =
  fun v1 t1 v2 t2  ->
  match classify_binint t1 t2 with
  | Coq_binint_case_ii sg ->
    (match v1 with
     | Vint n1 ->
       (match v2 with
        | Vint n2 -> Some (Vint (Int.coq_and n1 n2))
        | _ -> None)
     | _ -> None)
  | Coq_binint_default -> None

let sem_or :     coq_val -> coq_type -> coq_val -> coq_type -> coq_val option =
  fun v1 t1 v2 t2  ->
  match classify_binint t1 t2 with
  | Coq_binint_case_ii sg ->
    (match v1 with
     | Vint n1 ->
       (match v2 with
        | Vint n2 -> Some (Vint (Int.coq_or n1 n2))
        | _ -> None)
     | _ -> None)
  | Coq_binint_default -> None

let sem_xor :     coq_val -> coq_type -> coq_val -> coq_type -> coq_val option =
  fun v1 t1 v2 t2  ->
  match classify_binint t1 t2 with
  | Coq_binint_case_ii sg ->
    (match v1 with
     | Vint n1 ->
       (match v2 with
        | Vint n2 -> Some (Vint (Int.xor n1 n2))
        | _ -> None)
     | _ -> None)
  | Coq_binint_default -> None

type classify_shift_cases =
| Coq_shift_case_ii of signedness
| Coq_shift_default

let classify_shift :  coq_type -> coq_type -> classify_shift_cases =
  fun ty1 ty2  ->
  match typeconv ty1 with
  | Tint (i, s, a) ->
    (match i with
     | I32 ->
       (match typeconv ty2 with
        | Tint (i0, s0, a0) -> Coq_shift_case_ii s
        | _ -> Coq_shift_default)
     | _ ->
       (match typeconv ty2 with
        | Tint (i0, s0, a0) -> Coq_shift_case_ii Signed
        | _ -> Coq_shift_default))
  | _ -> Coq_shift_default

let sem_shl :     coq_val -> coq_type -> coq_val -> coq_type -> coq_val option =
  fun v1 t1 v2 t2  ->
  match classify_shift t1 t2 with
  | Coq_shift_case_ii sg ->
    (match v1 with
     | Vint n1 ->
       (match v2 with
        | Vint n2 ->
          if Int.ltu n2 Int.iwordsize
          then Some (Vint (Int.shl n1 n2))
          else None
        | _ -> None)
     | _ -> None)
  | Coq_shift_default -> None

let sem_shr :     coq_val -> coq_type -> coq_val -> coq_type -> coq_val option =
  fun v1 t1 v2 t2  ->
  match classify_shift t1 t2 with
  | Coq_shift_case_ii s ->
    (match s with
     | Signed ->
       (match v1 with
        | Vint n1 ->
          (match v2 with
           | Vint n2 ->
             if Int.ltu n2 Int.iwordsize
             then Some (Vint (Int.shr n1 n2))
             else None
           | _ -> None)
        | _ -> None)
     | Unsigned ->
       (match v1 with
        | Vint n1 ->
          (match v2 with
           | Vint n2 ->
             if Int.ltu n2 Int.iwordsize
             then Some (Vint (Int.shru n1 n2))
             else None
           | _ -> None)
        | _ -> None))
  | Coq_shift_default -> None

type classify_cmp_cases =
| Coq_cmp_case_ii of signedness
| Coq_cmp_case_pp
| Coq_cmp_case_ff
| Coq_cmp_case_if of signedness
| Coq_cmp_case_fi of signedness
| Coq_cmp_default

let classify_cmp :  coq_type -> coq_type -> classify_cmp_cases =
  fun ty1 ty2  ->
  match typeconv ty1 with
  | Tint (i, sg, a) ->
    (match i with
     | I32 ->
       (match sg with
        | Signed ->
          (match typeconv ty2 with
           | Tint (i0, s, a0) ->
             (match i0 with
              | I32 -> Coq_cmp_case_ii s
              | _ -> Coq_cmp_case_ii Signed)
           | Tfloat (f, a0) -> Coq_cmp_case_if sg
           | Tpointer (t, a0) -> Coq_cmp_case_pp
           | _ -> Coq_cmp_default)
        | Unsigned ->
          (match typeconv ty2 with
           | Tint (i0, s, a0) -> Coq_cmp_case_ii Unsigned
           | Tfloat (f, a0) -> Coq_cmp_case_if sg
           | Tpointer (t, a0) -> Coq_cmp_case_pp
           | _ -> Coq_cmp_default))
     | _ ->
       (match typeconv ty2 with
        | Tint (i0, s, a0) ->
          (match i0 with
           | I32 -> Coq_cmp_case_ii s
           | _ -> Coq_cmp_case_ii Signed)
        | Tfloat (f, a0) -> Coq_cmp_case_if sg
        | Tpointer (t, a0) -> Coq_cmp_case_pp
        | _ -> Coq_cmp_default))
  | Tfloat (f, a) ->
    (match typeconv ty2 with
     | Tint (i, sg, a0) -> Coq_cmp_case_fi sg
     | Tfloat (f0, a0) -> Coq_cmp_case_ff
     | _ -> Coq_cmp_default)
  | Tpointer (t, a) ->
    (match typeconv ty2 with
     | Tint (i, s, a0) -> Coq_cmp_case_pp
     | Tpointer (t0, a0) -> Coq_cmp_case_pp
     | _ -> Coq_cmp_default)
  | _ -> Coq_cmp_default

let sem_cmp :     comparison -> coq_val -> coq_type -> coq_val -> coq_type -> Mem.mem ->
    coq_val option =
  fun c v1 t1 v2 t2 m  ->
  match classify_cmp t1 t2 with
  | Coq_cmp_case_ii s ->
    (match s with
     | Signed ->
       (match v1 with
        | Vint n1 ->
          (match v2 with
           | Vint n2 -> Some (Val.of_bool (Int.cmp c n1 n2))
           | _ -> None)
        | _ -> None)
     | Unsigned ->
       (match v1 with
        | Vint n1 ->
          (match v2 with
           | Vint n2 -> Some (Val.of_bool (Int.cmpu c n1 n2))
           | _ -> None)
        | _ -> None))
  | Coq_cmp_case_pp ->
    (match v1 with
     | Vint n ->
       (match v2 with
        | Vint n2 -> Some (Val.of_bool (Int.cmpu c n n2))
        | Vptr (b, ofs) ->
          if Int.eq n Int.zero
          then option_map Val.of_bool (Val.cmp_different_blocks c)
          else None
        | _ -> None)
     | Vptr (b, ofs) ->
       (match v2 with
        | Vint n ->
          if Int.eq n Int.zero
          then option_map Val.of_bool (Val.cmp_different_blocks c)
          else None
        | Vptr (b2, ofs2) ->
          if zeq b b2
          then if (&&) (Mem.weak_valid_pointer m b (Int.unsigned ofs))
                    (Mem.weak_valid_pointer m b2 (Int.unsigned ofs2))
               then Some (Val.of_bool (Int.cmpu c ofs ofs2))
               else None
          else if (&&) (Mem.valid_pointer m b (Int.unsigned ofs))
                    (Mem.valid_pointer m b2 (Int.unsigned ofs2))
               then option_map Val.of_bool (Val.cmp_different_blocks c)
               else None
        | _ -> None)
     | _ -> None)
  | Coq_cmp_case_ff ->
    (match v1 with
     | Vfloat f1 ->
       (match v2 with
        | Vfloat f2 -> Some (Val.of_bool (Float.cmp c f1 f2))
        | _ -> None)
     | _ -> None)
  | Coq_cmp_case_if sg ->
    (match v1 with
     | Vint n1 ->
       (match v2 with
        | Vfloat n2 ->
          Some (Val.of_bool (Float.cmp c (cast_int_float sg n1) n2))
        | _ -> None)
     | _ -> None)
  | Coq_cmp_case_fi sg ->
    (match v1 with
     | Vfloat n1 ->
       (match v2 with
        | Vint n2 ->
          Some (Val.of_bool (Float.cmp c n1 (cast_int_float sg n2)))
        | _ -> None)
     | _ -> None)
  | Coq_cmp_default -> None

type classify_fun_cases =
| Coq_fun_case_f of typelist * coq_type
| Coq_fun_default

let classify_fun :  coq_type -> classify_fun_cases =
    function
| Tpointer (t, a) ->
  (match t with
   | Tfunction (args, res) -> Coq_fun_case_f (args, res)
   | _ -> Coq_fun_default)
| Tfunction (args, res) -> Coq_fun_case_f (args, res)
| _ -> Coq_fun_default

let sem_unary_operation :     unary_operation -> coq_val -> coq_type -> coq_val option =
  fun op v ty  ->
  match op with
  | Onotbool -> sem_notbool v ty
  | Onotint -> sem_notint v ty
  | Oneg -> sem_neg v ty

let sem_binary_operation :     binary_operation -> coq_val -> coq_type -> coq_val -> coq_type -> Mem.mem
    -> coq_val option =
  fun op v1 t1 v2 t2 m  ->
  match op with
  | Oadd -> sem_add v1 t1 v2 t2
  | Osub -> sem_sub v1 t1 v2 t2
  | Omul -> sem_mul v1 t1 v2 t2
  | Odiv -> sem_div v1 t1 v2 t2
  | Omod -> sem_mod v1 t1 v2 t2
  | Oand -> sem_and v1 t1 v2 t2
  | Oor -> sem_or v1 t1 v2 t2
  | Oxor -> sem_xor v1 t1 v2 t2
  | Oshl -> sem_shl v1 t1 v2 t2
  | Oshr -> sem_shr v1 t1 v2 t2
  | Oeq -> sem_cmp Ceq v1 t1 v2 t2 m
  | One -> sem_cmp Cne v1 t1 v2 t2 m
  | Olt -> sem_cmp Clt v1 t1 v2 t2 m
  | Ogt -> sem_cmp Cgt v1 t1 v2 t2 m
  | Ole -> sem_cmp Cle v1 t1 v2 t2 m
  | Oge -> sem_cmp Cge v1 t1 v2 t2 m

