open AST
open BinNums
open Csyntax
open Ctypes
open Globalenvs
open Integers
open List0
open Maps
open Memory
open Values

type genv = (fundef, coq_type) Genv.t

type env = (block * coq_type) PTree.t

let empty_env :  env =
  PTree.empty

let block_of_binding :     (ident * (block * coq_type)) -> (block * coq_Z) * coq_Z =
    function
| (id, p) -> let (b, ty) = p in ((b, Z0), (sizeof ty))

let blocks_of_env :  env -> ((block * coq_Z) * coq_Z) list =
  fun e  ->
  map block_of_binding (PTree.elements e)

let rec select_switch :     Int.int -> labeled_statements -> labeled_statements =
  fun n sl  ->
    match sl with
| LSdefault s -> sl
| LScase (c, s, sl') -> if Int.eq c n then sl else select_switch n sl'

let rec seq_of_labeled_statement :  labeled_statements -> statement =
    function
| LSdefault s -> s
| LScase (c, s, sl') -> Ssequence (s, (seq_of_labeled_statement sl'))

let type_of_global :  genv -> block -> coq_type option =
  fun ge b  ->
  match Genv.find_var_info ge b with
  | Some gv -> Some gv.gvar_info
  | None ->
    (match Genv.find_funct_ptr ge b with
     | Some fd -> Some (type_of_fundef fd)
     | None -> None)

type kind =
| LV
| RV

type cont =
| Kstop
| Kdo of cont
| Kseq of statement * cont
| Kifthenelse of statement * statement * cont
| Kwhile1 of expr * statement * cont
| Kwhile2 of expr * statement * cont
| Kdowhile1 of expr * statement * cont
| Kdowhile2 of expr * statement * cont
| Kfor2 of expr * statement * statement * cont
| Kfor3 of expr * statement * statement * cont
| Kfor4 of expr * statement * statement * cont
| Kswitch1 of labeled_statements * cont
| Kswitch2 of cont
| Kreturn of cont
| Kcall of coq_function * env * (expr -> expr) * coq_type * cont

let rec call_cont :  cont -> cont =
  fun k  ->
    match k with
| Kdo k0 -> k0
| Kseq (s, k0) -> call_cont k0
| Kifthenelse (s1, s2, k0) -> call_cont k0
| Kwhile1 (e, s, k0) -> call_cont k0
| Kwhile2 (e, s, k0) -> call_cont k0
| Kdowhile1 (e, s, k0) -> call_cont k0
| Kdowhile2 (e, s, k0) -> call_cont k0
| Kfor2 (e2, e3, s, k0) -> call_cont k0
| Kfor3 (e2, e3, s, k0) -> call_cont k0
| Kfor4 (e2, e3, s, k0) -> call_cont k0
| Kswitch1 (ls, k0) -> call_cont k0
| Kswitch2 k0 -> call_cont k0
| Kreturn k0 -> call_cont k0
| _ -> k

type state =
| State of coq_function * statement * cont * env * Mem.mem
| ExprState of coq_function * expr * cont * env * Mem.mem
| Callstate of fundef * coq_val list * cont * Mem.mem
| Returnstate of coq_val * cont * Mem.mem
| Stuckstate

let rec find_label :     label -> statement -> cont -> (statement * cont) option =
  fun lbl s k  ->
  match s with
  | Ssequence (s1, s2) ->
    (match find_label lbl s1 (Kseq (s2, k)) with
     | Some sk -> Some sk
     | None -> find_label lbl s2 k)
  | Sifthenelse (a, s1, s2) ->
    (match find_label lbl s1 k with
     | Some sk -> Some sk
     | None -> find_label lbl s2 k)
  | Swhile (a, s1) -> find_label lbl s1 (Kwhile2 (a, s1, k))
  | Sdowhile (a, s1) -> find_label lbl s1 (Kdowhile1 (a, s1, k))
  | Sfor (a1, a2, a3, s1) ->
    (match find_label lbl a1 (Kseq ((Sfor (Sskip, a2, a3, s1)), k)) with
     | Some sk -> Some sk
     | None ->
       (match find_label lbl s1 (Kfor3 (a2, a3, s1, k)) with
        | Some sk -> Some sk
        | None -> find_label lbl a3 (Kfor4 (a2, a3, s1, k))))
  | Sswitch (e, sl) -> find_label_ls lbl sl (Kswitch2 k)
  | Slabel (lbl', s') ->
    if ident_eq lbl lbl' then Some (s', k) else find_label lbl s' k
  | _ -> None

and find_label_ls :     label -> labeled_statements -> cont -> (statement * cont) option =
  fun lbl sl k  ->
  match sl with
  | LSdefault s -> find_label lbl s k
  | LScase (i, s, sl') ->
    (match find_label lbl s (Kseq ((seq_of_labeled_statement sl'), k)) with
     | Some sk -> Some sk
     | None -> find_label_ls lbl sl' k)

