open AST
open BinNums
open Clight
open Cminor
open Cop
open Coqlib
open Csharpminor
open Ctypes
open Datatypes
open Errors
open Floats
open Integers
open List0

let make_intconst :  Int.int -> expr =
  fun n  ->
  Econst (Ointconst n)

let make_floatconst :  float -> expr =
  fun f  ->
  Econst (Ofloatconst f)

let make_floatofint :  expr -> signedness -> expr =
  fun e  ->
    function
| Signed -> Eunop (Ofloatofint, e)
| Unsigned -> Eunop (Ofloatofintu, e)

let make_intoffloat :  expr -> signedness -> expr =
  fun e  ->
    function
| Signed -> Eunop (Ointoffloat, e)
| Unsigned -> Eunop (Ointuoffloat, e)

let make_cmp_ne_zero :  expr -> expr =
  fun e  ->
    match e with
| Ebinop (b, e1, e2) ->
  (match b with
   | Ocmp c -> e
   | Ocmpu c -> e
   | Ocmpf c -> e
   | _ -> Ebinop ((Ocmp Cne), e, (make_intconst Int.zero)))
| _ -> Ebinop ((Ocmp Cne), e, (make_intconst Int.zero))

let make_boolean :  expr -> coq_type -> expr =
  fun e ty  ->
  match classify_bool ty with
  | Coq_bool_case_i -> make_cmp_ne_zero e
  | Coq_bool_case_f -> Ebinop ((Ocmpf Cne), e, (make_floatconst Float.zero))
  | Coq_bool_case_p -> Ebinop ((Ocmpu Cne), e, (make_intconst Int.zero))
  | Coq_bool_default -> e

let make_notbool :  expr -> coq_type -> expr res =
  fun e ty  ->
  match classify_bool ty with
  | Coq_bool_case_i -> OK (Ebinop ((Ocmp Ceq), e, (make_intconst Int.zero)))
  | Coq_bool_case_f ->
    OK (Ebinop ((Ocmpf Ceq), e, (make_floatconst Float.zero)))
  | Coq_bool_case_p -> OK (Ebinop ((Ocmpu Ceq), e, (make_intconst Int.zero)))
  | Coq_bool_default ->
    Error
      (msg
        ('C'::('s'::('h'::('m'::('g'::('e'::('n'::('.'::('m'::('a'::('k'::('e'::('_'::('n'::('o'::('t'::('b'::('o'::('o'::('l'::[])))))))))))))))))))))

let make_neg :  expr -> coq_type -> expr res =
  fun e ty  ->
  match classify_neg ty with
  | Coq_neg_case_i s -> OK (Eunop (Onegint, e))
  | Coq_neg_case_f -> OK (Eunop (Onegf, e))
  | Coq_neg_default ->
    Error
      (msg
        ('C'::('s'::('h'::('m'::('g'::('e'::('n'::('.'::('m'::('a'::('k'::('e'::('_'::('n'::('e'::('g'::[])))))))))))))))))

let make_notint :  expr -> coq_type -> expr =
  fun e ty  ->
  Eunop (Cminor.Onotint, e)

let make_add :  expr -> coq_type -> expr -> coq_type -> expr res =
  fun e1 ty1 e2 ty2  ->
  match classify_add ty1 ty2 with
  | Coq_add_case_ii s -> OK (Ebinop (Cminor.Oadd, e1, e2))
  | Coq_add_case_ff -> OK (Ebinop (Oaddf, e1, e2))
  | Coq_add_case_if sg -> OK (Ebinop (Oaddf, (make_floatofint e1 sg), e2))
  | Coq_add_case_fi sg -> OK (Ebinop (Oaddf, e1, (make_floatofint e2 sg)))
  | Coq_add_case_pi (ty, a) ->
    let n = make_intconst (Int.repr (sizeof ty)) in
    OK (Ebinop (Cminor.Oadd, e1, (Ebinop (Cminor.Omul, n, e2))))
  | Coq_add_case_ip (ty, a) ->
    let n = make_intconst (Int.repr (sizeof ty)) in
    OK (Ebinop (Cminor.Oadd, e2, (Ebinop (Cminor.Omul, n, e1))))
  | Coq_add_default ->
    Error
      (msg
        ('C'::('s'::('h'::('m'::('g'::('e'::('n'::('.'::('m'::('a'::('k'::('e'::('_'::('a'::('d'::('d'::[])))))))))))))))))

let make_sub :  expr -> coq_type -> expr -> coq_type -> expr res =
  fun e1 ty1 e2 ty2  ->
  match classify_sub ty1 ty2 with
  | Coq_sub_case_ii s -> OK (Ebinop (Cminor.Osub, e1, e2))
  | Coq_sub_case_ff -> OK (Ebinop (Osubf, e1, e2))
  | Coq_sub_case_if sg -> OK (Ebinop (Osubf, (make_floatofint e1 sg), e2))
  | Coq_sub_case_fi sg -> OK (Ebinop (Osubf, e1, (make_floatofint e2 sg)))
  | Coq_sub_case_pi ty ->
    let n = make_intconst (Int.repr (sizeof ty)) in
    OK (Ebinop (Cminor.Osub, e1, (Ebinop (Cminor.Omul, n, e2))))
  | Coq_sub_case_pp ty ->
    let n = make_intconst (Int.repr (sizeof ty)) in
    OK (Ebinop (Odivu, (Ebinop (Cminor.Osub, e1, e2)), n))
  | Coq_sub_default ->
    Error
      (msg
        ('C'::('s'::('h'::('m'::('g'::('e'::('n'::('.'::('m'::('a'::('k'::('e'::('_'::('s'::('u'::('b'::[])))))))))))))))))

let make_mul :  expr -> coq_type -> expr -> coq_type -> expr res =
  fun e1 ty1 e2 ty2  ->
  match classify_mul ty1 ty2 with
  | Coq_mul_case_ii s -> OK (Ebinop (Cminor.Omul, e1, e2))
  | Coq_mul_case_ff -> OK (Ebinop (Omulf, e1, e2))
  | Coq_mul_case_if sg -> OK (Ebinop (Omulf, (make_floatofint e1 sg), e2))
  | Coq_mul_case_fi sg -> OK (Ebinop (Omulf, e1, (make_floatofint e2 sg)))
  | Coq_mul_default ->
    Error
      (msg
        ('C'::('s'::('h'::('m'::('g'::('e'::('n'::('.'::('m'::('a'::('k'::('e'::('_'::('m'::('u'::('l'::[])))))))))))))))))

let make_div :  expr -> coq_type -> expr -> coq_type -> expr res =
  fun e1 ty1 e2 ty2  ->
  match classify_div ty1 ty2 with
  | Coq_div_case_ii s ->
    (match s with
     | Signed -> OK (Ebinop (Cminor.Odiv, e1, e2))
     | Unsigned -> OK (Ebinop (Odivu, e1, e2)))
  | Coq_div_case_ff -> OK (Ebinop (Odivf, e1, e2))
  | Coq_div_case_if sg -> OK (Ebinop (Odivf, (make_floatofint e1 sg), e2))
  | Coq_div_case_fi sg -> OK (Ebinop (Odivf, e1, (make_floatofint e2 sg)))
  | Coq_div_default ->
    Error
      (msg
        ('C'::('s'::('h'::('m'::('g'::('e'::('n'::('.'::('m'::('a'::('k'::('e'::('_'::('d'::('i'::('v'::[])))))))))))))))))

let make_mod :  expr -> coq_type -> expr -> coq_type -> expr res =
  fun e1 ty1 e2 ty2  ->
  match classify_binint ty1 ty2 with
  | Coq_binint_case_ii s ->
    (match s with
     | Signed -> OK (Ebinop (Cminor.Omod, e1, e2))
     | Unsigned -> OK (Ebinop (Omodu, e1, e2)))
  | Coq_binint_default ->
    Error
      (msg
        ('C'::('s'::('h'::('m'::('g'::('e'::('n'::('.'::('m'::('a'::('k'::('e'::('_'::('m'::('o'::('d'::[])))))))))))))))))

let make_and :  expr -> coq_type -> expr -> coq_type -> expr res =
  fun e1 ty1 e2 ty2  ->
  OK (Ebinop (Cminor.Oand, e1, e2))

let make_or :  expr -> coq_type -> expr -> coq_type -> expr res =
  fun e1 ty1 e2 ty2  ->
  OK (Ebinop (Cminor.Oor, e1, e2))

let make_xor :  expr -> coq_type -> expr -> coq_type -> expr res =
  fun e1 ty1 e2 ty2  ->
  OK (Ebinop (Cminor.Oxor, e1, e2))

let make_shl :  expr -> coq_type -> expr -> coq_type -> expr res =
  fun e1 ty1 e2 ty2  ->
  OK (Ebinop (Cminor.Oshl, e1, e2))

let make_shr :  expr -> coq_type -> expr -> coq_type -> expr res =
  fun e1 ty1 e2 ty2  ->
  match classify_shift ty1 ty2 with
  | Coq_shift_case_ii s ->
    (match s with
     | Signed -> OK (Ebinop (Cminor.Oshr, e1, e2))
     | Unsigned -> OK (Ebinop (Oshru, e1, e2)))
  | Coq_shift_default ->
    Error
      (msg
        ('C'::('s'::('h'::('m'::('g'::('e'::('n'::('.'::('m'::('a'::('k'::('e'::('_'::('s'::('h'::('r'::[])))))))))))))))))

let make_cmp :     comparison -> expr -> coq_type -> expr -> coq_type -> expr res =
  fun c e1 ty1 e2 ty2  ->
  match classify_cmp ty1 ty2 with
  | Coq_cmp_case_ii s ->
    (match s with
     | Signed -> OK (Ebinop ((Ocmp c), e1, e2))
     | Unsigned -> OK (Ebinop ((Ocmpu c), e1, e2)))
  | Coq_cmp_case_pp -> OK (Ebinop ((Ocmpu c), e1, e2))
  | Coq_cmp_case_ff -> OK (Ebinop ((Ocmpf c), e1, e2))
  | Coq_cmp_case_if sg ->
    OK (Ebinop ((Ocmpf c), (make_floatofint e1 sg), e2))
  | Coq_cmp_case_fi sg ->
    OK (Ebinop ((Ocmpf c), e1, (make_floatofint e2 sg)))
  | Coq_cmp_default ->
    Error
      (msg
        ('C'::('s'::('h'::('m'::('g'::('e'::('n'::('.'::('m'::('a'::('k'::('e'::('_'::('c'::('m'::('p'::[])))))))))))))))))

let make_cast_int :  expr -> intsize -> signedness -> expr =
  fun e sz si  ->
  match sz with
  | I8 ->
    (match si with
     | Signed -> Eunop (Ocast8signed, e)
     | Unsigned -> Eunop (Ocast8unsigned, e))
  | I16 ->
    (match si with
     | Signed -> Eunop (Ocast16signed, e)
     | Unsigned -> Eunop (Ocast16unsigned, e))
  | I32 -> e
  | IBool -> make_cmp_ne_zero e

let make_cast_float :  expr -> floatsize -> expr =
  fun e  ->
    function
| F32 -> Eunop (Osingleoffloat, e)
| F64 -> e

let make_cast :  coq_type -> coq_type -> expr -> expr =
  fun from to0 e  ->
  match classify_cast from to0 with
  | Coq_cast_case_i2i (sz2, si2) -> make_cast_int e sz2 si2
  | Coq_cast_case_f2f sz2 -> make_cast_float e sz2
  | Coq_cast_case_i2f (si1, sz2) ->
    make_cast_float (make_floatofint e si1) sz2
  | Coq_cast_case_f2i (sz2, si2) ->
    make_cast_int (make_intoffloat e si2) sz2 si2
  | Coq_cast_case_f2bool ->
    Ebinop ((Ocmpf Cne), e, (make_floatconst Float.zero))
  | Coq_cast_case_p2bool -> Ebinop ((Ocmpu Cne), e, (make_intconst Int.zero))
  | _ -> e

let make_load :  expr -> coq_type -> expr res =
  fun addr ty_res  ->
  match access_mode ty_res with
  | By_value chunk -> OK (Eload (chunk, addr))
  | By_nothing ->
    Error
      (msg
        ('C'::('s'::('h'::('m'::('g'::('e'::('n'::('.'::('m'::('a'::('k'::('e'::('_'::('l'::('o'::('a'::('d'::[]))))))))))))))))))
  | _ -> OK addr

let make_memcpy :  expr -> expr -> coq_type -> stmt =
  fun dst src ty  ->
  Sbuiltin (None, (EF_memcpy ((sizeof ty), (alignof ty))),
    (dst :: (src :: [])))

let make_store :  expr -> coq_type -> expr -> stmt res =
  fun addr ty rhs  ->
  match access_mode ty with
  | By_value chunk -> OK (Sstore (chunk, addr, rhs))
  | By_copy -> OK (make_memcpy addr rhs ty)
  | _ ->
    Error
      (msg
        ('C'::('s'::('h'::('m'::('g'::('e'::('n'::('.'::('m'::('a'::('k'::('e'::('_'::('s'::('t'::('o'::('r'::('e'::[])))))))))))))))))))

let transl_unop :  Cop.unary_operation -> expr -> coq_type -> expr res =
  fun op a ta  ->
  match op with
  | Onotbool -> make_notbool a ta
  | Onotint -> OK (make_notint a ta)
  | Oneg -> make_neg a ta

let transl_binop :     Cop.binary_operation -> expr -> coq_type -> expr -> coq_type -> expr res =
  fun op a ta b tb  ->
  match op with
  | Oadd -> make_add a ta b tb
  | Osub -> make_sub a ta b tb
  | Omul -> make_mul a ta b tb
  | Odiv -> make_div a ta b tb
  | Omod -> make_mod a ta b tb
  | Oand -> make_and a ta b tb
  | Oor -> make_or a ta b tb
  | Oxor -> make_xor a ta b tb
  | Oshl -> make_shl a ta b tb
  | Oshr -> make_shr a ta b tb
  | Oeq -> make_cmp Ceq a ta b tb
  | One -> make_cmp Cne a ta b tb
  | Olt -> make_cmp Clt a ta b tb
  | Ogt -> make_cmp Cgt a ta b tb
  | Ole -> make_cmp Cle a ta b tb
  | Oge -> make_cmp Cge a ta b tb

let rec transl_expr :  Clight.expr -> expr res =
    function
| Econst_int (n, t) -> OK (make_intconst n)
| Econst_float (n, t) -> OK (make_floatconst n)
| Clight.Evar (id, ty) -> make_load (Eaddrof id) ty
| Etempvar (id, ty) -> OK (Evar id)
| Ederef (b, ty) ->
  (match transl_expr b with
   | OK x -> make_load x ty
   | Error msg0 -> Error msg0)
| Clight.Eaddrof (b, t) -> transl_lvalue b
| Clight.Eunop (op, b, t) ->
  (match transl_expr b with
   | OK x -> transl_unop op x (typeof b)
   | Error msg0 -> Error msg0)
| Clight.Ebinop (op, b, c, t) ->
  (match transl_expr b with
   | OK x ->
     (match transl_expr c with
      | OK x0 -> transl_binop op x (typeof b) x0 (typeof c)
      | Error msg0 -> Error msg0)
   | Error msg0 -> Error msg0)
| Ecast (b, ty) ->
  (match transl_expr b with
   | OK x -> OK (make_cast (typeof b) ty x)
   | Error msg0 -> Error msg0)
| Efield (b, i, ty) ->
  (match typeof b with
   | Tstruct (i0, fld, a0) ->
     (match transl_expr b with
      | OK x ->
        (match field_offset i fld with
         | OK x0 ->
           make_load (Ebinop (Cminor.Oadd, x, (make_intconst (Int.repr x0))))
             ty
         | Error msg0 -> Error msg0)
      | Error msg0 -> Error msg0)
   | Tunion (i0, fld, a0) ->
     (match transl_expr b with
      | OK x -> make_load x ty
      | Error msg0 -> Error msg0)
   | _ ->
     Error
       (msg
         ('C'::('s'::('h'::('m'::('g'::('e'::('n'::('.'::('t'::('r'::('a'::('n'::('s'::('l'::('_'::('e'::('x'::('p'::('r'::('('::('f'::('i'::('e'::('l'::('d'::(')'::[]))))))))))))))))))))))))))))

and transl_lvalue :  Clight.expr -> expr res =
    function
| Clight.Evar (id, t) -> OK (Eaddrof id)
| Ederef (b, t) -> transl_expr b
| Efield (b, i, ty) ->
  (match typeof b with
   | Tstruct (i0, fld, a0) ->
     (match transl_expr b with
      | OK x ->
        (match field_offset i fld with
         | OK x0 ->
           OK (Ebinop (Cminor.Oadd, x, (make_intconst (Int.repr x0))))
         | Error msg0 -> Error msg0)
      | Error msg0 -> Error msg0)
   | Tunion (i0, fld, a0) -> transl_expr b
   | _ ->
     Error
       (msg
         ('C'::('s'::('h'::('m'::('g'::('e'::('n'::('.'::('t'::('r'::('a'::('n'::('s'::('l'::('_'::('l'::('v'::('a'::('l'::('u'::('e'::('('::('f'::('i'::('e'::('l'::('d'::(')'::[]))))))))))))))))))))))))))))))
| _ ->
  Error
    (msg
      ('C'::('s'::('h'::('m'::('g'::('e'::('n'::('.'::('t'::('r'::('a'::('n'::('s'::('l'::('_'::('l'::('v'::('a'::('l'::('u'::('e'::[]))))))))))))))))))))))

let rec transl_arglist :  Clight.expr list -> typelist -> expr list res =
  fun al tyl  ->
  match al with
  | [] ->
    (match tyl with
     | Tnil -> OK []
     | Tcons (t, t0) ->
       Error
         (msg
           ('C'::('s'::('h'::('m'::('g'::('e'::('n'::('.'::('t'::('r'::('a'::('n'::('s'::('l'::('_'::('a'::('r'::('g'::('l'::('i'::('s'::('t'::(':'::(' '::('a'::('r'::('i'::('t'::('y'::(' '::('m'::('i'::('s'::('m'::('a'::('t'::('c'::('h'::[]))))))))))))))))))))))))))))))))))))))))
  | a1 :: a2 ->
    (match tyl with
     | Tnil ->
       Error
         (msg
           ('C'::('s'::('h'::('m'::('g'::('e'::('n'::('.'::('t'::('r'::('a'::('n'::('s'::('l'::('_'::('a'::('r'::('g'::('l'::('i'::('s'::('t'::(':'::(' '::('a'::('r'::('i'::('t'::('y'::(' '::('m'::('i'::('s'::('m'::('a'::('t'::('c'::('h'::[])))))))))))))))))))))))))))))))))))))))
     | Tcons (ty1, ty2) ->
       (match transl_expr a1 with
        | OK x ->
          (match transl_arglist a2 ty2 with
           | OK x0 -> OK ((make_cast (typeof a1) ty1 x) :: x0)
           | Error msg0 -> Error msg0)
        | Error msg0 -> Error msg0))

let rec transl_statement :     coq_type -> nat -> nat -> statement -> stmt res =
  fun tyret nbrk ncnt  ->
    function
| Clight.Sskip -> OK Sskip
| Clight.Sassign (b, c) ->
  (match transl_lvalue b with
   | OK x ->
     (match transl_expr c with
      | OK x0 -> make_store x (typeof b) (make_cast (typeof c) (typeof b) x0)
      | Error msg0 -> Error msg0)
   | Error msg0 -> Error msg0)
| Clight.Sset (x, b) ->
  (match transl_expr b with
   | OK x0 -> OK (Sset (x, x0))
   | Error msg0 -> Error msg0)
| Clight.Scall (x, b, cl) ->
  (match classify_fun (typeof b) with
   | Coq_fun_case_f (args, res0) ->
     (match transl_expr b with
      | OK x0 ->
        (match transl_arglist cl args with
         | OK x1 -> OK (Scall (x, (signature_of_type args res0), x0, x1))
         | Error msg0 -> Error msg0)
      | Error msg0 -> Error msg0)
   | Coq_fun_default ->
     Error
       (msg
         ('C'::('s'::('h'::('m'::('g'::('e'::('n'::('.'::('t'::('r'::('a'::('n'::('s'::('l'::('_'::('s'::('t'::('m'::('t'::('('::('c'::('a'::('l'::('l'::(')'::[])))))))))))))))))))))))))))
| Clight.Sbuiltin (x, ef, tyargs, bl) ->
  (match transl_arglist bl tyargs with
   | OK x0 -> OK (Sbuiltin (x, ef, x0))
   | Error msg0 -> Error msg0)
| Ssequence (s1, s2) ->
  (match transl_statement tyret nbrk ncnt s1 with
   | OK x ->
     (match transl_statement tyret nbrk ncnt s2 with
      | OK x0 -> OK (Sseq (x, x0))
      | Error msg0 -> Error msg0)
   | Error msg0 -> Error msg0)
| Clight.Sifthenelse (e, s1, s2) ->
  (match transl_expr e with
   | OK x ->
     (match transl_statement tyret nbrk ncnt s1 with
      | OK x0 ->
        (match transl_statement tyret nbrk ncnt s2 with
         | OK x1 -> OK (Sifthenelse ((make_boolean x (typeof e)), x0, x1))
         | Error msg0 -> Error msg0)
      | Error msg0 -> Error msg0)
   | Error msg0 -> Error msg0)
| Clight.Sloop (s1, s2) ->
  (match transl_statement tyret (S O) O s1 with
   | OK x ->
     (match transl_statement tyret O (S ncnt) s2 with
      | OK x0 -> OK (Sblock (Sloop (Sseq ((Sblock x), x0))))
      | Error msg0 -> Error msg0)
   | Error msg0 -> Error msg0)
| Sbreak -> OK (Sexit nbrk)
| Scontinue -> OK (Sexit ncnt)
| Clight.Sreturn o ->
  (match o with
   | Some e ->
     (match transl_expr e with
      | OK x -> OK (Sreturn (Some (make_cast (typeof e) tyret x)))
      | Error msg0 -> Error msg0)
   | None -> OK (Sreturn None))
| Clight.Sswitch (a, sl) ->
  (match transl_expr a with
   | OK x ->
     (match transl_lbl_stmt tyret O (S ncnt) sl with
      | OK x0 -> OK (Sblock (Sswitch (x, x0)))
      | Error msg0 -> Error msg0)
   | Error msg0 -> Error msg0)
| Clight.Slabel (lbl, s0) ->
  (match transl_statement tyret nbrk ncnt s0 with
   | OK x -> OK (Slabel (lbl, x))
   | Error msg0 -> Error msg0)
| Clight.Sgoto lbl -> OK (Sgoto lbl)

and transl_lbl_stmt :     coq_type -> nat -> nat -> labeled_statements -> lbl_stmt res =
  fun tyret nbrk ncnt  ->
    function
| Clight.LSdefault s ->
  (match transl_statement tyret nbrk ncnt s with
   | OK x -> OK (LSdefault x)
   | Error msg0 -> Error msg0)
| Clight.LScase (n, s, sl') ->
  (match transl_statement tyret nbrk ncnt s with
   | OK x ->
     (match transl_lbl_stmt tyret nbrk ncnt sl' with
      | OK x0 -> OK (LScase (n, x, x0))
      | Error msg0 -> Error msg0)
   | Error msg0 -> Error msg0)

let transl_var :  (ident * coq_type) -> ident * coq_Z =
  fun v  ->
  ((fst v), (sizeof (snd v)))

let signature_of_function :  Clight.coq_function -> signature =
  fun f  ->
  { sig_args = (map typ_of_type (map snd (Clight.fn_params f))); sig_res =
    (opttyp_of_type f.fn_return) }

let transl_function :  Clight.coq_function -> coq_function res =
  fun f  ->
  match transl_statement f.fn_return (S O) O (Clight.fn_body f) with
  | OK x ->
    OK { fn_sig = (signature_of_function f); fn_params =
      (map fst (Clight.fn_params f)); fn_vars =
      (map transl_var (Clight.fn_vars f)); fn_temps =
      (map fst (Clight.fn_temps f)); fn_body = x }
  | Error msg0 -> Error msg0

let list_typ_eq :  typ list -> typ list -> bool =
  fun x y ->
  list_eq_dec typ_eq x y

let transl_fundef :  Clight.fundef -> fundef res =
    function
| Internal g ->
  (match transl_function g with
   | OK x -> OK (AST.Internal x)
   | Error msg0 -> Error msg0)
| External (ef, args, res0) ->
  if (&&)
       (proj_sumbool
         (list_typ_eq (ef_sig ef).sig_args (typlist_of_typelist args)))
       (proj_sumbool (opt_typ_eq (ef_sig ef).sig_res (opttyp_of_type res0)))
  then OK (AST.External ef)
  else Error
         (msg
           ('C'::('s'::('h'::('m'::('g'::('e'::('n'::('.'::('t'::('r'::('a'::('n'::('s'::('l'::('_'::('f'::('u'::('n'::('d'::('e'::('f'::(':'::(' '::('w'::('r'::('o'::('n'::('g'::(' '::('e'::('x'::('t'::('e'::('r'::('n'::('a'::('l'::(' '::('s'::('i'::('g'::('n'::('a'::('t'::('u'::('r'::('e'::[]))))))))))))))))))))))))))))))))))))))))))))))))

let transl_globvar :  coq_type -> unit res =
  fun ty  ->
  OK ()

let transl_program :  Clight.program -> program res =
  fun p  ->
  transform_partial_program2 transl_fundef transl_globvar p

