open AST
open BinInt
open BinNums
open Bool
open Coqlib
open Errors

type signedness =
| Signed
| Unsigned

type intsize =
| I8
| I16
| I32
| IBool

type floatsize =
| F32
| F64

type attr =
  bool
  (* singleton inductive, whose constructor was mk_attr *)

let attr_volatile :  attr -> bool =
  fun a  ->
  a

let noattr :  attr =
  false

type coq_type =
| Tvoid
| Tint of intsize * signedness * attr
| Tfloat of floatsize * attr
| Tpointer of coq_type * attr
| Tarray of coq_type * coq_Z * attr
| Tfunction of typelist * coq_type
| Tstruct of ident * fieldlist * attr
| Tunion of ident * fieldlist * attr
| Tcomp_ptr of ident * attr
and typelist =
| Tnil
| Tcons of coq_type * typelist
and fieldlist =
| Fnil
| Fcons of ident * coq_type * fieldlist

let rec type_eq :  coq_type -> coq_type -> bool =
  fun ty1 ty2 ->
  let h = fun x y ->
    match x with
    | I8 ->
      (match y with
       | I8 -> true
       | _ -> false)
    | I16 ->
      (match y with
       | I16 -> true
       | _ -> false)
    | I32 ->
      (match y with
       | I32 -> true
       | _ -> false)
    | IBool ->
      (match y with
       | IBool -> true
       | _ -> false)
  in
  let h0 = fun x y ->
    match x with
    | Signed ->
      (match y with
       | Signed -> true
       | Unsigned -> false)
    | Unsigned ->
      (match y with
       | Signed -> false
       | Unsigned -> true)
  in
  let h1 = fun x y ->
    match x with
    | F32 ->
      (match y with
       | F32 -> true
       | F64 -> false)
    | F64 ->
      (match y with
       | F32 -> false
       | F64 -> true)
  in
  let h2 = fun x y -> bool_dec x y in
  (
  match ty1 with
  | Tvoid ->
    (match ty2 with
     | Tvoid -> true
     | _ -> false)
  | Tint (i, s, a) ->
    (match ty2 with
     | Tint (i0, s0, a0) ->
       if h i i0 then if h0 s s0 then h2 a a0 else false else false
     | _ -> false)
  | Tfloat (f, a) ->
    (match ty2 with
     | Tfloat (f0, a0) -> if h1 f f0 then h2 a a0 else false
     | _ -> false)
  | Tpointer (t, a) ->
    (match ty2 with
     | Tpointer (t0, a0) -> if type_eq t t0 then h2 a a0 else false
     | _ -> false)
  | Tarray (t, z, a) ->
    (match ty2 with
     | Tarray (t0, z0, a0) ->
       if type_eq t t0 then if zeq z z0 then h2 a a0 else false else false
     | _ -> false)
  | Tfunction (t, t0) ->
    (match ty2 with
     | Tfunction (t1, t2) ->
       if typelist_eq t t1 then type_eq t0 t2 else false
     | _ -> false)
  | Tstruct (i, f, a) ->
    (match ty2 with
     | Tstruct (i0, f0, a0) ->
       if ident_eq i i0
       then if fieldlist_eq f f0 then h2 a a0 else false
       else false
     | _ -> false)
  | Tunion (i, f, a) ->
    (match ty2 with
     | Tunion (i0, f0, a0) ->
       if ident_eq i i0
       then if fieldlist_eq f f0 then h2 a a0 else false
       else false
     | _ -> false)
  | Tcomp_ptr (i, a) ->
    (match ty2 with
     | Tcomp_ptr (i0, a0) -> if ident_eq i i0 then h2 a a0 else false
     | _ -> false))

and typelist_eq :  typelist -> typelist -> bool =
  fun tyl1 tyl2  ->
  match tyl1 with
  | Tnil ->
    (match tyl2 with
     | Tnil -> true
     | Tcons (t, t0) -> false)
  | Tcons (t, t0) ->
    (match tyl2 with
     | Tnil -> false
     | Tcons (t1, t2) -> if type_eq t t1 then typelist_eq t0 t2 else false)

and fieldlist_eq :  fieldlist -> fieldlist -> bool =
  fun fld1 fld2  ->
  match fld1 with
  | Fnil ->
    (match fld2 with
     | Fnil -> true
     | Fcons (i, t, f) -> false)
  | Fcons (i, t, f) ->
    (match fld2 with
     | Fnil -> false
     | Fcons (i0, t0, f0) ->
       if ident_eq i i0
       then if type_eq t t0 then fieldlist_eq f f0 else false
       else false)

let attr_of_type :  coq_type -> attr =
    function
| Tint (sz, si, a) -> a
| Tfloat (sz, a) -> a
| Tpointer (elt, a) -> a
| Tarray (elt, sz, a) -> a
| Tstruct (id, fld, a) -> a
| Tunion (id, fld, a) -> a
| Tcomp_ptr (id, a) -> a
| _ -> noattr

let type_int32s :  coq_type =
  Tint (I32, Signed, noattr)

let type_bool :  coq_type =
  Tint (IBool, Signed, noattr)

let typeconv :  coq_type -> coq_type =
  fun ty  ->
    match ty with
| Tint (i, s, a) ->
  (match i with
   | I32 ->
     (match s with
      | Signed -> Tint (I32, Signed, a)
      | Unsigned -> ty)
   | _ -> Tint (I32, Signed, a))
| Tarray (t, sz, a) -> Tpointer (t, a)
| Tfunction (t, t0) -> Tpointer (ty, noattr)
| _ -> ty

let rec alignof :  coq_type -> coq_Z =
    function
| Tvoid -> Zpos Coq_xH
| Tint (i, s, a) ->
  (match i with
   | I16 -> Zpos (Coq_xO Coq_xH)
   | I32 -> Zpos (Coq_xO (Coq_xO Coq_xH))
   | _ -> Zpos Coq_xH)
| Tfloat (f, a) ->
  (match f with
   | F32 -> Zpos (Coq_xO (Coq_xO Coq_xH))
   | F64 -> Zpos (Coq_xO (Coq_xO (Coq_xO Coq_xH))))
| Tarray (t', z, a) -> alignof t'
| Tfunction (t0, t1) -> Zpos Coq_xH
| Tstruct (i, fld, a) -> alignof_fields fld
| Tunion (i, fld, a) -> alignof_fields fld
| _ -> Zpos (Coq_xO (Coq_xO Coq_xH))

and alignof_fields :  fieldlist -> coq_Z =
    function
| Fnil -> Zpos Coq_xH
| Fcons (id, t, f') -> Z.max (alignof t) (alignof_fields f')

let rec sizeof :  coq_type -> coq_Z =
  fun t  ->
    match t with
| Tvoid -> Zpos Coq_xH
| Tint (i, s, a) ->
  (match i with
   | I16 -> Zpos (Coq_xO Coq_xH)
   | I32 -> Zpos (Coq_xO (Coq_xO Coq_xH))
   | _ -> Zpos Coq_xH)
| Tfloat (f, a) ->
  (match f with
   | F32 -> Zpos (Coq_xO (Coq_xO Coq_xH))
   | F64 -> Zpos (Coq_xO (Coq_xO (Coq_xO Coq_xH))))
| Tarray (t', n, a) -> Z.mul (sizeof t') (Z.max (Zpos Coq_xH) n)
| Tfunction (t0, t1) -> Zpos Coq_xH
| Tstruct (i, fld, a) ->
  align (Z.max (Zpos Coq_xH) (sizeof_struct fld Z0)) (alignof t)
| Tunion (i, fld, a) ->
  align (Z.max (Zpos Coq_xH) (sizeof_union fld)) (alignof t)
| _ -> Zpos (Coq_xO (Coq_xO Coq_xH))

and sizeof_struct :  fieldlist -> coq_Z -> coq_Z =
  fun fld pos  ->
  match fld with
  | Fnil -> pos
  | Fcons (id, t, fld') ->
    sizeof_struct fld' (Z.add (align pos (alignof t)) (sizeof t))

and sizeof_union :  fieldlist -> coq_Z =
    function
| Fnil -> Z0
| Fcons (id, t, fld') -> Z.max (sizeof t) (sizeof_union fld')

let rec field_offset_rec :  ident -> fieldlist -> coq_Z -> coq_Z res =
  fun id fld pos  ->
  match fld with
  | Fnil ->
    Error ((MSG
      ('U'::('n'::('k'::('n'::('o'::('w'::('n'::(' '::('f'::('i'::('e'::('l'::('d'::(' '::[]))))))))))))))) :: ((CTX
      id) :: []))
  | Fcons (id', t, fld') ->
    if ident_eq id id'
    then OK (align pos (alignof t))
    else field_offset_rec id fld' (Z.add (align pos (alignof t)) (sizeof t))

let field_offset :  ident -> fieldlist -> coq_Z res =
  fun id fld  ->
  field_offset_rec id fld Z0

type mode =
| By_value of memory_chunk
| By_reference
| By_copy
| By_nothing

let access_mode :  coq_type -> mode =
    function
| Tvoid -> By_nothing
| Tint (i, s, a) ->
  (match i with
   | I8 ->
     (match s with
      | Signed -> By_value Mint8signed
      | Unsigned -> By_value Mint8unsigned)
   | I16 ->
     (match s with
      | Signed -> By_value Mint16signed
      | Unsigned -> By_value Mint16unsigned)
   | I32 -> By_value Mint32
   | IBool -> By_value Mint8unsigned)
| Tfloat (f, a) ->
  (match f with
   | F32 -> By_value Mfloat32
   | F64 -> By_value Mfloat64)
| Tpointer (t, a) -> By_value Mint32
| Tstruct (i, f, a) -> By_copy
| Tunion (i, f, a) -> By_copy
| Tcomp_ptr (i, a) -> By_nothing
| _ -> By_reference

let type_is_volatile :  coq_type -> bool =
  fun ty  ->
  match access_mode ty with
  | By_value m -> attr_volatile (attr_of_type ty)
  | _ -> false

let unroll_composite :  ident -> coq_type -> coq_type -> coq_type =
  fun cid comp zza ->
  let rec unroll_composite0 ty = match ty with
  | Tpointer (t1, a) -> Tpointer ((unroll_composite0 t1), a)
  | Tarray (t1, sz, a) -> Tarray ((unroll_composite0 t1), sz, a)
  | Tfunction (t1, t2) ->
    Tfunction ((unroll_composite_list t1), (unroll_composite0 t2))
  | Tstruct (id, fld, a) ->
    if ident_eq id cid
    then ty
    else Tstruct (id, (unroll_composite_fields fld), a)
  | Tunion (id, fld, a) ->
    if ident_eq id cid
    then ty
    else Tunion (id, (unroll_composite_fields fld), a)
  | Tcomp_ptr (id, a) -> if ident_eq id cid then Tpointer (comp, a) else ty
  | _ -> ty
  and unroll_composite_list = function
  | Tnil -> Tnil
  | Tcons (t1, tl') ->
    Tcons ((unroll_composite0 t1), (unroll_composite_list tl'))
  and unroll_composite_fields = function
  | Fnil -> Fnil
  | Fcons (id, ty, fld') ->
    Fcons (id, (unroll_composite0 ty), (unroll_composite_fields fld'))
  in unroll_composite0 zza

let rec type_of_params :  (ident * coq_type) list -> typelist =
    function
| [] -> Tnil
| p :: rem -> let (id, ty) = p in Tcons (ty, (type_of_params rem))

let typ_of_type :  coq_type -> typ =
    function
| Tfloat (f, a) -> AST.Tfloat
| _ -> AST.Tint

let opttyp_of_type :  coq_type -> typ option =
    function
| Tvoid -> None
| Tfloat (f, a) -> Some AST.Tfloat
| _ -> Some AST.Tint

let rec typlist_of_typelist :  typelist -> typ list =
    function
| Tnil -> []
| Tcons (hd, tl0) -> (typ_of_type hd) :: (typlist_of_typelist tl0)

let signature_of_type :  typelist -> coq_type -> signature =
  fun args res0  ->
  { sig_args = (typlist_of_typelist args); sig_res = (opttyp_of_type res0) }

