open AST
open BinInt
open BinNums
open BinPos
open Coqlib
open Datatypes
open Errors
open Integers
open List0
open Maps
open Op
open RTL
open Registers

type __ = Obj.t
let __ = let rec f _ = Obj.repr f in Obj.repr f

type funenv = coq_function PTree.t

let add_globdef :     funenv -> (ident * (fundef, unit) globdef) -> funenv =
  fun fenv  ->
    function
| (id, g) ->
  (match g with
   | Gfun f0 ->
     (match f0 with
      | Internal f ->
        if Inliningaux.should_inline id f
        then PTree.set id f fenv
        else PTree.remove id fenv
      | External e -> PTree.remove id fenv)
   | Gvar v -> PTree.remove id fenv)

let funenv_program :  program -> funenv =
  fun p  ->
  fold_left add_globdef p.prog_defs PTree.empty

let max_pc_function :  coq_function -> positive =
  fun f  ->
  PTree.fold (fun m pc i -> Pos.max m pc) f.fn_code Coq_xH

let max_def_instr :  instruction -> positive =
    function
| Iop (op, args, res0, s) -> res0
| Iload (chunk, addr, args, dst, s) -> dst
| Icall (sg, ros, args, res0, s) -> res0
| Ibuiltin (ef, args, res0, s) -> res0
| _ -> Coq_xH

let max_def_function :  coq_function -> positive =
  fun f  ->
  Pos.max
    (PTree.fold (fun m pc i -> Pos.max m (max_def_instr i)) f.fn_code Coq_xH)
    (fold_left (fun m r -> Pos.max m r) f.fn_params Coq_xH)

type state = { st_nextreg : positive; st_nextnode : positive; st_code : 
               code; st_stksize : coq_Z }

let st_nextreg :  state -> positive =
  fun x  ->
    x.st_nextreg

let st_nextnode :  state -> positive =
  fun x  ->
    x.st_nextnode

let st_code :  state -> code =
  fun x  ->
    x.st_code

let st_stksize :  state -> coq_Z =
  fun x  ->
    x.st_stksize

type 'a res =
| R of 'a * state

type 'a mon = state -> 'a res

let initstate :  state =
  { st_nextreg = Coq_xH; st_nextnode = Coq_xH; st_code = PTree.empty;
    st_stksize = Z0 }

let set_instr :  node -> instruction -> unit mon =
  fun pc i s  ->
  R ((), { st_nextreg = s.st_nextreg; st_nextnode = s.st_nextnode; st_code =
    (PTree.set pc i s.st_code); st_stksize = s.st_stksize })

let add_instr :  instruction -> node mon =
  fun i s  ->
  let pc = Pos.succ s.st_nextnode in
  R (pc, { st_nextreg = s.st_nextreg; st_nextnode = pc; st_code =
  (PTree.set pc i s.st_code); st_stksize = s.st_stksize })

let reserve_nodes :  positive -> positive mon =
  fun numnodes s  ->
  R (s.st_nextnode, { st_nextreg = s.st_nextreg; st_nextnode =
    (Pos.add s.st_nextnode numnodes); st_code = s.st_code; st_stksize =
    s.st_stksize })

let reserve_regs :  positive -> positive mon =
  fun numregs s  ->
  R (s.st_nextreg, { st_nextreg = (Pos.add s.st_nextreg numregs);
    st_nextnode = s.st_nextnode; st_code = s.st_code; st_stksize =
    s.st_stksize })

let request_stack :  coq_Z -> unit mon =
  fun sz s  ->
  R ((), { st_nextreg = s.st_nextreg; st_nextnode = s.st_nextnode; st_code =
    s.st_code; st_stksize = (Z.max s.st_stksize sz) })

let rec mlist_iter2 :     ('a1 -> 'a2 -> unit mon) -> ('a1 * 'a2) list -> unit mon =
  fun f l s  ->
  match l with
  | [] -> R ((), s)
  | p :: l' ->
    let (x, y) = p in let (R (vx, s2)) = f x y s in mlist_iter2 f l' s2

type context = { dpc : positive; dreg : positive; dstk : coq_Z;
                 mreg : positive; mstk : coq_Z; retinfo : (node * reg) option }

let dpc :  context -> positive =
  fun x  ->
    x.dpc

let dreg :  context -> positive =
  fun x  ->
    x.dreg

let dstk :  context -> coq_Z =
  fun x  ->
    x.dstk

let mstk :  context -> coq_Z =
  fun x  ->
    x.mstk

let retinfo :  context -> (node * reg) option =
  fun x  ->
    x.retinfo

let spc :  context -> node -> positive =
  fun ctx pc  ->
  Pos.add pc ctx.dpc

let sreg :  context -> reg -> positive =
  fun ctx r  ->
  Pos.add r ctx.dreg

let sregs :  context -> reg list -> positive list =
  fun ctx rl  ->
  map (sreg ctx) rl

let sros :  context -> (reg, ident) sum -> (positive, ident) sum =
  fun ctx ros  ->
  sum_left_map (sreg ctx) ros

let sop :  context -> operation -> operation =
  fun ctx op  ->
  shift_stack_operation (Int.repr ctx.dstk) op

let saddr :  context -> addressing -> addressing =
  fun ctx addr  ->
  shift_stack_addressing (Int.repr ctx.dstk) addr

let initcontext :     positive -> positive -> positive -> coq_Z -> context =
  fun dpc0 dreg0 nreg sz  ->
  { dpc = dpc0; dreg = dreg0; dstk = Z0; mreg = nreg; mstk = (Z.max sz Z0);
    retinfo = None }

let min_alignment :  coq_Z -> coq_Z =
  fun sz  ->
  if zle sz (Zpos Coq_xH)
  then Zpos Coq_xH
  else if zle sz (Zpos (Coq_xO Coq_xH))
       then Zpos (Coq_xO Coq_xH)
       else if zle sz (Zpos (Coq_xO (Coq_xO Coq_xH)))
            then Zpos (Coq_xO (Coq_xO Coq_xH))
            else Zpos (Coq_xO (Coq_xO (Coq_xO Coq_xH)))

let callcontext :     context -> positive -> positive -> positive -> coq_Z -> node -> reg ->
    context =
  fun ctx dpc0 dreg0 nreg sz retpc retreg  ->
  { dpc = dpc0; dreg = dreg0; dstk =
    (align (Z.add ctx.dstk ctx.mstk) (min_alignment sz)); mreg = nreg; mstk =
    (Z.max sz Z0); retinfo = (Some ((spc ctx retpc), (sreg ctx retreg))) }

let tailcontext :     context -> positive -> positive -> positive -> coq_Z -> context =
  fun ctx dpc0 dreg0 nreg sz  ->
  { dpc = dpc0; dreg = dreg0; dstk = (align ctx.dstk (min_alignment sz));
    mreg = nreg; mstk = (Z.max sz Z0); retinfo = ctx.retinfo }

let rec add_moves :  reg list -> reg list -> node -> node mon =
  fun srcs dsts succ0 s  ->
  match srcs with
  | [] -> R (succ0, s)
  | s1 :: sl ->
    (match dsts with
     | [] -> R (succ0, s)
     | d1 :: dl ->
       let (R (vx, s2)) = add_instr (Iop (Omove, (s1 :: []), d1, succ0)) s in
       add_moves sl dl vx s2)

type inline_decision =
| Cannot_inline
| Can_inline of ident * coq_function

let can_inline :  funenv -> (reg, ident) sum -> inline_decision =
  fun fenv  ->
    function
| Coq_inl r -> Cannot_inline
| Coq_inr id ->
  let filtered_var = PTree.get id fenv in
  (match filtered_var with
   | Some f -> Can_inline (id, f)
   | None -> Cannot_inline)

let inline_function :     funenv -> (funenv -> __ -> context -> coq_function -> unit mon) ->
    context -> ident -> coq_function -> reg list -> node -> reg -> node mon =
  fun fenv rec0 ctx id f args retpc retreg  ->
  let npc = max_pc_function f in
  let nreg = max_def_function f in
  (fun s1 ->
  let (R (vx, s2)) = reserve_nodes npc s1 in
  let (R (vx0, s3)) = reserve_regs nreg s2 in
  let ctx' = callcontext ctx vx vx0 nreg f.fn_stacksize retpc retreg in
  let (R (vx1, s4)) = rec0 (PTree.remove id fenv) __ ctx' f s3 in
  add_moves (sregs ctx args) (sregs ctx' f.fn_params)
    (spc ctx' f.fn_entrypoint) s4)

let inline_tail_function :     funenv -> (funenv -> __ -> context -> coq_function -> unit mon) ->
    context -> ident -> coq_function -> reg list -> node mon =
  fun fenv rec0 ctx id f args  ->
  let npc = max_pc_function f in
  let nreg = max_def_function f in
  (fun s1 ->
  let (R (vx, s2)) = reserve_nodes npc s1 in
  let (R (vx0, s3)) = reserve_regs nreg s2 in
  let ctx' = tailcontext ctx vx vx0 nreg f.fn_stacksize in
  let (R (vx1, s4)) = rec0 (PTree.remove id fenv) __ ctx' f s3 in
  add_moves (sregs ctx args) (sregs ctx' f.fn_params)
    (spc ctx' f.fn_entrypoint) s4)

let inline_return :     context -> reg option -> (node * reg) -> instruction =
  fun ctx or0  ->
    function
| (retpc, retreg) ->
  (match or0 with
   | Some r -> Iop (Omove, ((sreg ctx r) :: []), retreg, retpc)
   | None -> Inop retpc)

let expand_instr :     funenv -> (funenv -> __ -> context -> coq_function -> unit mon) ->
    context -> node -> instruction -> unit mon =
  fun fenv rec0 ctx pc  ->
    function
| Inop s -> set_instr (spc ctx pc) (Inop (spc ctx s))
| Iop (op, args, res0, s) ->
  set_instr (spc ctx pc) (Iop ((sop ctx op), (sregs ctx args),
    (sreg ctx res0), (spc ctx s)))
| Iload (chunk, addr, args, dst, s) ->
  set_instr (spc ctx pc) (Iload (chunk, (saddr ctx addr), (sregs ctx args),
    (sreg ctx dst), (spc ctx s)))
| Istore (chunk, addr, args, src, s) ->
  set_instr (spc ctx pc) (Istore (chunk, (saddr ctx addr), (sregs ctx args),
    (sreg ctx src), (spc ctx s)))
| Icall (sg, ros, args, res0, s) ->
  (match can_inline fenv ros with
   | Cannot_inline ->
     set_instr (spc ctx pc) (Icall (sg, (sros ctx ros), (sregs ctx args),
       (sreg ctx res0), (spc ctx s)))
   | Can_inline (id, f) ->
     (fun s1 ->
       let (R (vx, s2)) = inline_function fenv rec0 ctx id f args s res0 s1 in
       set_instr (spc ctx pc) (Inop vx) s2))
| Itailcall (sg, ros, args) ->
  (match can_inline fenv ros with
   | Cannot_inline ->
     (match ctx.retinfo with
      | Some p ->
        let (rpc, rreg) = p in
        set_instr (spc ctx pc) (Icall (sg, (sros ctx ros), (sregs ctx args),
          rreg, rpc))
      | None ->
        set_instr (spc ctx pc) (Itailcall (sg, (sros ctx ros),
          (sregs ctx args))))
   | Can_inline (id, f) ->
     (fun s1 ->
       let (R (vx, s2)) = inline_tail_function fenv rec0 ctx id f args s1 in
       set_instr (spc ctx pc) (Inop vx) s2))
| Ibuiltin (ef, args, res0, s) ->
  set_instr (spc ctx pc) (Ibuiltin (ef, (sregs ctx args), (sreg ctx res0),
    (spc ctx s)))
| Icond (cond, args, s1, s2) ->
  set_instr (spc ctx pc) (Icond (cond, (sregs ctx args), (spc ctx s1),
    (spc ctx s2)))
| Ijumptable (r, tbl) ->
  set_instr (spc ctx pc) (Ijumptable ((sreg ctx r), (map (spc ctx) tbl)))
| Ireturn or0 ->
  (match ctx.retinfo with
   | Some rinfo -> set_instr (spc ctx pc) (inline_return ctx or0 rinfo)
   | None -> set_instr (spc ctx pc) (Ireturn (option_map (sreg ctx) or0)))

let expand_cfg_rec :     funenv -> (funenv -> __ -> context -> coq_function -> unit mon) ->
    context -> coq_function -> unit mon =
  fun fenv rec0 ctx f s1  ->
  let (R (vx, s2)) = request_stack (Z.add ctx.dstk ctx.mstk) s1 in
  mlist_iter2 (expand_instr fenv rec0 ctx) (PTree.elements f.fn_code) s2

let rec expand_cfg :  funenv -> context -> coq_function -> unit mon =
  fun x a b ->
  expand_cfg_rec x (fun y _ -> expand_cfg y) a b

let expand_function :  funenv -> coq_function -> context mon =
  fun fenv f  ->
  let npc = max_pc_function f in
  let nreg = max_def_function f in
  (fun s1 ->
  let (R (vx, s2)) = reserve_nodes npc s1 in
  let (R (vx0, s3)) = reserve_regs nreg s2 in
  let ctx = initcontext vx vx0 nreg f.fn_stacksize in
  let (R (vx1, s4)) = expand_cfg fenv ctx f s3 in R (ctx, s4))

let transf_function :     funenv -> coq_function -> coq_function Errors.res =
  fun fenv f  ->
  let (R (ctx, s)) = expand_function fenv f initstate in
  if zlt s.st_stksize Int.max_unsigned
  then OK { fn_sig = f.fn_sig; fn_params = (sregs ctx f.fn_params);
         fn_stacksize = s.st_stksize; fn_code = s.st_code; fn_entrypoint =
         (spc ctx f.fn_entrypoint) }
  else Error
         (msg
           ('I'::('n'::('l'::('i'::('n'::('i'::('n'::('g'::(':'::(' '::('s'::('t'::('a'::('c'::('k'::(' '::('t'::('o'::('o'::(' '::('b'::('i'::('g'::[]))))))))))))))))))))))))

let transf_fundef :  funenv -> fundef -> fundef Errors.res =
  fun fenv fd  ->
  transf_partial_fundef (transf_function fenv) fd

let transf_program :  program -> program Errors.res =
  fun p  ->
  let fenv = funenv_program p in
  transform_partial_program (transf_fundef fenv) p

