open AST
open BinInt
open BinNums
open Coqlib
open Datatypes
open EqNat
open Floats
open Integers
open List0
open Values

let size_chunk :  memory_chunk -> coq_Z =
    function
| Mint8signed -> Zpos Coq_xH
| Mint8unsigned -> Zpos Coq_xH
| Mint32 -> Zpos (Coq_xO (Coq_xO Coq_xH))
| Mfloat32 -> Zpos (Coq_xO (Coq_xO Coq_xH))
| Mfloat64 -> Zpos (Coq_xO (Coq_xO (Coq_xO Coq_xH)))
| Mfloat64al32 -> Zpos (Coq_xO (Coq_xO (Coq_xO Coq_xH)))
| _ -> Zpos (Coq_xO Coq_xH)

let size_chunk_nat :  memory_chunk -> nat =
  fun chunk  ->
  nat_of_Z (size_chunk chunk)

let align_chunk :  memory_chunk -> coq_Z =
    function
| Mint8signed -> Zpos Coq_xH
| Mint8unsigned -> Zpos Coq_xH
| Mint16signed -> Zpos (Coq_xO Coq_xH)
| Mint16unsigned -> Zpos (Coq_xO Coq_xH)
| Mfloat64 -> Zpos (Coq_xO (Coq_xO (Coq_xO Coq_xH)))
| _ -> Zpos (Coq_xO (Coq_xO Coq_xH))

type memval =
| Undef
| Byte of Byte.int
| Pointer of block * Int.int * nat

let rec bytes_of_int :  nat -> coq_Z -> Byte.int list =
  fun n x  ->
  match n with
  | O -> []
  | S m ->
    (Byte.repr x) :: (bytes_of_int m
                       (Z.div x (Zpos (Coq_xO (Coq_xO (Coq_xO (Coq_xO (Coq_xO
                         (Coq_xO (Coq_xO (Coq_xO Coq_xH)))))))))))

let rec int_of_bytes :  Byte.int list -> coq_Z =
    function
| [] -> Z0
| b :: l' ->
  Z.add (Byte.unsigned b)
    (Z.mul (int_of_bytes l') (Zpos (Coq_xO (Coq_xO (Coq_xO (Coq_xO (Coq_xO
      (Coq_xO (Coq_xO (Coq_xO Coq_xH))))))))))

let big_endian :  bool =
    Memdataaux.big_endian

let rev_if_be :  Byte.int list -> Byte.int list =
  fun l  ->
  if big_endian then rev l else l

let encode_int :  nat -> coq_Z -> Byte.int list =
  fun sz x  ->
  rev_if_be (bytes_of_int sz x)

let decode_int :  Byte.int list -> coq_Z =
  fun b  ->
  int_of_bytes (rev_if_be b)

let inj_bytes :  Byte.int list -> memval list =
  fun bl  ->
  map (fun x -> Byte x) bl

let rec proj_bytes :  memval list -> Byte.int list option =
    function
| [] -> Some []
| m :: vl' ->
  (match m with
   | Byte b ->
     (match proj_bytes vl' with
      | Some bl -> Some (b :: bl)
      | None -> None)
   | _ -> None)

let rec inj_pointer :  nat -> block -> Int.int -> memval list =
  fun n b ofs  ->
  match n with
  | O -> []
  | S m -> (Pointer (b, ofs, m)) :: (inj_pointer m b ofs)

let rec check_pointer :  nat -> block -> Int.int -> memval list -> bool =
  fun n b ofs vl  ->
  match n with
  | O ->
    (match vl with
     | [] -> true
     | m :: l -> false)
  | S m ->
    (match vl with
     | [] -> false
     | m0 :: vl' ->
       (match m0 with
        | Pointer (b', ofs', m') ->
          (&&)
            ((&&)
              ((&&) (proj_sumbool (eq_block b b'))
                (proj_sumbool (Int.eq_dec ofs ofs'))) (beq_nat m m'))
            (check_pointer m b ofs vl')
        | _ -> false))

let proj_pointer :  memval list -> coq_val =
  fun vl  ->
    match vl with
| [] -> Vundef
| m :: vl' ->
  (match m with
   | Pointer (b, ofs, n) ->
     if check_pointer (S (S (S (S O)))) b ofs vl
     then Vptr (b, ofs)
     else Vundef
   | _ -> Vundef)

let encode_val :  memory_chunk -> coq_val -> memval list =
  fun chunk  ->
    function
| Vundef -> list_repeat (size_chunk_nat chunk) Undef
| Vint n ->
  (match chunk with
   | Mint8signed -> inj_bytes (encode_int (S O) (Int.unsigned n))
   | Mint8unsigned -> inj_bytes (encode_int (S O) (Int.unsigned n))
   | Mint16signed -> inj_bytes (encode_int (S (S O)) (Int.unsigned n))
   | Mint16unsigned -> inj_bytes (encode_int (S (S O)) (Int.unsigned n))
   | Mint32 -> inj_bytes (encode_int (S (S (S (S O)))) (Int.unsigned n))
   | _ -> list_repeat (size_chunk_nat chunk) Undef)
| Vfloat n ->
  (match chunk with
   | Mfloat32 ->
     inj_bytes
       (encode_int (S (S (S (S O)))) (Int.unsigned (Float.bits_of_single n)))
   | Mfloat64 ->
     inj_bytes
       (encode_int (S (S (S (S (S (S (S (S O))))))))
         (Int64.unsigned (Float.bits_of_double n)))
   | Mfloat64al32 ->
     inj_bytes
       (encode_int (S (S (S (S (S (S (S (S O))))))))
         (Int64.unsigned (Float.bits_of_double n)))
   | _ -> list_repeat (size_chunk_nat chunk) Undef)
| Vptr (b, ofs) ->
  (match chunk with
   | Mint32 -> inj_pointer (S (S (S (S O)))) b ofs
   | _ -> list_repeat (size_chunk_nat chunk) Undef)

let decode_val :  memory_chunk -> memval list -> coq_val =
  fun chunk vl  ->
  match proj_bytes vl with
  | Some bl ->
    (match chunk with
     | Mint8signed ->
       Vint
         (Int.sign_ext (Zpos (Coq_xO (Coq_xO (Coq_xO Coq_xH))))
           (Int.repr (decode_int bl)))
     | Mint8unsigned ->
       Vint
         (Int.zero_ext (Zpos (Coq_xO (Coq_xO (Coq_xO Coq_xH))))
           (Int.repr (decode_int bl)))
     | Mint16signed ->
       Vint
         (Int.sign_ext (Zpos (Coq_xO (Coq_xO (Coq_xO (Coq_xO Coq_xH)))))
           (Int.repr (decode_int bl)))
     | Mint16unsigned ->
       Vint
         (Int.zero_ext (Zpos (Coq_xO (Coq_xO (Coq_xO (Coq_xO Coq_xH)))))
           (Int.repr (decode_int bl)))
     | Mint32 -> Vint (Int.repr (decode_int bl))
     | Mfloat32 -> Vfloat (Float.single_of_bits (Int.repr (decode_int bl)))
     | _ -> Vfloat (Float.double_of_bits (Int64.repr (decode_int bl))))
  | None ->
    (match chunk with
     | Mint32 -> proj_pointer vl
     | _ -> Vundef)

