open AST
open Coqlib
open Floats
open Integers
open Memdata

type condition =
| Ccomp of comparison
| Ccompu of comparison
| Ccompimm of comparison * Int.int
| Ccompuimm of comparison * Int.int
| Ccompf of comparison
| Cnotcompf of comparison
| Cmaskzero of Int.int
| Cmasknotzero of Int.int

type addressing =
| Aindexed of Int.int
| Aindexed2 of Int.int
| Ascaled of Int.int * Int.int
| Aindexed2scaled of Int.int * Int.int
| Aglobal of ident * Int.int
| Abased of ident * Int.int
| Abasedscaled of Int.int * ident * Int.int
| Ainstack of Int.int

type operation =
| Omove
| Ointconst of Int.int
| Ofloatconst of float
| Oindirectsymbol of ident
| Ocast8signed
| Ocast8unsigned
| Ocast16signed
| Ocast16unsigned
| Oneg
| Osub
| Omul
| Omulimm of Int.int
| Odiv
| Odivu
| Omod
| Omodu
| Oand
| Oandimm of Int.int
| Oor
| Oorimm of Int.int
| Oxor
| Oxorimm of Int.int
| Oshl
| Oshlimm of Int.int
| Oshr
| Oshrimm of Int.int
| Oshrximm of Int.int
| Oshru
| Oshruimm of Int.int
| Ororimm of Int.int
| Olea of addressing
| Onegf
| Oabsf
| Oaddf
| Osubf
| Omulf
| Odivf
| Osingleoffloat
| Ointoffloat
| Ofloatofint
| Ocmp of condition

let coq_Oaddrsymbol :  ident -> Int.int -> operation =
  fun id ofs  ->
  Olea (Aglobal (id, ofs))

let coq_Oaddrstack :  Int.int -> operation =
  fun ofs  ->
  Olea (Ainstack ofs)

let coq_Oaddimm :  Int.int -> operation =
  fun n  ->
  Olea (Aindexed n)

let eq_addressing :  addressing -> addressing -> bool =
  fun x y  ->
  match x with
  | Aindexed x0 ->
    (match y with
     | Aindexed i0 -> Int.eq_dec x0 i0
     | _ -> false)
  | Aindexed2 x0 ->
    (match y with
     | Aindexed2 i0 -> Int.eq_dec x0 i0
     | _ -> false)
  | Ascaled (x0, x1) ->
    (match y with
     | Ascaled (i1, i2) ->
       if Int.eq_dec x0 i1 then Int.eq_dec x1 i2 else false
     | _ -> false)
  | Aindexed2scaled (x0, x1) ->
    (match y with
     | Aindexed2scaled (i1, i2) ->
       if Int.eq_dec x0 i1 then Int.eq_dec x1 i2 else false
     | _ -> false)
  | Aglobal (x0, x1) ->
    (match y with
     | Aglobal (i1, i2) -> if peq x0 i1 then Int.eq_dec x1 i2 else false
     | _ -> false)
  | Abased (x0, x1) ->
    (match y with
     | Abased (i1, i2) -> if peq x0 i1 then Int.eq_dec x1 i2 else false
     | _ -> false)
  | Abasedscaled (x0, x1, x2) ->
    (match y with
     | Abasedscaled (i2, i3, i4) ->
       if Int.eq_dec x0 i2
       then if peq x1 i3 then Int.eq_dec x2 i4 else false
       else false
     | _ -> false)
  | Ainstack x0 ->
    (match y with
     | Ainstack i0 -> Int.eq_dec x0 i0
     | _ -> false)

let eq_operation :  operation -> operation -> bool =
  fun x y  ->
  let h2 = fun x0 y0 ->
    match x0 with
    | Ceq ->
      (match y0 with
       | Ceq -> true
       | _ -> false)
    | Cne ->
      (match y0 with
       | Cne -> true
       | _ -> false)
    | Clt ->
      (match y0 with
       | Clt -> true
       | _ -> false)
    | Cle ->
      (match y0 with
       | Cle -> true
       | _ -> false)
    | Cgt ->
      (match y0 with
       | Cgt -> true
       | _ -> false)
    | Cge ->
      (match y0 with
       | Cge -> true
       | _ -> false)
  in
  let h3 = fun x0 y0 ->
    match x0 with
    | Ccomp x1 ->
      (match y0 with
       | Ccomp c0 -> h2 x1 c0
       | _ -> false)
    | Ccompu x1 ->
      (match y0 with
       | Ccompu c0 -> h2 x1 c0
       | _ -> false)
    | Ccompimm (x1, x2) ->
      (match y0 with
       | Ccompimm (c0, i0) -> if h2 x1 c0 then Int.eq_dec x2 i0 else false
       | _ -> false)
    | Ccompuimm (x1, x2) ->
      (match y0 with
       | Ccompuimm (c0, i0) -> if h2 x1 c0 then Int.eq_dec x2 i0 else false
       | _ -> false)
    | Ccompf x1 ->
      (match y0 with
       | Ccompf c0 -> h2 x1 c0
       | _ -> false)
    | Cnotcompf x1 ->
      (match y0 with
       | Cnotcompf c0 -> h2 x1 c0
       | _ -> false)
    | Cmaskzero x1 ->
      (match y0 with
       | Cmaskzero i0 -> Int.eq_dec x1 i0
       | _ -> false)
    | Cmasknotzero x1 ->
      (match y0 with
       | Cmasknotzero i0 -> Int.eq_dec x1 i0
       | _ -> false)
  in
  (match x with
   | Omove ->
     (match y with
      | Omove -> true
      | _ -> false)
   | Ointconst x0 ->
     (match y with
      | Ointconst i0 -> Int.eq_dec x0 i0
      | _ -> false)
   | Ofloatconst x0 ->
     (match y with
      | Ofloatconst f0 -> Float.eq_dec x0 f0
      | _ -> false)
   | Oindirectsymbol x0 ->
     (match y with
      | Oindirectsymbol i0 -> peq x0 i0
      | _ -> false)
   | Ocast8signed ->
     (match y with
      | Ocast8signed -> true
      | _ -> false)
   | Ocast8unsigned ->
     (match y with
      | Ocast8unsigned -> true
      | _ -> false)
   | Ocast16signed ->
     (match y with
      | Ocast16signed -> true
      | _ -> false)
   | Ocast16unsigned ->
     (match y with
      | Ocast16unsigned -> true
      | _ -> false)
   | Oneg ->
     (match y with
      | Oneg -> true
      | _ -> false)
   | Osub ->
     (match y with
      | Osub -> true
      | _ -> false)
   | Omul ->
     (match y with
      | Omul -> true
      | _ -> false)
   | Omulimm x0 ->
     (match y with
      | Omulimm i0 -> Int.eq_dec x0 i0
      | _ -> false)
   | Odiv ->
     (match y with
      | Odiv -> true
      | _ -> false)
   | Odivu ->
     (match y with
      | Odivu -> true
      | _ -> false)
   | Omod ->
     (match y with
      | Omod -> true
      | _ -> false)
   | Omodu ->
     (match y with
      | Omodu -> true
      | _ -> false)
   | Oand ->
     (match y with
      | Oand -> true
      | _ -> false)
   | Oandimm x0 ->
     (match y with
      | Oandimm i0 -> Int.eq_dec x0 i0
      | _ -> false)
   | Oor ->
     (match y with
      | Oor -> true
      | _ -> false)
   | Oorimm x0 ->
     (match y with
      | Oorimm i0 -> Int.eq_dec x0 i0
      | _ -> false)
   | Oxor ->
     (match y with
      | Oxor -> true
      | _ -> false)
   | Oxorimm x0 ->
     (match y with
      | Oxorimm i0 -> Int.eq_dec x0 i0
      | _ -> false)
   | Oshl ->
     (match y with
      | Oshl -> true
      | _ -> false)
   | Oshlimm x0 ->
     (match y with
      | Oshlimm i0 -> Int.eq_dec x0 i0
      | _ -> false)
   | Oshr ->
     (match y with
      | Oshr -> true
      | _ -> false)
   | Oshrimm x0 ->
     (match y with
      | Oshrimm i0 -> Int.eq_dec x0 i0
      | _ -> false)
   | Oshrximm x0 ->
     (match y with
      | Oshrximm i0 -> Int.eq_dec x0 i0
      | _ -> false)
   | Oshru ->
     (match y with
      | Oshru -> true
      | _ -> false)
   | Oshruimm x0 ->
     (match y with
      | Oshruimm i0 -> Int.eq_dec x0 i0
      | _ -> false)
   | Ororimm x0 ->
     (match y with
      | Ororimm i0 -> Int.eq_dec x0 i0
      | _ -> false)
   | Olea x0 ->
     (match y with
      | Olea a0 -> eq_addressing x0 a0
      | _ -> false)
   | Onegf ->
     (match y with
      | Onegf -> true
      | _ -> false)
   | Oabsf ->
     (match y with
      | Oabsf -> true
      | _ -> false)
   | Oaddf ->
     (match y with
      | Oaddf -> true
      | _ -> false)
   | Osubf ->
     (match y with
      | Osubf -> true
      | _ -> false)
   | Omulf ->
     (match y with
      | Omulf -> true
      | _ -> false)
   | Odivf ->
     (match y with
      | Odivf -> true
      | _ -> false)
   | Osingleoffloat ->
     (match y with
      | Osingleoffloat -> true
      | _ -> false)
   | Ointoffloat ->
     (match y with
      | Ointoffloat -> true
      | _ -> false)
   | Ofloatofint ->
     (match y with
      | Ofloatofint -> true
      | _ -> false)
   | Ocmp x0 ->
     (match y with
      | Ocmp c0 -> h3 x0 c0
      | _ -> false))

let type_of_condition :  condition -> typ list =
    function
| Ccomp c0 -> Tint :: (Tint :: [])
| Ccompu c0 -> Tint :: (Tint :: [])
| Ccompf c0 -> Tfloat :: (Tfloat :: [])
| Cnotcompf c0 -> Tfloat :: (Tfloat :: [])
| _ -> Tint :: []

let type_of_addressing :  addressing -> typ list =
    function
| Aindexed2 i -> Tint :: (Tint :: [])
| Aindexed2scaled (i, i0) -> Tint :: (Tint :: [])
| Aglobal (i, i0) -> []
| Ainstack i -> []
| _ -> Tint :: []

let type_of_operation :  operation -> typ list * typ =
    function
| Omove -> ([], Tint)
| Ointconst i -> ([], Tint)
| Ofloatconst f -> ([], Tfloat)
| Oindirectsymbol i -> ([], Tint)
| Osub -> ((Tint :: (Tint :: [])), Tint)
| Omul -> ((Tint :: (Tint :: [])), Tint)
| Odiv -> ((Tint :: (Tint :: [])), Tint)
| Odivu -> ((Tint :: (Tint :: [])), Tint)
| Omod -> ((Tint :: (Tint :: [])), Tint)
| Omodu -> ((Tint :: (Tint :: [])), Tint)
| Oand -> ((Tint :: (Tint :: [])), Tint)
| Oor -> ((Tint :: (Tint :: [])), Tint)
| Oxor -> ((Tint :: (Tint :: [])), Tint)
| Oshl -> ((Tint :: (Tint :: [])), Tint)
| Oshr -> ((Tint :: (Tint :: [])), Tint)
| Oshru -> ((Tint :: (Tint :: [])), Tint)
| Olea addr -> ((type_of_addressing addr), Tint)
| Onegf -> ((Tfloat :: []), Tfloat)
| Oabsf -> ((Tfloat :: []), Tfloat)
| Oaddf -> ((Tfloat :: (Tfloat :: [])), Tfloat)
| Osubf -> ((Tfloat :: (Tfloat :: [])), Tfloat)
| Omulf -> ((Tfloat :: (Tfloat :: [])), Tfloat)
| Odivf -> ((Tfloat :: (Tfloat :: [])), Tfloat)
| Osingleoffloat -> ((Tfloat :: []), Tfloat)
| Ointoffloat -> ((Tfloat :: []), Tint)
| Ofloatofint -> ((Tint :: []), Tfloat)
| Ocmp c -> ((type_of_condition c), Tint)
| _ -> ((Tint :: []), Tint)

let is_move_operation :  operation -> 'a1 list -> 'a1 option =
  fun op args  ->
  match op with
  | Omove ->
    (match args with
     | [] -> None
     | arg :: l ->
       (match l with
        | [] -> Some arg
        | a :: l0 -> None))
  | _ -> None

let negate_condition :  condition -> condition =
    function
| Ccomp c -> Ccomp (negate_comparison c)
| Ccompu c -> Ccompu (negate_comparison c)
| Ccompimm (c, n) -> Ccompimm ((negate_comparison c), n)
| Ccompuimm (c, n) -> Ccompuimm ((negate_comparison c), n)
| Ccompf c -> Cnotcompf c
| Cnotcompf c -> Ccompf c
| Cmaskzero n -> Cmasknotzero n
| Cmasknotzero n -> Cmaskzero n

let shift_stack_addressing :  Int.int -> addressing -> addressing =
  fun delta addr  ->
    match addr with
| Ainstack ofs -> Ainstack (Int.add delta ofs)
| _ -> addr

let shift_stack_operation :  Int.int -> operation -> operation =
  fun delta op  ->
    match op with
| Olea addr -> Olea (shift_stack_addressing delta addr)
| _ -> op

let op_for_binary_addressing :  addressing -> operation =
  fun addr  ->
  Olea addr

let two_address_op :  operation -> bool =
    function
| Omove -> false
| Ointconst i -> false
| Ofloatconst f -> false
| Oindirectsymbol i -> false
| Ocast8signed -> false
| Ocast8unsigned -> false
| Ocast16signed -> false
| Ocast16unsigned -> false
| Olea addr -> false
| Osingleoffloat -> false
| Ointoffloat -> false
| Ofloatofint -> false
| Ocmp c -> false
| _ -> true

let is_trivial_op :  operation -> bool =
    function
| Omove -> true
| Ointconst i -> true
| Olea a ->
  (match a with
   | Aglobal (i, i0) -> true
   | Ainstack i -> true
   | _ -> false)
| _ -> false

let op_depends_on_memory :  operation -> bool =
    function
| Ocmp c ->
  (match c with
   | Ccompu c0 -> true
   | _ -> false)
| _ -> false

let addressing_separated :     memory_chunk -> addressing -> memory_chunk -> addressing -> bool =
  fun chunk1 addr1 chunk2 addr2  ->
  match addr1 with
  | Aindexed ofs1 ->
    (match addr2 with
     | Aindexed ofs2 ->
       Int.no_overlap ofs1 (size_chunk chunk1) ofs2 (size_chunk chunk2)
     | _ -> false)
  | Aglobal (s1, ofs1) ->
    (match addr2 with
     | Aglobal (s2, ofs2) ->
       if ident_eq s1 s2
       then Int.no_overlap ofs1 (size_chunk chunk1) ofs2 (size_chunk chunk2)
       else true
     | _ -> false)
  | Abased (s1, ofs1) ->
    (match addr2 with
     | Abased (s2, ofs2) ->
       if ident_eq s1 s2
       then Int.no_overlap ofs1 (size_chunk chunk1) ofs2 (size_chunk chunk2)
       else true
     | _ -> false)
  | Ainstack ofs1 ->
    (match addr2 with
     | Ainstack ofs2 ->
       Int.no_overlap ofs1 (size_chunk chunk1) ofs2 (size_chunk chunk2)
     | _ -> false)
  | _ -> false

