open AST
open Conventions1
open Coqlib
open Datatypes
open Linear
open List0
open Locations
open Machregs
open Op

type equation = { e_reg : mreg; e_slot : slot }

let e_reg :  equation -> mreg =
  fun x  ->
    x.e_reg

let e_slot :  equation -> slot =
  fun x  ->
    x.e_slot

type equations = equation list

let rec find_reg_containing :  slot -> equations -> mreg option =
  fun s  ->
    function
| [] -> None
| e :: eqs' ->
  if slot_eq e.e_slot s then Some e.e_reg else find_reg_containing s eqs'

let eq_equation :  equation -> equation -> bool =
  fun eq1 eq2  ->
  let { e_reg = x; e_slot = x0 } = eq1 in
  let { e_reg = e_reg1; e_slot = e_slot1 } = eq2 in
  if mreg_eq x e_reg1 then slot_eq x0 e_slot1 else false

let contains_equation :  slot -> mreg -> equations -> bool =
  fun s r eqs  ->
  proj_sumbool (in_dec eq_equation { e_reg = r; e_slot = s } eqs)

let rec kill_loc :  loc -> equations -> equations =
  fun l  ->
    function
| [] -> []
| e :: eqs' ->
  if (&&) (proj_sumbool (Loc.diff_dec (S e.e_slot) l))
       (proj_sumbool (Loc.diff_dec (R e.e_reg) l))
  then e :: (kill_loc l eqs')
  else kill_loc l eqs'

let kill_locs :  loc list -> equations -> equations =
  fun ll eqs  ->
  fold_left (fun eqs0 l -> kill_loc l eqs0) ll eqs

let kill_temps :  equations -> equations =
  fun eqs  ->
  kill_locs temporaries eqs

let kill_at_move :  equations -> equations =
  fun eqs  ->
  kill_locs destroyed_at_move eqs

let kill_op :  operation -> equations -> equations =
  fun op eqs  ->
  match op with
  | Omove -> kill_at_move eqs
  | _ -> kill_temps eqs

let is_incoming :  slot -> bool =
    function
| Incoming (z, t) -> true
| _ -> false

let rec safe_move_insertion :  code -> bool =
    function
| [] -> false
| i :: c' ->
  (match i with
   | Lgetstack (s, r) ->
     (&&) (negb (proj_sumbool (in_dec mreg_eq r destroyed_at_move_regs)))
       (safe_move_insertion c')
   | Lsetstack (r, s) ->
     negb (proj_sumbool (in_dec mreg_eq r destroyed_at_move_regs))
   | Lop (op, args, res) ->
     proj_sumbool (list_disjoint_dec mreg_eq args destroyed_at_move_regs)
   | Lload (chunk, addr, args, res) ->
     proj_sumbool (list_disjoint_dec mreg_eq args destroyed_at_move_regs)
   | Lstore (chunk, addr, args, src) ->
     proj_sumbool
       (list_disjoint_dec mreg_eq (src :: args) destroyed_at_move_regs)
   | Lbuiltin (ef, args, res) ->
     proj_sumbool (list_disjoint_dec mreg_eq args destroyed_at_move_regs)
   | Lcond (cond, args, lbl) ->
     proj_sumbool (list_disjoint_dec mreg_eq args destroyed_at_move_regs)
   | Ljumptable (arg, tbl) ->
     negb (proj_sumbool (in_dec mreg_eq arg destroyed_at_move_regs))
   | _ -> false)

let rec transf_code :  equations -> code -> code =
  fun eqs  ->
    function
| [] -> []
| i :: c0 ->
  (match i with
   | Lgetstack (s, r) ->
     if is_incoming s
     then (Lgetstack (s,
            r)) :: (transf_code (kill_loc (R r) (kill_loc (R IT1) eqs)) c0)
     else if contains_equation s r eqs
          then transf_code eqs c0
          else (match find_reg_containing s eqs with
                | Some r' ->
                  if safe_move_insertion c0
                  then (Lop (Omove, (r' :: []),
                         r)) :: (transf_code
                                  (kill_at_move ({ e_reg = r; e_slot =
                                    s } :: (kill_loc (R r) eqs))) c0)
                  else (Lgetstack (s,
                         r)) :: (transf_code ({ e_reg = r; e_slot =
                                  s } :: (kill_loc (R r) eqs)) c0)
                | None ->
                  (Lgetstack (s,
                    r)) :: (transf_code ({ e_reg = r; e_slot =
                             s } :: (kill_loc (R r) eqs)) c0))
   | Lsetstack (r, s) ->
     (Lsetstack (r,
       s)) :: (transf_code
                (kill_at_move ({ e_reg = r; e_slot =
                  s } :: (kill_loc (S s) eqs))) c0)
   | Lop (op, args, res) ->
     (Lop (op, args,
       res)) :: (transf_code (kill_loc (R res) (kill_op op eqs)) c0)
   | Lload (chunk, addr, args, res) ->
     (Lload (chunk, addr, args,
       res)) :: (transf_code (kill_loc (R res) (kill_temps eqs)) c0)
   | Lstore (chunk, addr, args, src) ->
     (Lstore (chunk, addr, args, src)) :: (transf_code (kill_temps eqs) c0)
   | Lbuiltin (ef, args, res) ->
     (Lbuiltin (ef, args,
       res)) :: (transf_code (kill_loc (R res) (kill_temps eqs)) c0)
   | Lannot (ef, args) -> (Lannot (ef, args)) :: (transf_code eqs c0)
   | Lcond (cond, args, lbl) ->
     (Lcond (cond, args, lbl)) :: (transf_code (kill_temps eqs) c0)
   | x -> x :: (transf_code [] c0))

let transf_function :  coq_function -> coq_function =
  fun f  ->
  { fn_sig = f.fn_sig; fn_stacksize = f.fn_stacksize; fn_code =
    (transf_code [] f.fn_code) }

let transf_fundef :  fundef -> fundef =
  fun fd  ->
  transf_fundef transf_function fd

let transf_program :  program -> program =
  fun p  ->
  transform_program transf_fundef p

