open AST
open BinNums
open BinPos
open Cminor
open CminorSel
open Coqlib
open Datatypes
open Errors
open Integers
open List0
open Maps
open Op
open RTL
open Registers
open Switch

type mapping = { map_vars : reg PTree.t; map_letvars : reg list }

let map_vars :  mapping -> reg PTree.t =
  fun x  ->
    x.map_vars

let map_letvars :  mapping -> reg list =
  fun x  ->
    x.map_letvars

type state = { st_nextreg : positive; st_nextnode : positive; st_code : code }

let st_nextreg :  state -> positive =
  fun x  ->
    x.st_nextreg

let st_nextnode :  state -> positive =
  fun x  ->
    x.st_nextnode

let st_code :  state -> code =
  fun x  ->
    x.st_code

type 'a res =
| Error of errmsg
| OK of 'a * state

type 'a mon = state -> 'a res

let handle_error :  'a1 mon -> 'a1 mon -> 'a1 mon =
  fun f g s  ->
  match f s with
  | Error e -> g s
  | OK (a, s') -> OK (a, s')

let init_state :  state =
  { st_nextreg = Coq_xH; st_nextnode = Coq_xH; st_code = PTree.empty }

let add_instr :  instruction -> node mon =
  fun i s  ->
  let n = s.st_nextnode in
  OK (n, { st_nextreg = s.st_nextreg; st_nextnode = (Pos.succ n); st_code =
  (PTree.set n i s.st_code) })

let reserve_instr :  node mon =
  fun s  ->
  let n = s.st_nextnode in
  OK (n, { st_nextreg = s.st_nextreg; st_nextnode = (Pos.succ n); st_code =
  s.st_code })

let check_empty_node :  state -> node -> bool =
  fun s n  ->
  match PTree.get n s.st_code with
  | Some i -> false
  | None -> true

let update_instr :  node -> instruction -> unit mon =
  fun n i s  ->
  if plt n s.st_nextnode
  then if check_empty_node s n
       then OK ((), { st_nextreg = s.st_nextreg; st_nextnode = s.st_nextnode;
              st_code = (PTree.set n i s.st_code) })
       else Error
              (msg
                ('R'::('T'::('L'::('g'::('e'::('n'::('.'::('u'::('p'::('d'::('a'::('t'::('e'::('_'::('i'::('n'::('s'::('t'::('r'::[]))))))))))))))))))))
  else Error
         (msg
           ('R'::('T'::('L'::('g'::('e'::('n'::('.'::('u'::('p'::('d'::('a'::('t'::('e'::('_'::('i'::('n'::('s'::('t'::('r'::[]))))))))))))))))))))

let new_reg :  reg mon =
  fun s  ->
  OK (s.st_nextreg, { st_nextreg = (Pos.succ s.st_nextreg); st_nextnode =
    s.st_nextnode; st_code = s.st_code })

let init_mapping :  mapping =
  { map_vars = PTree.empty; map_letvars = [] }

let add_var :  mapping -> ident -> (reg * mapping) mon =
  fun map name s  ->
  match new_reg s with
  | Error msg0 -> Error msg0
  | OK (a, s') ->
    OK ((a, { map_vars = (PTree.set name a map.map_vars); map_letvars =
      map.map_letvars }), s')

let rec add_vars :  mapping -> ident list -> (reg list * mapping) mon =
  fun map names s  ->
  match names with
  | [] -> OK (([], map), s)
  | n1 :: nl ->
    (match add_vars map nl s with
     | Error msg0 -> Error msg0
     | OK (a, s') ->
       (match add_var (snd a) n1 s' with
        | Error msg0 -> Error msg0
        | OK (a0, s'0) -> OK ((((fst a0) :: (fst a)), (snd a0)), s'0)))

let find_var :  mapping -> ident -> reg mon =
  fun map name s  ->
  match PTree.get name map.map_vars with
  | Some r -> OK (r, s)
  | None ->
    Error ((MSG
      ('R'::('T'::('L'::('g'::('e'::('n'::(':'::(' '::('u'::('n'::('b'::('o'::('u'::('n'::('d'::(' '::('v'::('a'::('r'::('i'::('a'::('b'::('l'::('e'::(' '::[])))))))))))))))))))))))))) :: ((CTX
      name) :: []))

let add_letvar :  mapping -> reg -> mapping =
  fun map r  ->
  { map_vars = map.map_vars; map_letvars = (r :: map.map_letvars) }

let find_letvar :  mapping -> nat -> reg mon =
  fun map idx s  ->
  match nth_error map.map_letvars idx with
  | Some r -> OK (r, s)
  | None ->
    Error
      (msg
        ('R'::('T'::('L'::('g'::('e'::('n'::(':'::(' '::('u'::('n'::('b'::('o'::('u'::('n'::('d'::(' '::('l'::('e'::('t'::(' '::('v'::('a'::('r'::('i'::('a'::('b'::('l'::('e'::[])))))))))))))))))))))))))))))

let alloc_reg :  mapping -> expr -> reg mon =
  fun map  ->
    function
| Evar id -> find_var map id
| Eletvar n -> find_letvar map n
| _ -> new_reg

let rec alloc_regs :  mapping -> exprlist -> reg list mon =
  fun map al s  ->
  match al with
  | Enil -> OK ([], s)
  | Econs (a, bl) ->
    (match alloc_reg map a s with
     | Error msg0 -> Error msg0
     | OK (a0, s') ->
       (match alloc_regs map bl s' with
        | Error msg0 -> Error msg0
        | OK (a1, s'0) -> OK ((a0 :: a1), s'0)))

let alloc_regs_2addr :  mapping -> exprlist -> reg -> reg list mon =
  fun map al rd s  ->
  match al with
  | Enil -> OK ([], s)
  | Econs (a, bl) ->
    (match alloc_regs map bl s with
     | Error msg0 -> Error msg0
     | OK (a0, s') -> OK ((rd :: a0), s'))

let alloc_optreg :  mapping -> ident option -> reg mon =
  fun map  ->
    function
| Some id -> find_var map id
| None -> new_reg

let add_move :  reg -> reg -> node -> node mon =
  fun rs rd nd  ->
  if Reg.eq rs rd
  then (fun s -> OK (nd, s))
  else add_instr (Iop (Omove, (rs :: []), rd, nd))

let rec transl_expr :  mapping -> expr -> reg -> node -> node mon =
  fun map a rd nd s  ->
  match a with
  | Evar v ->
    (match find_var map v s with
     | Error msg0 -> Error msg0
     | OK (a0, s') -> add_move a0 rd nd s')
  | Eop (op, al) ->
    (match if two_address_op op
           then alloc_regs_2addr map al rd s
           else alloc_regs map al s with
     | Error msg0 -> Error msg0
     | OK (a0, s') ->
       (match add_instr (Iop (op, a0, rd, nd)) s' with
        | Error msg0 -> Error msg0
        | OK (a1, s'0) -> transl_exprlist map al a0 a1 s'0))
  | Eload (chunk, addr, al) ->
    (match alloc_regs map al s with
     | Error msg0 -> Error msg0
     | OK (a0, s') ->
       (match add_instr (Iload (chunk, addr, a0, rd, nd)) s' with
        | Error msg0 -> Error msg0
        | OK (a1, s'0) -> transl_exprlist map al a0 a1 s'0))
  | Econdition (cond, al, b, c) ->
    (match alloc_regs map al s with
     | Error msg0 -> Error msg0
     | OK (a0, s') ->
       (match transl_expr map c rd nd s' with
        | Error msg0 -> Error msg0
        | OK (a1, s'0) ->
          (match transl_expr map b rd nd s'0 with
           | Error msg0 -> Error msg0
           | OK (a2, s'1) ->
             (match add_instr (Icond (cond, a0, a2, a1)) s'1 with
              | Error msg0 -> Error msg0
              | OK (a3, s'2) -> transl_exprlist map al a0 a3 s'2))))
  | Elet (b, c) ->
    (match new_reg s with
     | Error msg0 -> Error msg0
     | OK (a0, s') ->
       (match transl_expr (add_letvar map a0) c rd nd s' with
        | Error msg0 -> Error msg0
        | OK (a1, s'0) -> transl_expr map b a0 a1 s'0))
  | Eletvar n ->
    (match find_letvar map n s with
     | Error msg0 -> Error msg0
     | OK (a0, s') -> add_move a0 rd nd s')

and transl_exprlist :     mapping -> exprlist -> reg list -> node -> node mon =
  fun map al rl nd s  ->
  match al with
  | Enil ->
    (match rl with
     | [] -> OK (nd, s)
     | r :: l ->
       Error
         (msg
           ('R'::('T'::('L'::('g'::('e'::('n'::('.'::('t'::('r'::('a'::('n'::('s'::('l'::('_'::('e'::('x'::('p'::('r'::('l'::('i'::('s'::('t'::[]))))))))))))))))))))))))
  | Econs (b, bs) ->
    (match rl with
     | [] ->
       Error
         (msg
           ('R'::('T'::('L'::('g'::('e'::('n'::('.'::('t'::('r'::('a'::('n'::('s'::('l'::('_'::('e'::('x'::('p'::('r'::('l'::('i'::('s'::('t'::[])))))))))))))))))))))))
     | r :: rs ->
       (match transl_exprlist map bs rs nd s with
        | Error msg0 -> Error msg0
        | OK (a, s') -> transl_expr map b r a s'))

let more_likely :  condition -> stmt -> stmt -> bool =
  fun x y z ->
    RTLgenaux.more_likely x y z

let compile_switch :  nat -> table -> comptree =
  fun x y ->
    RTLgenaux.compile_switch x y

let transl_exit :  node list -> nat -> node mon =
  fun nexits n s  ->
  match nth_error nexits n with
  | Some ne -> OK (ne, s)
  | None ->
    Error
      (msg
        ('R'::('T'::('L'::('g'::('e'::('n'::(':'::(' '::('w'::('r'::('o'::('n'::('g'::(' '::('e'::('x'::('i'::('t'::[])))))))))))))))))))

let rec transl_jumptable :  node list -> nat list -> node list mon =
  fun nexits tbl s  ->
  match tbl with
  | [] -> OK ([], s)
  | t1 :: tl ->
    (match transl_exit nexits t1 s with
     | Error msg0 -> Error msg0
     | OK (a, s') ->
       (match transl_jumptable nexits tl s' with
        | Error msg0 -> Error msg0
        | OK (a0, s'0) -> OK ((a :: a0), s'0)))

let rec transl_switch :  reg -> node list -> comptree -> node mon =
  fun r nexits  ->
    function
| CTaction act -> transl_exit nexits act
| CTifeq (key, act, t') ->
  (fun s ->
    match transl_switch r nexits t' s with
    | Error msg0 -> Error msg0
    | OK (a, s') ->
      (match transl_exit nexits act s' with
       | Error msg0 -> Error msg0
       | OK (a0, s'0) ->
         add_instr (Icond ((Ccompimm (Ceq, key)), (r :: []), a0, a)) s'0))
| CTiflt (key, t1, t2) ->
  (fun s ->
    match transl_switch r nexits t2 s with
    | Error msg0 -> Error msg0
    | OK (a, s') ->
      (match transl_switch r nexits t1 s' with
       | Error msg0 -> Error msg0
       | OK (a0, s'0) ->
         add_instr (Icond ((Ccompuimm (Clt, key)), (r :: []), a0, a)) s'0))
| CTjumptable (ofs, sz, tbl, t') ->
  (fun s ->
    match new_reg s with
    | Error msg0 -> Error msg0
    | OK (a, s') ->
      (match transl_jumptable nexits tbl s' with
       | Error msg0 -> Error msg0
       | OK (a0, s'0) ->
         (match add_instr (Ijumptable (a, a0)) s'0 with
          | Error msg0 -> Error msg0
          | OK (a1, s'1) ->
            (match transl_switch r nexits t' s'1 with
             | Error msg0 -> Error msg0
             | OK (a2, s'2) ->
               (match add_instr (Icond ((Ccompuimm (Clt, sz)), (a :: []), a1,
                        a2)) s'2 with
                | Error msg0 -> Error msg0
                | OK (a3, s'3) ->
                  let op =
                    if Int.eq ofs Int.zero
                    then Omove
                    else coq_Oaddimm (Int.neg ofs)
                  in
                  add_instr (Iop (op, (r :: []), a, a3)) s'3)))))

let rec expr_is_2addr_op :  expr -> bool =
    function
| Eop (op, e0) -> two_address_op op
| Econdition (cond, el, e2, e3) ->
  (||) (expr_is_2addr_op e2) (expr_is_2addr_op e3)
| Elet (e1, e2) -> expr_is_2addr_op e2
| _ -> false

type labelmap = node PTree.t

let rec transl_stmt :     mapping -> stmt -> node -> node list -> labelmap -> node -> reg option ->
    node mon =
  fun map s nd nexits ngoto nret rret  ->
  match s with
  | Sskip -> (fun s0 -> OK (nd, s0))
  | Sassign (v, b) ->
    (fun s0 ->
      match find_var map v s0 with
      | Error msg0 -> Error msg0
      | OK (a, s') ->
        if expr_is_2addr_op b
        then (match new_reg s' with
              | Error msg0 -> Error msg0
              | OK (a0, s'0) ->
                (match add_move a0 a nd s'0 with
                 | Error msg0 -> Error msg0
                 | OK (a1, s'1) -> transl_expr map b a0 a1 s'1))
        else transl_expr map b a nd s')
  | Sstore (chunk, addr, al, b) ->
    (fun s0 ->
      match alloc_regs map al s0 with
      | Error msg0 -> Error msg0
      | OK (a, s') ->
        (match alloc_reg map b s' with
         | Error msg0 -> Error msg0
         | OK (a0, s'0) ->
           (match add_instr (Istore (chunk, addr, a, a0, nd)) s'0 with
            | Error msg0 -> Error msg0
            | OK (a1, s'1) ->
              (match transl_expr map b a0 a1 s'1 with
               | Error msg0 -> Error msg0
               | OK (a2, s'2) -> transl_exprlist map al a a2 s'2))))
  | Scall (optid, sig0, s0, cl) ->
    (fun s1 ->
      match s0 with
      | Coq_inl b ->
        (match alloc_reg map b s1 with
         | Error msg0 -> Error msg0
         | OK (a, s') ->
           (match alloc_regs map cl s' with
            | Error msg0 -> Error msg0
            | OK (a0, s'0) ->
              (match alloc_optreg map optid s'0 with
               | Error msg0 -> Error msg0
               | OK (a1, s'1) ->
                 (match add_instr (Icall (sig0, (Coq_inl a), a0, a1, nd)) s'1 with
                  | Error msg0 -> Error msg0
                  | OK (a2, s'2) ->
                    (match transl_exprlist map cl a0 a2 s'2 with
                     | Error msg0 -> Error msg0
                     | OK (a3, s'3) -> transl_expr map b a a3 s'3)))))
      | Coq_inr id ->
        (match alloc_regs map cl s1 with
         | Error msg0 -> Error msg0
         | OK (a, s') ->
           (match alloc_optreg map optid s' with
            | Error msg0 -> Error msg0
            | OK (a0, s'0) ->
              (match add_instr (Icall (sig0, (Coq_inr id), a, a0, nd)) s'0 with
               | Error msg0 -> Error msg0
               | OK (a1, s'1) -> transl_exprlist map cl a a1 s'1))))
  | Stailcall (sig0, s0, cl) ->
    (fun s1 ->
      match s0 with
      | Coq_inl b ->
        (match alloc_reg map b s1 with
         | Error msg0 -> Error msg0
         | OK (a, s') ->
           (match alloc_regs map cl s' with
            | Error msg0 -> Error msg0
            | OK (a0, s'0) ->
              (match add_instr (Itailcall (sig0, (Coq_inl a), a0)) s'0 with
               | Error msg0 -> Error msg0
               | OK (a1, s'1) ->
                 (match transl_exprlist map cl a0 a1 s'1 with
                  | Error msg0 -> Error msg0
                  | OK (a2, s'2) -> transl_expr map b a a2 s'2))))
      | Coq_inr id ->
        (match alloc_regs map cl s1 with
         | Error msg0 -> Error msg0
         | OK (a, s') ->
           (match add_instr (Itailcall (sig0, (Coq_inr id), a)) s' with
            | Error msg0 -> Error msg0
            | OK (a0, s'0) -> transl_exprlist map cl a a0 s'0)))
  | Sbuiltin (optid, ef, al) ->
    (fun s0 ->
      match alloc_regs map al s0 with
      | Error msg0 -> Error msg0
      | OK (a, s') ->
        (match alloc_optreg map optid s' with
         | Error msg0 -> Error msg0
         | OK (a0, s'0) ->
           (match add_instr (Ibuiltin (ef, a, a0, nd)) s'0 with
            | Error msg0 -> Error msg0
            | OK (a1, s'1) -> transl_exprlist map al a a1 s'1)))
  | Sseq (s1, s2) ->
    (fun s0 ->
      match transl_stmt map s2 nd nexits ngoto nret rret s0 with
      | Error msg0 -> Error msg0
      | OK (a, s') -> transl_stmt map s1 a nexits ngoto nret rret s')
  | Sifthenelse (cond, al, strue, sfalse) ->
    (fun s0 ->
      match alloc_regs map al s0 with
      | Error msg0 -> Error msg0
      | OK (a, s') ->
        if more_likely cond strue sfalse
        then (match transl_stmt map sfalse nd nexits ngoto nret rret s' with
              | Error msg0 -> Error msg0
              | OK (a0, s'0) ->
                (match transl_stmt map strue nd nexits ngoto nret rret s'0 with
                 | Error msg0 -> Error msg0
                 | OK (a1, s'1) ->
                   (match add_instr (Icond (cond, a, a1, a0)) s'1 with
                    | Error msg0 -> Error msg0
                    | OK (a2, s'2) -> transl_exprlist map al a a2 s'2)))
        else (match transl_stmt map strue nd nexits ngoto nret rret s' with
              | Error msg0 -> Error msg0
              | OK (a0, s'0) ->
                (match transl_stmt map sfalse nd nexits ngoto nret rret s'0 with
                 | Error msg0 -> Error msg0
                 | OK (a1, s'1) ->
                   (match add_instr (Icond (cond, a, a0, a1)) s'1 with
                    | Error msg0 -> Error msg0
                    | OK (a2, s'2) -> transl_exprlist map al a a2 s'2))))
  | Sloop sbody ->
    (fun s0 ->
      match reserve_instr s0 with
      | Error msg0 -> Error msg0
      | OK (a, s') ->
        (match transl_stmt map sbody a nexits ngoto nret rret s' with
         | Error msg0 -> Error msg0
         | OK (a0, s'0) ->
           (match update_instr a (Inop a0) s'0 with
            | Error msg0 -> Error msg0
            | OK (a1, s'1) -> add_instr (Inop a0) s'1)))
  | Sblock sbody -> transl_stmt map sbody nd (nd :: nexits) ngoto nret rret
  | Sexit n -> transl_exit nexits n
  | Sswitch (a, cases, coqzz_default) ->
    let t0 = compile_switch coqzz_default cases in
    (fun s0 ->
    if validate_switch coqzz_default cases t0
    then (match alloc_reg map a s0 with
          | Error msg0 -> Error msg0
          | OK (a0, s') ->
            (match transl_switch a0 nexits t0 s' with
             | Error msg0 -> Error msg0
             | OK (a1, s'0) -> transl_expr map a a0 a1 s'0))
    else Error
           (msg
             ('R'::('T'::('L'::('g'::('e'::('n'::(':'::(' '::('w'::('r'::('o'::('n'::('g'::(' '::('s'::('w'::('i'::('t'::('c'::('h'::[]))))))))))))))))))))))
  | Sreturn opt_a ->
    (match opt_a with
     | Some a ->
       (match rret with
        | Some r -> transl_expr map a r nret
        | None ->
          (fun s0 -> Error
            (msg
              ('R'::('T'::('L'::('g'::('e'::('n'::(':'::(' '::('t'::('y'::('p'::('e'::(' '::('m'::('i'::('s'::('m'::('a'::('t'::('c'::('h'::(' '::('o'::('n'::(' '::('r'::('e'::('t'::('u'::('r'::('n'::[]))))))))))))))))))))))))))))))))))
     | None -> (fun s0 -> OK (nret, s0)))
  | Slabel (lbl, s') ->
    (fun s0 ->
      match transl_stmt map s' nd nexits ngoto nret rret s0 with
      | Error msg0 -> Error msg0
      | OK (a, s'0) ->
        (match PTree.get lbl ngoto with
         | Some n ->
           (match handle_error (update_instr n (Inop a)) (fun s1 -> Error
                    ((MSG
                    ('M'::('u'::('l'::('t'::('i'::('p'::('l'::('y'::('-'::('d'::('e'::('f'::('i'::('n'::('e'::('d'::(' '::('l'::('a'::('b'::('e'::('l'::(' '::[])))))))))))))))))))))))) :: ((CTX
                    lbl) :: []))) s'0 with
            | Error msg0 -> Error msg0
            | OK (a0, s'1) -> OK (a, s'1))
         | None ->
           Error
             (msg
               ('R'::('T'::('L'::('g'::('e'::('n'::(':'::(' '::('u'::('n'::('b'::('o'::('u'::('n'::('d'::(' '::('l'::('a'::('b'::('e'::('l'::[]))))))))))))))))))))))))
  | Sgoto lbl ->
    (fun s0 ->
      match PTree.get lbl ngoto with
      | Some n -> OK (n, s0)
      | None ->
        Error ((MSG
          ('U'::('n'::('d'::('e'::('f'::('i'::('n'::('e'::('d'::(' '::('d'::('e'::('f'::('i'::('n'::('e'::('d'::(' '::('l'::('a'::('b'::('e'::('l'::(' '::[]))))))))))))))))))))))))) :: ((CTX
          lbl) :: [])))

let alloc_label :  label -> (labelmap * state) -> labelmap * state =
  fun lbl  ->
    function
| (map, s) ->
  let n = s.st_nextnode in
  ((PTree.set lbl n map), { st_nextreg = s.st_nextreg; st_nextnode =
  (Pos.succ s.st_nextnode); st_code = s.st_code })

let rec reserve_labels :  stmt -> (labelmap * state) -> labelmap * state =
  fun s ms  ->
  match s with
  | Sseq (s1, s2) -> reserve_labels s1 (reserve_labels s2 ms)
  | Sifthenelse (cond, al, s1, s2) ->
    reserve_labels s1 (reserve_labels s2 ms)
  | Sloop s1 -> reserve_labels s1 ms
  | Sblock s1 -> reserve_labels s1 ms
  | Slabel (lbl, s1) -> alloc_label lbl (reserve_labels s1 ms)
  | _ -> ms

let ret_reg :  signature -> reg -> reg option =
  fun sig0 rd  ->
  match sig0.sig_res with
  | Some ty -> Some rd
  | None -> None

let transl_fun :     CminorSel.coq_function -> labelmap -> (node * reg list) mon =
  fun f ngoto s  ->
  match add_vars init_mapping (CminorSel.fn_params f) s with
  | Error msg0 -> Error msg0
  | OK (a, s') ->
    (match add_vars (snd a) f.fn_vars s' with
     | Error msg0 -> Error msg0
     | OK (a0, s'0) ->
       (match new_reg s'0 with
        | Error msg0 -> Error msg0
        | OK (a1, s'1) ->
          let orret = ret_reg (CminorSel.fn_sig f) a1 in
          (match add_instr (Ireturn orret) s'1 with
           | Error msg0 -> Error msg0
           | OK (a2, s'2) ->
             (match transl_stmt (snd a0) f.fn_body a2 [] ngoto a2 orret s'2 with
              | Error msg0 -> Error msg0
              | OK (a3, s'3) -> OK ((a3, (fst a)), s'3)))))

let transl_function :     CminorSel.coq_function -> coq_function Errors.res =
  fun f  ->
  let (ngoto, s0) = reserve_labels f.fn_body (PTree.empty, init_state) in
  (match transl_fun f ngoto s0 with
   | Error msg0 -> Errors.Error msg0
   | OK (p, s) ->
     let (nentry, rparams) = p in
     Errors.OK { fn_sig = CminorSel.fn_sig f; fn_params = rparams;
     fn_stacksize = f.fn_stackspace; fn_code = s.st_code; fn_entrypoint =
     nentry })

let transl_fundef :     CminorSel.coq_function AST.fundef -> coq_function AST.fundef Errors.res =
  fun z ->
  transf_partial_fundef transl_function z

let transl_program :  CminorSel.program -> program Errors.res =
  fun p  ->
  transform_partial_program transl_fundef p

