open AST
open BinInt
open BinNums
open Conventions
open Coqlib
open Datatypes
open Errors
open Integers
open List0
open Maps
open Op
open RTL
open Registers

type regenv = reg -> typ

let infer_type_environment :     coq_function -> (node * instruction) list -> regenv option = fun x y ->
    RTLtypingaux.infer_type_environment x y

let check_reg :  regenv -> reg -> typ -> bool =
  fun env r ty  ->
  if typ_eq (env r) ty then true else false

let rec check_regs :  regenv -> reg list -> typ list -> bool =
  fun env rl tyl  ->
  match rl with
  | [] ->
    (match tyl with
     | [] -> true
     | t :: l -> false)
  | r1 :: rs ->
    (match tyl with
     | [] -> false
     | ty :: tys -> (&&) (check_reg env r1 ty) (check_regs env rs tys))

let check_op :  regenv -> operation -> reg list -> reg -> bool =
  fun env op args res0  ->
  let (targs, tres) = type_of_operation op in
  (&&) (check_regs env args targs) (check_reg env res0 tres)

let check_successor :  coq_function -> node -> bool =
  fun funct s  ->
  match PTree.get s funct.fn_code with
  | Some i -> true
  | None -> false

let check_instr :  coq_function -> regenv -> instruction -> bool =
  fun funct env  ->
    function
| Inop s -> check_successor funct s
| Iop (op, args, res0, s) ->
  (match op with
   | Omove ->
     (match args with
      | [] -> false
      | arg :: l ->
        (match l with
         | [] ->
           if typ_eq (env arg) (env res0)
           then check_successor funct s
           else false
         | r :: l0 -> false))
   | _ -> (&&) (check_op env op args res0) (check_successor funct s))
| Iload (chunk, addr, args, dst, s) ->
  (&&)
    ((&&) (check_regs env args (type_of_addressing addr))
      (check_reg env dst (type_of_chunk chunk))) (check_successor funct s)
| Istore (chunk, addr, args, src, s) ->
  (&&)
    ((&&) (check_regs env args (type_of_addressing addr))
      (check_reg env src (type_of_chunk chunk))) (check_successor funct s)
| Icall (sig0, ros, args, res0, s) ->
  (&&)
    ((&&)
      ((&&)
        (match ros with
         | Coq_inl r -> check_reg env r Tint
         | Coq_inr s0 -> true) (check_regs env args sig0.sig_args))
      (check_reg env res0 (proj_sig_res sig0))) (check_successor funct s)
| Itailcall (sig0, ros, args) ->
  (&&)
    ((&&)
      ((&&)
        (match ros with
         | Coq_inl r -> check_reg env r Tint
         | Coq_inr s -> true) (check_regs env args sig0.sig_args))
      (proj_sumbool (opt_typ_eq sig0.sig_res funct.fn_sig.sig_res)))
    (tailcall_is_possible sig0)
| Ibuiltin (ef, args, res0, s) ->
  (&&)
    ((&&)
      ((&&) (check_regs env args (ef_sig ef).sig_args)
        (check_reg env res0 (proj_sig_res (ef_sig ef))))
      (if ef_reloads ef then arity_ok (ef_sig ef).sig_args else true))
    (check_successor funct s)
| Icond (cond, args, s1, s2) ->
  (&&)
    ((&&) (check_regs env args (type_of_condition cond))
      (check_successor funct s1)) (check_successor funct s2)
| Ijumptable (arg, tbl) ->
  (&&) ((&&) (check_reg env arg Tint) (forallb (check_successor funct) tbl))
    (proj_sumbool
      (zle (Z.mul (list_length_z tbl) (Zpos (Coq_xO (Coq_xO Coq_xH))))
        Int.max_unsigned))
| Ireturn optres ->
  (match optres with
   | Some r ->
     (match funct.fn_sig.sig_res with
      | Some t -> check_reg env r t
      | None -> false)
   | None ->
     (match funct.fn_sig.sig_res with
      | Some t -> false
      | None -> true))

let check_params_norepet :  reg list -> bool =
  fun coqzz_params  ->
  if list_norepet_dec Reg.eq coqzz_params then true else false

let rec check_instrs :     coq_function -> regenv -> (node * instruction) list -> bool =
  fun funct env  ->
    function
| [] -> true
| p :: rem ->
  let (pc, i) = p in
  (&&) (check_instr funct env i) (check_instrs funct env rem)

let type_function :  coq_function -> regenv res =
  fun f  ->
  let instrs = PTree.elements f.fn_code in
  (match infer_type_environment f instrs with
   | Some env ->
     if (&&)
          ((&&)
            ((&&) (check_regs env f.fn_params f.fn_sig.sig_args)
              (check_params_norepet f.fn_params))
            (check_instrs f env instrs)) (check_successor f f.fn_entrypoint)
     then OK env
     else Error
            (msg
              ('R'::('T'::('L'::(' '::('t'::('y'::('p'::('e'::(' '::('c'::('h'::('e'::('c'::('k'::('i'::('n'::('g'::(' '::('e'::('r'::('r'::('o'::('r'::[]))))))))))))))))))))))))
   | None ->
     Error
       (msg
         ('R'::('T'::('L'::(' '::('t'::('y'::('p'::('e'::(' '::('i'::('n'::('f'::('e'::('r'::('e'::('n'::('c'::('e'::(' '::('e'::('r'::('r'::('o'::('r'::[]))))))))))))))))))))))))))

