open AST
open Conventions
open Conventions1
open Datatypes
open Integers
open LTLin
open Linear
open List0
open Locations
open Machregs
open Op
open Parallelmove

let reg_for :  loc -> mreg =
    function
| R r -> r
| S s ->
  (match slot_type s with
   | Tint -> IT1
   | Tfloat -> FT1)

let rec regs_for_rec :  loc list -> mreg list -> mreg list -> mreg list =
  fun locs itmps ftmps  ->
  match locs with
  | [] -> []
  | l :: ls ->
    (match l with
     | R r -> r :: (regs_for_rec ls itmps ftmps)
     | S s ->
       (match slot_type s with
        | Tint ->
          (match itmps with
           | [] -> []
           | it1 :: its -> it1 :: (regs_for_rec ls its ftmps))
        | Tfloat ->
          (match ftmps with
           | [] -> []
           | ft1 :: fts -> ft1 :: (regs_for_rec ls itmps fts))))

let regs_for :  loc list -> mreg list =
  fun locs  ->
  regs_for_rec locs int_temporaries float_temporaries

let rec enough_temporaries_rec :     loc list -> mreg list -> mreg list -> bool =
  fun locs itmps ftmps  ->
  match locs with
  | [] -> true
  | l :: ls ->
    (match l with
     | R r -> enough_temporaries_rec ls itmps ftmps
     | S s ->
       (match slot_type s with
        | Tint ->
          (match itmps with
           | [] -> false
           | it1 :: its -> enough_temporaries_rec ls its ftmps)
        | Tfloat ->
          (match ftmps with
           | [] -> false
           | ft1 :: fts -> enough_temporaries_rec ls itmps fts)))

let enough_temporaries :  loc list -> bool =
  fun locs  ->
  enough_temporaries_rec locs int_temporaries float_temporaries

let add_spill :  mreg -> loc -> code -> code =
  fun src dst k  ->
  match dst with
  | R rd -> if mreg_eq src rd then k else (Lop (Omove, (src :: []), rd)) :: k
  | S sl -> (Lsetstack (src, sl)) :: k

let add_reload :  loc -> mreg -> code -> code =
  fun src dst k  ->
  match src with
  | R rs -> if mreg_eq rs dst then k else (Lop (Omove, (rs :: []), dst)) :: k
  | S sl -> (Lgetstack (sl, dst)) :: k

let rec add_reloads :  loc list -> mreg list -> code -> code =
  fun srcs dsts k  ->
  match srcs with
  | [] -> k
  | s1 :: sl ->
    (match dsts with
     | [] -> k
     | t1 :: tl -> add_reload s1 t1 (add_reloads sl tl k))

let add_move :  loc -> loc -> code -> code =
  fun src dst k  ->
  if Loc.eq src dst
  then k
  else (match src with
        | R rs -> add_spill rs dst k
        | S ss ->
          (match dst with
           | R rd -> add_reload src rd k
           | S sd ->
             let tmp =
               match slot_type ss with
               | Tint -> IT1
               | Tfloat -> FT1
             in
             add_reload src tmp (add_spill tmp dst k)))

let parallel_move :  loc list -> loc list -> code -> code =
  fun srcs dsts k  ->
  fold_right (fun p k0 -> add_move (fst p) (snd p) k0) k (parmove srcs dsts)

let transf_instr :     LTLin.coq_function -> LTLin.instruction -> code -> code =
  fun f instr k  ->
  match instr with
  | LTLin.Lop (op, args, res) ->
    (match is_move_operation op args with
     | Some src -> add_move src res k
     | None ->
       let rargs = regs_for args in
       let rres = reg_for res in
       add_reloads args rargs ((Lop (op, rargs,
         rres)) :: (add_spill rres res k)))
  | LTLin.Lload (chunk, addr, args, dst) ->
    let rargs = regs_for args in
    let rdst = reg_for dst in
    add_reloads args rargs ((Lload (chunk, addr, rargs,
      rdst)) :: (add_spill rdst dst k))
  | LTLin.Lstore (chunk, addr, args, src) ->
    if enough_temporaries (src :: args)
    then (match regs_for (src :: args) with
          | [] -> k
          | rsrc :: rargs ->
            add_reloads (src :: args) (rsrc :: rargs) ((Lstore (chunk, addr,
              rargs, rsrc)) :: k))
    else let rargs = regs_for args in
         let rsrc = reg_for src in
         add_reloads args rargs ((Lop ((op_for_binary_addressing addr),
           rargs,
           IT2)) :: (add_reload src rsrc ((Lstore (chunk, (Aindexed
                      Int.zero), (IT2 :: []), rsrc)) :: k)))
  | LTLin.Lcall (sig0, los, args, res) ->
    let largs = loc_arguments sig0 in
    let rres = loc_result sig0 in
    (match los with
     | Coq_inl fn ->
       parallel_move args largs
         (add_reload fn (reg_for fn) ((Lcall (sig0, (Coq_inl
           (reg_for fn)))) :: (add_spill rres res k)))
     | Coq_inr id ->
       parallel_move args largs ((Lcall (sig0, (Coq_inr
         id))) :: (add_spill rres res k)))
  | LTLin.Ltailcall (sig0, los, args) ->
    let largs = loc_arguments sig0 in
    (match los with
     | Coq_inl fn ->
       parallel_move args largs
         (add_reload fn IT1 ((Ltailcall (sig0, (Coq_inl IT1))) :: k))
     | Coq_inr id ->
       parallel_move args largs ((Ltailcall (sig0, (Coq_inr id))) :: k))
  | LTLin.Lbuiltin (ef, args, dst) ->
    if ef_reloads ef
    then let rargs = regs_for args in
         let rdst = reg_for dst in
         add_reloads args rargs ((Lbuiltin (ef, rargs,
           rdst)) :: (add_spill rdst dst k))
    else (Lannot (ef, args)) :: k
  | LTLin.Llabel lbl -> (Llabel lbl) :: k
  | LTLin.Lgoto lbl -> (Lgoto lbl) :: k
  | LTLin.Lcond (cond, args, lbl) ->
    let rargs = regs_for args in
    add_reloads args rargs ((Lcond (cond, rargs, lbl)) :: k)
  | LTLin.Ljumptable (arg, tbl) ->
    let rarg = reg_for arg in
    add_reload arg rarg ((Ljumptable (rarg, tbl)) :: k)
  | LTLin.Lreturn o ->
    (match o with
     | Some loc0 ->
       add_reload loc0 (loc_result (LTLin.fn_sig f)) (Lreturn :: k)
     | None -> Lreturn :: k)

let transf_code :  LTLin.coq_function -> LTLin.code -> code =
  fun f c  ->
  fold_right (transf_instr f) [] c

let transf_function :  LTLin.coq_function -> coq_function =
  fun f  ->
  { fn_sig = LTLin.fn_sig f;
    fn_stacksize = LTLin.fn_stacksize f;
    fn_code = (parallel_move (loc_parameters (LTLin.fn_sig f)) f.fn_params
      (transf_code f (LTLin.fn_code f))) }

let transf_fundef :  LTLin.fundef -> fundef =
  fun fd  ->
  transf_fundef transf_function fd

let transf_program :  LTLin.program -> program =
  fun p  ->
  transform_program transf_fundef p

