open AST
open Cminor
open CminorSel
open Datatypes
open Globalenvs
open Integers
open Op
open SelectOp

let condition_of_expr :  expr -> condition * exprlist =
  fun e  ->
    match e with
| Eop (o, el) ->
  (match o with
   | Ocmp c -> (c, el)
   | _ -> ((Ccompuimm (Cne, Int.zero)), (Econs (e, Enil))))
| _ -> ((Ccompuimm (Cne, Int.zero)), (Econs (e, Enil)))

let load :  memory_chunk -> expr -> expr =
  fun chunk e1  ->
  let (mode, args) = addressing chunk e1 in Eload (chunk, mode, args)

let store :  memory_chunk -> expr -> expr -> stmt =
  fun chunk e1 e2  ->
  let (mode, args) = addressing chunk e1 in Sstore (chunk, mode, args, e2)

let sel_constant :  constant -> expr =
    function
| Cminor.Ointconst n -> Eop ((Ointconst n), Enil)
| Cminor.Ofloatconst f -> Eop ((Ofloatconst f), Enil)
| Oaddrsymbol (id, ofs) -> addrsymbol id ofs
| Oaddrstack ofs -> addrstack ofs

let sel_unop :  unary_operation -> expr -> expr =
  fun op arg  ->
  match op with
  | Cminor.Ocast8unsigned -> cast8unsigned arg
  | Cminor.Ocast8signed -> cast8signed arg
  | Cminor.Ocast16unsigned -> cast16unsigned arg
  | Cminor.Ocast16signed -> cast16signed arg
  | Onegint -> negint arg
  | Onotint -> notint arg
  | Cminor.Onegf -> negf arg
  | Cminor.Oabsf -> absf arg
  | Cminor.Osingleoffloat -> singleoffloat arg
  | Cminor.Ointoffloat -> intoffloat arg
  | Ointuoffloat -> intuoffloat arg
  | Cminor.Ofloatofint -> floatofint arg
  | Ofloatofintu -> floatofintu arg

let sel_binop :  binary_operation -> expr -> expr -> expr =
  fun op arg1 arg2  ->
  match op with
  | Oadd -> add arg1 arg2
  | Cminor.Osub -> sub arg1 arg2
  | Cminor.Omul -> mul arg1 arg2
  | Cminor.Odiv -> divs arg1 arg2
  | Cminor.Odivu -> divu arg1 arg2
  | Cminor.Omod -> mods arg1 arg2
  | Cminor.Omodu -> modu arg1 arg2
  | Cminor.Oand -> coq_and arg1 arg2
  | Cminor.Oor -> coq_or arg1 arg2
  | Cminor.Oxor -> xor arg1 arg2
  | Cminor.Oshl -> shl arg1 arg2
  | Cminor.Oshr -> shr arg1 arg2
  | Cminor.Oshru -> shru arg1 arg2
  | Cminor.Oaddf -> addf arg1 arg2
  | Cminor.Osubf -> subf arg1 arg2
  | Cminor.Omulf -> mulf arg1 arg2
  | Cminor.Odivf -> divf arg1 arg2
  | Cminor.Ocmp c -> comp c arg1 arg2
  | Ocmpu c -> compu c arg1 arg2
  | Ocmpf c -> compf c arg1 arg2

let rec sel_expr :  Cminor.expr -> expr =
    function
| Cminor.Evar id -> Evar id
| Econst cst -> sel_constant cst
| Eunop (op, arg) -> sel_unop op (sel_expr arg)
| Ebinop (op, arg1, arg2) -> sel_binop op (sel_expr arg1) (sel_expr arg2)
| Cminor.Eload (chunk, addr) -> load chunk (sel_expr addr)

let rec sel_exprlist :  Cminor.expr list -> exprlist =
    function
| [] -> Enil
| a :: bl -> Econs ((sel_expr a), (sel_exprlist bl))

type call_kind =
| Call_default
| Call_imm of ident
| Call_builtin of external_function

let expr_is_addrof_ident :  Cminor.expr -> ident option =
    function
| Econst c ->
  (match c with
   | Oaddrsymbol (id, ofs) -> if Int.eq ofs Int.zero then Some id else None
   | _ -> None)
| _ -> None

let classify_call :  genv -> Cminor.expr -> call_kind =
  fun ge e  ->
  match expr_is_addrof_ident e with
  | Some id ->
    (match Genv.find_symbol ge id with
     | Some b ->
       (match Genv.find_funct_ptr ge b with
        | Some f ->
          (match f with
           | Internal f0 -> Call_imm id
           | External ef ->
             if ef_inline ef then Call_builtin ef else Call_imm id)
        | None -> Call_imm id)
     | None -> Call_imm id)
  | None -> Call_default

let rec sel_stmt :  genv -> Cminor.stmt -> stmt =
  fun ge  ->
    function
| Cminor.Sskip -> Sskip
| Cminor.Sassign (id, e) -> Sassign (id, (sel_expr e))
| Cminor.Sstore (chunk, addr, rhs) ->
  store chunk (sel_expr addr) (sel_expr rhs)
| Cminor.Scall (optid, sg, fn, args) ->
  (match classify_call ge fn with
   | Call_default ->
     Scall (optid, sg, (Coq_inl (sel_expr fn)), (sel_exprlist args))
   | Call_imm id -> Scall (optid, sg, (Coq_inr id), (sel_exprlist args))
   | Call_builtin ef -> Sbuiltin (optid, ef, (sel_exprlist args)))
| Cminor.Stailcall (sg, fn, args) ->
  (match classify_call ge fn with
   | Call_imm id -> Stailcall (sg, (Coq_inr id), (sel_exprlist args))
   | _ -> Stailcall (sg, (Coq_inl (sel_expr fn)), (sel_exprlist args)))
| Cminor.Sbuiltin (optid, ef, args) ->
  Sbuiltin (optid, ef, (sel_exprlist args))
| Cminor.Sseq (s1, s2) -> Sseq ((sel_stmt ge s1), (sel_stmt ge s2))
| Cminor.Sifthenelse (e, ifso, ifnot) ->
  let (cond, args) = condition_of_expr (sel_expr e) in
  Sifthenelse (cond, args, (sel_stmt ge ifso), (sel_stmt ge ifnot))
| Cminor.Sloop body -> Sloop (sel_stmt ge body)
| Cminor.Sblock body -> Sblock (sel_stmt ge body)
| Cminor.Sexit n -> Sexit n
| Cminor.Sswitch (e, cases, dfl) -> Sswitch ((sel_expr e), cases, dfl)
| Cminor.Sreturn o ->
  (match o with
   | Some e -> Sreturn (Some (sel_expr e))
   | None -> Sreturn None)
| Cminor.Slabel (lbl, body) -> Slabel (lbl, (sel_stmt ge body))
| Cminor.Sgoto lbl -> Sgoto lbl

let sel_function :  genv -> Cminor.coq_function -> coq_function =
  fun ge f  ->
  { fn_sig = Cminor.fn_sig f;
    fn_params = Cminor.fn_params f;
    fn_vars = Cminor.fn_vars f;
    fn_stackspace = Cminor.fn_stackspace f;
    fn_body = sel_stmt ge (Cminor.fn_body f) }

let sel_fundef :  genv -> Cminor.fundef -> fundef =
  fun ge f  ->
  transf_fundef (sel_function ge) f

let sel_program :  Cminor.program -> program =
  fun p  ->
  let ge = Genv.globalenv p in transform_program (sel_fundef ge) p

