open AST
open BinInt
open BinNums
open Bounds
open Conventions1
open Coqlib
open Errors
open Integers
open Linear
open List0
open Locations
open Mach
open Machregs
open Op
open Stacklayout

type frame_index =
| FI_link
| FI_retaddr
| FI_local of coq_Z * typ
| FI_arg of coq_Z * typ
| FI_saved_int of coq_Z
| FI_saved_float of coq_Z

let offset_of_index :  frame_env -> frame_index -> coq_Z =
  fun fe  ->
    function
| FI_link -> fe.fe_ofs_link
| FI_retaddr -> fe.fe_ofs_retaddr
| FI_local (x, t) ->
  (match t with
   | Tint ->
     Z.add fe.fe_ofs_int_local (Z.mul (Zpos (Coq_xO (Coq_xO Coq_xH))) x)
   | Tfloat ->
     Z.add fe.fe_ofs_float_local
       (Z.mul (Zpos (Coq_xO (Coq_xO (Coq_xO Coq_xH)))) x))
| FI_arg (x, ty) ->
  Z.add fe_ofs_arg (Z.mul (Zpos (Coq_xO (Coq_xO Coq_xH))) x)
| FI_saved_int x ->
  Z.add fe.fe_ofs_int_callee_save (Z.mul (Zpos (Coq_xO (Coq_xO Coq_xH))) x)
| FI_saved_float x ->
  Z.add fe.fe_ofs_float_callee_save
    (Z.mul (Zpos (Coq_xO (Coq_xO (Coq_xO Coq_xH)))) x)

let save_callee_save_reg :     (frame_env -> coq_Z) -> (mreg -> coq_Z) -> (coq_Z -> frame_index) -> typ
    -> frame_env -> mreg -> code -> instruction list =
  fun bound number mkindex ty fe cs k  ->
  let i = number cs in
  if zlt i (bound fe)
  then (Msetstack (cs, (Int.repr (offset_of_index fe (mkindex i))), ty)) :: k
  else k

let save_callee_save_regs :     (frame_env -> coq_Z) -> (mreg -> coq_Z) -> (coq_Z -> frame_index) -> typ
    -> frame_env -> mreg list -> code -> code =
  fun bound number mkindex ty fe csl k  ->
  fold_right (save_callee_save_reg bound number mkindex ty fe) k csl

let save_callee_save_int :  frame_env -> code -> code =
  fun fe z ->
  save_callee_save_regs fe_num_int_callee_save index_int_callee_save
    (fun x -> FI_saved_int x) Tint fe int_callee_save_regs z

let save_callee_save_float :  frame_env -> code -> code =
  fun fe z ->
  save_callee_save_regs fe_num_float_callee_save index_float_callee_save
    (fun x -> FI_saved_float x) Tfloat fe float_callee_save_regs z

let save_callee_save :  frame_env -> code -> code =
  fun fe k  ->
  save_callee_save_int fe (save_callee_save_float fe k)

let restore_callee_save_reg :     (frame_env -> coq_Z) -> (mreg -> coq_Z) -> (coq_Z -> frame_index) -> typ
    -> frame_env -> mreg -> code -> instruction list =
  fun bound number mkindex ty fe cs k  ->
  let i = number cs in
  if zlt i (bound fe)
  then (Mgetstack ((Int.repr (offset_of_index fe (mkindex i))), ty, cs)) :: k
  else k

let restore_callee_save_regs :     (frame_env -> coq_Z) -> (mreg -> coq_Z) -> (coq_Z -> frame_index) -> typ
    -> frame_env -> mreg list -> code -> code =
  fun bound number mkindex ty fe csl k  ->
  fold_right (restore_callee_save_reg bound number mkindex ty fe) k csl

let restore_callee_save_int :  frame_env -> code -> code =
  fun fe z ->
  restore_callee_save_regs fe_num_int_callee_save index_int_callee_save
    (fun x -> FI_saved_int x) Tint fe int_callee_save_regs z

let restore_callee_save_float :  frame_env -> code -> code =
  fun fe z ->
  restore_callee_save_regs fe_num_float_callee_save index_float_callee_save
    (fun x -> FI_saved_float x) Tfloat fe float_callee_save_regs z

let restore_callee_save :  frame_env -> code -> code =
  fun fe k  ->
  restore_callee_save_int fe (restore_callee_save_float fe k)

let transl_op :  frame_env -> operation -> operation =
  fun fe op  ->
  shift_stack_operation (Int.repr fe.fe_stack_data) op

let transl_addr :  frame_env -> addressing -> addressing =
  fun fe addr  ->
  shift_stack_addressing (Int.repr fe.fe_stack_data) addr

let transl_annot_param :  frame_env -> loc -> annot_param =
  fun fe  ->
    function
| R r -> APreg r
| S s ->
  (match s with
   | Local (ofs, ty) ->
     APstack ((chunk_of_type ty), (offset_of_index fe (FI_local (ofs, ty))))
   | _ -> APstack (Mint32, (Zneg Coq_xH)))

let transl_instr :  frame_env -> Linear.instruction -> code -> code =
  fun fe i k  ->
  match i with
  | Lgetstack (s, r) ->
    (match s with
     | Local (ofs, ty) ->
       (Mgetstack ((Int.repr (offset_of_index fe (FI_local (ofs, ty)))), ty,
         r)) :: k
     | Incoming (ofs, ty) ->
       (Mgetparam ((Int.repr (offset_of_index fe (FI_arg (ofs, ty)))), ty,
         r)) :: k
     | Outgoing (ofs, ty) ->
       (Mgetstack ((Int.repr (offset_of_index fe (FI_arg (ofs, ty)))), ty,
         r)) :: k)
  | Lsetstack (r, s) ->
    (match s with
     | Local (ofs, ty) ->
       (Msetstack (r, (Int.repr (offset_of_index fe (FI_local (ofs, ty)))),
         ty)) :: k
     | Incoming (ofs, ty) -> k
     | Outgoing (ofs, ty) ->
       (Msetstack (r, (Int.repr (offset_of_index fe (FI_arg (ofs, ty)))),
         ty)) :: k)
  | Lop (op, args, res0) -> (Mop ((transl_op fe op), args, res0)) :: k
  | Lload (chunk, addr, args, dst) ->
    (Mload (chunk, (transl_addr fe addr), args, dst)) :: k
  | Lstore (chunk, addr, args, src) ->
    (Mstore (chunk, (transl_addr fe addr), args, src)) :: k
  | Lcall (sig0, ros) -> (Mcall (sig0, ros)) :: k
  | Ltailcall (sig0, ros) ->
    restore_callee_save fe ((Mtailcall (sig0, ros)) :: k)
  | Lbuiltin (ef, args, dst) -> (Mbuiltin (ef, args, dst)) :: k
  | Lannot (ef, args) ->
    (Mannot (ef, (map (transl_annot_param fe) args))) :: k
  | Llabel lbl -> (Mlabel lbl) :: k
  | Lgoto lbl -> (Mgoto lbl) :: k
  | Lcond (cond, args, lbl) -> (Mcond (cond, args, lbl)) :: k
  | Ljumptable (arg, tbl) -> (Mjumptable (arg, tbl)) :: k
  | Lreturn -> restore_callee_save fe (Mreturn :: k)

let transl_code :  frame_env -> Linear.instruction list -> code =
  fun fe il  ->
  fold_right (transl_instr fe) [] il

let transl_body :  Linear.coq_function -> frame_env -> code =
  fun f fe  ->
  save_callee_save fe (transl_code fe (Linear.fn_code f))

let transf_function :  Linear.coq_function -> coq_function res =
  fun f  ->
  let fe = make_env (function_bounds f) in
  if zlt Int.max_unsigned fe.fe_size
  then Error
         (msg
           ('T'::('o'::('o'::(' '::('m'::('a'::('n'::('y'::(' '::('s'::('p'::('i'::('l'::('l'::('e'::('d'::(' '::('v'::('a'::('r'::('i'::('a'::('b'::('l'::('e'::('s'::(','::(' '::('s'::('t'::('a'::('c'::('k'::(' '::('s'::('i'::('z'::('e'::(' '::('e'::('x'::('c'::('e'::('e'::('d'::('e'::('d'::[]))))))))))))))))))))))))))))))))))))))))))))))))
  else OK { fn_sig = Linear.fn_sig f; fn_code = (transl_body f fe);
         fn_stacksize = fe.fe_size; fn_link_ofs = (Int.repr fe.fe_ofs_link);
         fn_retaddr_ofs = (Int.repr fe.fe_ofs_retaddr) }

let transf_fundef :  Linear.fundef -> fundef res =
  fun f  ->
  transf_partial_fundef transf_function f

let transf_program :  Linear.program -> program res =
  fun p  ->
  transform_partial_program transf_fundef p

