open BinInt
open BinNums
open Coqlib
open Datatypes
open EqNat
open Integers
open Maps

module IntIndexed = 
 struct 
  type t = Int.int
  
  (** val index : Int.int -> positive **)
  
  let index n =
    match Int.unsigned n with
    | Z0 -> Coq_xH
    | Zpos p -> Coq_xO p
    | Zneg p -> Coq_xI p
  
  (** val eq : Int.int -> Int.int -> bool **)
  
  let eq x y =
    Int.eq_dec x y
 end

module IntMap = 
 struct 
  type elt = IntIndexed.t
  
  (** val elt_eq : IntIndexed.t -> IntIndexed.t -> bool **)
  
  let elt_eq x y =
    IntIndexed.eq x y
  
  type 'x t = 'x PMap.t
  
  (** val init : 'a1 -> 'a1 * 'a1 PTree.t **)
  
  let init x =
    PMap.init x
  
  (** val get : IntIndexed.t -> 'a1 t -> 'a1 **)
  
  let get i m =
    PMap.get (IntIndexed.index i) m
  
  (** val set : IntIndexed.t -> 'a1 -> 'a1 t -> 'a1 * 'a1 PTree.t **)
  
  let set i v m =
    PMap.set (IntIndexed.index i) v m
  
  (** val map : ('a1 -> 'a2) -> 'a1 t -> 'a2 t **)
  
  let map f m =
    PMap.map f m
 end

type table = (Int.int * nat) list

type comptree =
| CTaction of nat
| CTifeq of Int.int * nat * comptree
| CTiflt of Int.int * comptree * comptree
| CTjumptable of Int.int * Int.int * nat list * comptree

let rec split_lt :  Int.int -> table -> table * table =
  fun pivot  ->
    function
| [] -> ([], [])
| p :: rem ->
  let (key, act) = p in
  let (l, r) = split_lt pivot rem in
  if Int.ltu key pivot
  then (((key, act) :: l), r)
  else (l, ((key, act) :: r))

let rec split_eq :  Int.int -> table -> nat option * table =
  fun pivot  ->
    function
| [] -> (None, [])
| p :: rem ->
  let (key, act) = p in
  let (same, others) = split_eq pivot rem in
  if Int.eq key pivot
  then ((Some act), others)
  else (same, ((key, act) :: others))

let rec split_between :     nat -> Int.int -> Int.int -> table -> nat IntMap.t * table =
  fun dfl ofs sz  ->
    function
| [] -> ((IntMap.init dfl), [])
| p :: rem ->
  let (key, act) = p in
  let (inside, outside) = split_between dfl ofs sz rem in
  if Int.ltu (Int.sub key ofs) sz
  then ((IntMap.set key act inside), outside)
  else (inside, ((key, act) :: outside))

let refine_low_bound :  coq_Z -> coq_Z -> coq_Z =
  fun v lo  ->
  if zeq v lo then Z.add lo (Zpos Coq_xH) else lo

let refine_high_bound :  coq_Z -> coq_Z -> coq_Z =
  fun v hi  ->
  if zeq v hi then Z.sub hi (Zpos Coq_xH) else hi

let rec validate_jumptable :  nat IntMap.t -> nat list -> Int.int -> bool =
  fun cases tbl n  ->
  match tbl with
  | [] -> true
  | act :: rem ->
    (&&) (beq_nat act (IntMap.get n cases))
      (validate_jumptable cases rem (Int.add n Int.one))

let rec validate :  nat -> table -> comptree -> coq_Z -> coq_Z -> bool =
  fun coqzz_default cases t0 lo hi  ->
  match t0 with
  | CTaction act ->
    (match cases with
     | [] -> beq_nat act coqzz_default
     | p :: l ->
       let (key1, act1) = p in
       (&&)
         ((&&) (proj_sumbool (zeq (Int.unsigned key1) lo))
           (proj_sumbool (zeq lo hi))) (beq_nat act act1))
  | CTifeq (pivot, act, t') ->
    let (o, others) = split_eq pivot cases in
    (match o with
     | Some act' ->
       (&&) (beq_nat act act')
         (validate coqzz_default others t'
           (refine_low_bound (Int.unsigned pivot) lo)
           (refine_high_bound (Int.unsigned pivot) hi))
     | None -> false)
  | CTiflt (pivot, t1, t2) ->
    let (lcases, rcases) = split_lt pivot cases in
    (&&)
      (validate coqzz_default lcases t1 lo
        (Z.sub (Int.unsigned pivot) (Zpos Coq_xH)))
      (validate coqzz_default rcases t2 (Int.unsigned pivot) hi)
  | CTjumptable (ofs, sz, tbl, t') ->
    let tbl_len = list_length_z tbl in
    let (inside, outside) = split_between coqzz_default ofs sz cases in
    (&&)
      ((&&)
        ((&&) (proj_sumbool (zle (Int.unsigned sz) tbl_len))
          (proj_sumbool (zle tbl_len Int.max_signed)))
        (validate_jumptable inside tbl ofs))
      (validate coqzz_default outside t' lo hi)

let validate_switch :  nat -> table -> comptree -> bool =
  fun coqzz_default cases t0 ->
  validate coqzz_default cases t0 Z0 Int.max_unsigned

