open AST
open BinInt
open BinNums
open Coqlib
open Datatypes
open Floats
open Integers

type block = coq_Z

let eq_block :  coq_Z -> coq_Z -> bool =
fun x0 x1  -> zeq x0 x1 

type coq_val =
| Vundef
| Vint of Int.int
| Vfloat of float
| Vptr of block * Int.int

let coq_Vzero :  coq_val =
  Vint Int.zero

let coq_Vtrue :  coq_val =
  Vint Int.one

let coq_Vfalse :  coq_val =
  Vint Int.zero

module Val = 
 struct 
  (** val neg : coq_val -> coq_val **)
  
  let neg = function
  | Vint n -> Vint (Int.neg n)
  | _ -> Vundef
  
  (** val negf : coq_val -> coq_val **)
  
  let negf = function
  | Vfloat f -> Vfloat (Float.neg f)
  | _ -> Vundef
  
  (** val absf : coq_val -> coq_val **)
  
  let absf = function
  | Vfloat f -> Vfloat (Float.abs f)
  | _ -> Vundef
  
  (** val maketotal : coq_val option -> coq_val **)
  
  let maketotal = function
  | Some v -> v
  | None -> Vundef
  
  (** val intoffloat : coq_val -> coq_val option **)
  
  let intoffloat = function
  | Vfloat f -> option_map (fun x -> Vint x) (Float.intoffloat f)
  | _ -> None
  
  (** val intuoffloat : coq_val -> coq_val option **)
  
  let intuoffloat = function
  | Vfloat f -> option_map (fun x -> Vint x) (Float.intuoffloat f)
  | _ -> None
  
  (** val floatofint : coq_val -> coq_val option **)
  
  let floatofint = function
  | Vint n -> Some (Vfloat (Float.floatofint n))
  | _ -> None
  
  (** val floatofintu : coq_val -> coq_val option **)
  
  let floatofintu = function
  | Vint n -> Some (Vfloat (Float.floatofintu n))
  | _ -> None
  
  (** val floatofwords : coq_val -> coq_val -> coq_val **)
  
  let floatofwords v1 v2 =
    match v1 with
    | Vint n1 ->
      (match v2 with
       | Vint n2 -> Vfloat (Float.from_words n1 n2)
       | _ -> Vundef)
    | _ -> Vundef
  
  (** val negint : coq_val -> coq_val **)
  
  let negint = function
  | Vint n -> Vint (Int.neg n)
  | _ -> Vundef
  
  (** val notint : coq_val -> coq_val **)
  
  let notint = function
  | Vint n -> Vint (Int.not n)
  | _ -> Vundef
  
  (** val of_bool : bool -> coq_val **)
  
  let of_bool = function
  | true -> coq_Vtrue
  | false -> coq_Vfalse
  
  (** val boolval : coq_val -> coq_val **)
  
  let boolval = function
  | Vint n -> of_bool (negb (Int.eq n Int.zero))
  | Vptr (b, ofs) -> coq_Vtrue
  | _ -> Vundef
  
  (** val notbool : coq_val -> coq_val **)
  
  let notbool = function
  | Vint n -> of_bool (Int.eq n Int.zero)
  | Vptr (b, ofs) -> coq_Vfalse
  | _ -> Vundef
  
  (** val zero_ext : coq_Z -> coq_val -> coq_val **)
  
  let zero_ext nbits = function
  | Vint n -> Vint (Int.zero_ext nbits n)
  | _ -> Vundef
  
  (** val sign_ext : coq_Z -> coq_val -> coq_val **)
  
  let sign_ext nbits = function
  | Vint n -> Vint (Int.sign_ext nbits n)
  | _ -> Vundef
  
  (** val singleoffloat : coq_val -> coq_val **)
  
  let singleoffloat = function
  | Vfloat f -> Vfloat (Float.singleoffloat f)
  | _ -> Vundef
  
  (** val add : coq_val -> coq_val -> coq_val **)
  
  let add v1 v2 =
    match v1 with
    | Vint n1 ->
      (match v2 with
       | Vint n2 -> Vint (Int.add n1 n2)
       | Vptr (b2, ofs2) -> Vptr (b2, (Int.add ofs2 n1))
       | _ -> Vundef)
    | Vptr (b1, ofs1) ->
      (match v2 with
       | Vint n2 -> Vptr (b1, (Int.add ofs1 n2))
       | _ -> Vundef)
    | _ -> Vundef
  
  (** val sub : coq_val -> coq_val -> coq_val **)
  
  let sub v1 v2 =
    match v1 with
    | Vint n1 ->
      (match v2 with
       | Vint n2 -> Vint (Int.sub n1 n2)
       | _ -> Vundef)
    | Vptr (b1, ofs1) ->
      (match v2 with
       | Vint n2 -> Vptr (b1, (Int.sub ofs1 n2))
       | Vptr (b2, ofs2) ->
         if zeq b1 b2 then Vint (Int.sub ofs1 ofs2) else Vundef
       | _ -> Vundef)
    | _ -> Vundef
  
  (** val mul : coq_val -> coq_val -> coq_val **)
  
  let mul v1 v2 =
    match v1 with
    | Vint n1 ->
      (match v2 with
       | Vint n2 -> Vint (Int.mul n1 n2)
       | _ -> Vundef)
    | _ -> Vundef
  
  (** val divs : coq_val -> coq_val -> coq_val option **)
  
  let divs v1 v2 =
    match v1 with
    | Vint n1 ->
      (match v2 with
       | Vint n2 ->
         if (||) (Int.eq n2 Int.zero)
              ((&&) (Int.eq n1 (Int.repr Int.min_signed))
                (Int.eq n2 Int.mone))
         then None
         else Some (Vint (Int.divs n1 n2))
       | _ -> None)
    | _ -> None
  
  (** val mods : coq_val -> coq_val -> coq_val option **)
  
  let mods v1 v2 =
    match v1 with
    | Vint n1 ->
      (match v2 with
       | Vint n2 ->
         if (||) (Int.eq n2 Int.zero)
              ((&&) (Int.eq n1 (Int.repr Int.min_signed))
                (Int.eq n2 Int.mone))
         then None
         else Some (Vint (Int.mods n1 n2))
       | _ -> None)
    | _ -> None
  
  (** val divu : coq_val -> coq_val -> coq_val option **)
  
  let divu v1 v2 =
    match v1 with
    | Vint n1 ->
      (match v2 with
       | Vint n2 ->
         if Int.eq n2 Int.zero then None else Some (Vint (Int.divu n1 n2))
       | _ -> None)
    | _ -> None
  
  (** val modu : coq_val -> coq_val -> coq_val option **)
  
  let modu v1 v2 =
    match v1 with
    | Vint n1 ->
      (match v2 with
       | Vint n2 ->
         if Int.eq n2 Int.zero then None else Some (Vint (Int.modu n1 n2))
       | _ -> None)
    | _ -> None
  
  (** val add_carry : coq_val -> coq_val -> coq_val -> coq_val **)
  
  let add_carry v1 v2 cin =
    match v1 with
    | Vint n1 ->
      (match v2 with
       | Vint n2 ->
         (match cin with
          | Vint c -> Vint (Int.add_carry n1 n2 c)
          | _ -> Vundef)
       | _ -> Vundef)
    | _ -> Vundef
  
  (** val coq_and : coq_val -> coq_val -> coq_val **)
  
  let coq_and v1 v2 =
    match v1 with
    | Vint n1 ->
      (match v2 with
       | Vint n2 -> Vint (Int.coq_and n1 n2)
       | _ -> Vundef)
    | _ -> Vundef
  
  (** val coq_or : coq_val -> coq_val -> coq_val **)
  
  let coq_or v1 v2 =
    match v1 with
    | Vint n1 ->
      (match v2 with
       | Vint n2 -> Vint (Int.coq_or n1 n2)
       | _ -> Vundef)
    | _ -> Vundef
  
  (** val xor : coq_val -> coq_val -> coq_val **)
  
  let xor v1 v2 =
    match v1 with
    | Vint n1 ->
      (match v2 with
       | Vint n2 -> Vint (Int.xor n1 n2)
       | _ -> Vundef)
    | _ -> Vundef
  
  (** val shl : coq_val -> coq_val -> coq_val **)
  
  let shl v1 v2 =
    match v1 with
    | Vint n1 ->
      (match v2 with
       | Vint n2 ->
         if Int.ltu n2 Int.iwordsize then Vint (Int.shl n1 n2) else Vundef
       | _ -> Vundef)
    | _ -> Vundef
  
  (** val shr : coq_val -> coq_val -> coq_val **)
  
  let shr v1 v2 =
    match v1 with
    | Vint n1 ->
      (match v2 with
       | Vint n2 ->
         if Int.ltu n2 Int.iwordsize then Vint (Int.shr n1 n2) else Vundef
       | _ -> Vundef)
    | _ -> Vundef
  
  (** val shr_carry : coq_val -> coq_val -> coq_val **)
  
  let shr_carry v1 v2 =
    match v1 with
    | Vint n1 ->
      (match v2 with
       | Vint n2 ->
         if Int.ltu n2 Int.iwordsize
         then Vint (Int.shr_carry n1 n2)
         else Vundef
       | _ -> Vundef)
    | _ -> Vundef
  
  (** val shrx : coq_val -> coq_val -> coq_val option **)
  
  let shrx v1 v2 =
    match v1 with
    | Vint n1 ->
      (match v2 with
       | Vint n2 ->
         if Int.ltu n2
              (Int.repr (Zpos (Coq_xI (Coq_xI (Coq_xI (Coq_xI Coq_xH))))))
         then Some (Vint (Int.shrx n1 n2))
         else None
       | _ -> None)
    | _ -> None
  
  (** val shru : coq_val -> coq_val -> coq_val **)
  
  let shru v1 v2 =
    match v1 with
    | Vint n1 ->
      (match v2 with
       | Vint n2 ->
         if Int.ltu n2 Int.iwordsize then Vint (Int.shru n1 n2) else Vundef
       | _ -> Vundef)
    | _ -> Vundef
  
  (** val rolm : coq_val -> Int.int -> Int.int -> coq_val **)
  
  let rolm v amount mask =
    match v with
    | Vint n -> Vint (Int.rolm n amount mask)
    | _ -> Vundef
  
  (** val ror : coq_val -> coq_val -> coq_val **)
  
  let ror v1 v2 =
    match v1 with
    | Vint n1 ->
      (match v2 with
       | Vint n2 ->
         if Int.ltu n2 Int.iwordsize then Vint (Int.ror n1 n2) else Vundef
       | _ -> Vundef)
    | _ -> Vundef
  
  (** val addf : coq_val -> coq_val -> coq_val **)
  
  let addf v1 v2 =
    match v1 with
    | Vfloat f1 ->
      (match v2 with
       | Vfloat f2 -> Vfloat (Float.add f1 f2)
       | _ -> Vundef)
    | _ -> Vundef
  
  (** val subf : coq_val -> coq_val -> coq_val **)
  
  let subf v1 v2 =
    match v1 with
    | Vfloat f1 ->
      (match v2 with
       | Vfloat f2 -> Vfloat (Float.sub f1 f2)
       | _ -> Vundef)
    | _ -> Vundef
  
  (** val mulf : coq_val -> coq_val -> coq_val **)
  
  let mulf v1 v2 =
    match v1 with
    | Vfloat f1 ->
      (match v2 with
       | Vfloat f2 -> Vfloat (Float.mul f1 f2)
       | _ -> Vundef)
    | _ -> Vundef
  
  (** val divf : coq_val -> coq_val -> coq_val **)
  
  let divf v1 v2 =
    match v1 with
    | Vfloat f1 ->
      (match v2 with
       | Vfloat f2 -> Vfloat (Float.div f1 f2)
       | _ -> Vundef)
    | _ -> Vundef
  
  (** val cmp_bool : comparison -> coq_val -> coq_val -> bool option **)
  
  let cmp_bool c v1 v2 =
    match v1 with
    | Vint n1 ->
      (match v2 with
       | Vint n2 -> Some (Int.cmp c n1 n2)
       | _ -> None)
    | _ -> None
  
  (** val cmp_different_blocks : comparison -> bool option **)
  
  let cmp_different_blocks = function
  | Ceq -> Some false
  | Cne -> Some true
  | _ -> None
  
  (** val cmpu_bool :
      (block -> coq_Z -> bool) -> comparison -> coq_val -> coq_val -> bool
      option **)
  
  let cmpu_bool valid_ptr c v1 v2 =
    let weak_valid_ptr = fun b ofs ->
      (||) (valid_ptr b ofs) (valid_ptr b (Z.sub ofs (Zpos Coq_xH)))
    in
    (
    match v1 with
    | Vint n1 ->
      (match v2 with
       | Vint n2 -> Some (Int.cmpu c n1 n2)
       | Vptr (b2, ofs2) ->
         if Int.eq n1 Int.zero then cmp_different_blocks c else None
       | _ -> None)
    | Vptr (b1, ofs1) ->
      (match v2 with
       | Vint n2 ->
         if Int.eq n2 Int.zero then cmp_different_blocks c else None
       | Vptr (b2, ofs2) ->
         if zeq b1 b2
         then if (&&) (weak_valid_ptr b1 (Int.unsigned ofs1))
                   (weak_valid_ptr b2 (Int.unsigned ofs2))
              then Some (Int.cmpu c ofs1 ofs2)
              else None
         else if (&&) (valid_ptr b1 (Int.unsigned ofs1))
                   (valid_ptr b2 (Int.unsigned ofs2))
              then cmp_different_blocks c
              else None
       | _ -> None)
    | _ -> None)
  
  (** val cmpf_bool : comparison -> coq_val -> coq_val -> bool option **)
  
  let cmpf_bool c v1 v2 =
    match v1 with
    | Vfloat f1 ->
      (match v2 with
       | Vfloat f2 -> Some (Float.cmp c f1 f2)
       | _ -> None)
    | _ -> None
  
  (** val of_optbool : bool option -> coq_val **)
  
  let of_optbool = function
  | Some b -> if b then coq_Vtrue else coq_Vfalse
  | None -> Vundef
  
  (** val cmp : comparison -> coq_val -> coq_val -> coq_val **)
  
  let cmp c v1 v2 =
    of_optbool (cmp_bool c v1 v2)
  
  (** val cmpu :
      (block -> coq_Z -> bool) -> comparison -> coq_val -> coq_val -> coq_val **)
  
  let cmpu valid_ptr c v1 v2 =
    of_optbool (cmpu_bool valid_ptr c v1 v2)
  
  (** val cmpf : comparison -> coq_val -> coq_val -> coq_val **)
  
  let cmpf c v1 v2 =
    of_optbool (cmpf_bool c v1 v2)
  
  (** val load_result : memory_chunk -> coq_val -> coq_val **)
  
  let load_result chunk v =
    match chunk with
    | Mint8signed ->
      (match v with
       | Vint n ->
         Vint (Int.sign_ext (Zpos (Coq_xO (Coq_xO (Coq_xO Coq_xH)))) n)
       | _ -> Vundef)
    | Mint8unsigned ->
      (match v with
       | Vint n ->
         Vint (Int.zero_ext (Zpos (Coq_xO (Coq_xO (Coq_xO Coq_xH)))) n)
       | _ -> Vundef)
    | Mint16signed ->
      (match v with
       | Vint n ->
         Vint
           (Int.sign_ext (Zpos (Coq_xO (Coq_xO (Coq_xO (Coq_xO Coq_xH))))) n)
       | _ -> Vundef)
    | Mint16unsigned ->
      (match v with
       | Vint n ->
         Vint
           (Int.zero_ext (Zpos (Coq_xO (Coq_xO (Coq_xO (Coq_xO Coq_xH))))) n)
       | _ -> Vundef)
    | Mint32 ->
      (match v with
       | Vfloat f -> Vundef
       | x -> x)
    | Mfloat32 ->
      (match v with
       | Vfloat f -> Vfloat (Float.singleoffloat f)
       | _ -> Vundef)
    | _ ->
      (match v with
       | Vfloat f -> Vfloat f
       | _ -> Vundef)
 end

type meminj = block -> (block * coq_Z) option

